
/***************************************************************************
 *                                                                         *
 *   chan_dgv.c - Channel Driver for Asterisk (c)                          *
 *                                                                         *
 *   Copyright (c) 2006-2007, Digivoice Tecnologia em Eletronica Ltda      *
 *                                                                         *
 *   Module: Asterisk Channel for Digivoice Cards                          *
 *                                                                         *
 *   Author: DigiVoice Tecnologia em Eletronica Ltda                       *
 *   desenvolvimento@digivoice.com.br                                      *
 *                                                                         *
 *   This software is licensed under GPL                                   *
 *   Check LICENSE file for details                                        *
 *                                                                         *
 ***************************************************************************/

//ONLY TO VERSION 0.9.0, 0.9.1, 0.9.2, 0.9.3, 0.9.4, 0.9.5 AND 0.9.6
//TO USE THE CHANNEL WITH ISDN SIGNALLING UNCOMMENT THE LINE "//#define USE_LIBPRI" LIKE "#define USE_LIBPRI" AND RECOMPILE ALL

//#define USE_LIBPRI

#define CHANDGV_VERSION      "0.9.6_rc7"
#define VOICERLIB_DEP        4095                 //VoicerLib dependency Version

//#define DEBUG_Q

#ifdef USE_LIBPRI
	//#define DEBUG_PRI      1
#endif//#ifdef USE_LIBPRI

#define DETAILED_VERBOSE     8

#define	DATALEN              160

#define	GSM_DATALEN          33

#define TX_CUT_FACTOR        33                   //8 = 160ms   , 16 = 320ms, etc...

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include <voicerlib/voicerlib.h>

#include <asterisk.h>

#define NUM_SPANS            12

#define SILENCE_TIMEOUT      3000

#define DEFAULT_MOHINTERPRET "default"

#define PRI_DEBUG_LEVEL      PRI_DEBUG_ALL /*PRI_DEBUG_Q931_ANOMALY*/ /*PRI_DEBUG_Q921_STATE PRI_DEBUG_ALL */

#include "config.h"

#ifdef VOICERLIB_VERSIONNUMBER
    #if VOICERLIB_VERSIONNUMBER < VOICERLIB_DEP
    	#error "**** Incompatible VoicerLib Version. Please check the dependencies ****"
    #endif//#if VOICERLIB_VERSIONNUMBER < VOICERLIB_DEP
#else//#ifdef VOICERLIB_VERSIONNUMBER
    #error "**** Incompatible VoicerLib Version. Please check the dependencies ****"
#endif//#ifdef VOICERLIB_VERSIONNUMBER

u32 last_time_pri;

ASTERISK_FILE_VERSION(__FILE__, "$Revision: 4772 $")

#include <pthread.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <sys/ioctl.h>

#include <asterisk/lock.h>
#include <asterisk/utils.h>
#include <asterisk/channel.h>
#include <asterisk/config.h>
#include <asterisk/logger.h>
#include <asterisk/module.h>
#include <asterisk/pbx.h>
#include <asterisk/options.h>
#include <asterisk/callerid.h>
#include <asterisk/dsp.h>
#include <asterisk/features.h>
#include <asterisk/causes.h>
#include <asterisk/cli.h>
#include <asterisk/sched.h>
#include <asterisk/translate.h>
#include <asterisk/app.h>
#include <asterisk/transcap.h>
#include <asterisk/musiconhold.h>

#ifdef USE_LIBPRI
    #include <libpri.h>
#endif//#ifdef USE_LIBPRI

#ifdef CC_AST_HAS_VERSION_1_4
	#include "asterisk/abstract_jb.h"
	#define AST_MODULE "chan_dgv"	
#else//#ifdef CC_AST_HAS_VERSION_1_4
	static const char desc[] = "Digivoice Boards Support";
#endif//#ifdef CC_AST_HAS_VERSION_1_4

#include "dgvchannel.h"

//static const char desc[] = "Digivoice Boards Support";
static const char type[] = "DGV";
static const char tdesc[] = "DiGiVoice voice cards driver";
static const char config[] = "digivoice.conf";

#define START_PBX		1
#define DONT_START_PBX	0

//default echo cancelation taps
#define DEFAULT_TAPS    ECHO_TAPS_64

static ast_cond_t timer_pending_cond;

static const char *auto_scan_port_desc[] = {"auto","manual"};

static const char *taps_desc[] = { "ECHO OFF","ECHO_TAPS_32","ECHO_TAPS_64","ECHO_TAPS_128","ECHO_TAPS_256","ECHO_TAPS_512"};

static const char *sync_mode_desc[] = {"","SYNC_LINE_A","SYNC_LINE_B","SYNC_INTERNAL"};

static const char *consolelanguage_desc[] = {"English", "Portuguese"};

static const char *showalarmstatus_desc[] = {"Off", "On "};

/*const short senoide[] = { 0x0000, 0x09cb, 0x121a, 0x17a6, 0x1999, 0x17a6, 0x121a, 
                                 0x09cb, 0x0000, 0xf635, 0xede6, 0xe85a, 
                                 0xe667, 0xe85a, 0xede6, 0xf635 };*/

/*const char senoide[] = { 0xd5,0x9d,0x83,0x88,
                          0xb5,0x88,0x83,0x9d,
                          0x55,0x1d,0x03,0x08,
                          0x35,0x08,0x03,0x1d }; senoide 500hz alaw   */

const char senoide[] = { 0xd5, 0xed, 0x9d, 0x84, 0x83, 0x8f, 0x88, 0x8a, 0x8a, 0x8a, 0x88, 0x8f, 0x83, 0x84, 0x9d, 0xed, 
                         0x55, 0x6d, 0x1d, 0x04, 0x03, 0x0f, 0x08, 0x0a, 0x35, 0x0a, 0x08, 0x0f, 0x03, 0x04, 0x1d, 0x6d };

/* Default language */
static char global_language[MAX_LANGUAGE] = "";

/* Signalling types */
/* These types will be used to differentiate cards and signalling types of
the same card */
enum {
	DG_SIG_FXO,
	DG_SIG_R2_NDIS_FIXED,
    DG_SIG_R2_NDIS_VARIABLE,
    DG_SIG_PRI,
    DG_SIG_CUSTOMCAS,
    DG_SIG_FXS
};

/* used to identify span interface type - ccs or cas */
enum {
	DG_SPAN_CCS,
    DG_SPAN_CAS
};

//enum used to index span structure
enum {
    DG_SPAN_NUMBER,
    DG_FRAME_TYPE,
    DG_SIDE,
    DG_SWITCHTYPE,
    DG_OVERLAPDIAL,
    DG_PRIDIALPLAN,
    DG_PRILOCALDIALPLAN
};

static struct sched_context *sched;

static short channel_up = 0;		//flag indicating that channel is up and running

static int usecnt =0;

int supported_fmts = AST_FORMAT_SLINEAR | AST_FORMAT_ULAW | AST_FORMAT_ALAW |  AST_FORMAT_GSM;

#ifndef CC_AST_HAS_VERSION_1_4
	AST_MUTEX_DEFINE_STATIC(usecnt_lock);
#endif//#ifndef CC_AST_HAS_VERSION_1_4

/* Protect the interface list (of ast_modem_pvt's) */
//AST_MUTEX_DEFINE_STATIC(iflock);
ast_mutex_t iflock;
/* these are used to wake up the timer thread when there's work to do */
//AST_MUTEX_DEFINE_STATIC(timer_pending_lock);
ast_mutex_t timer_pending_lock;

static short span_type[NUM_SPANS];

#ifdef USE_LIBPRI
	/* global variables to config cards */
	static short span_pritype[NUM_SPANS];
	static short span_switchtype[NUM_SPANS];
	static short span_overlapdial[NUM_SPANS];
	static short span_pridialplan[NUM_SPANS];
	static short span_prilocaldialplan[NUM_SPANS];
	
	static int thd_socket_running = 0;
	static int thd_pri_running = 0;

	ast_mutex_t ccs_lock;

	//static int dialplan = PRI_NATIONAL_ISDN + 1;
	//static int localdialplan = PRI_NATIONAL_ISDN + 1;

	static int dev_idx = 0;//used only in set_card_config
#endif//#ifdef USE_LIBPRI

/*! \brief How long to wait for following digits */
static int gendigittimeout = 7000;

/*! \brief How long to wait for an extra digit, if there is an ambiguous match */
static int matchdigittimeout = 3000;

static int ringtimeout = 7000;                   //How long to wait for other ring (FXO logic)

static struct {
	short auto_scan_port;						 //1- auto allocates ports (default) - 0 - manual allocations (future implementation)
	//types
	short default_dialtype;
	short default_detectiontype;

	char szDefaultCallProgressFile[TAMSTR];

	short afterdialpause;
	short ringbacktone1;
	char szRingbacktone1Name[TAMSTR];
	char szRingbacktone2Name[TAMSTR];
	char szBusytone1Name[TAMSTR];
	char szBusytone2Name[TAMSTR];
	short ringbacktone2;
	short busytone1;
	short busytone2;
  	unsigned int ringback_times[4];			     //time for ringback
	unsigned int busy_times[4];					 //times for busy
	//E1 configurations
	char szDefPortID[TAMSTR];
	//-----
	
    //codec available
	short allow_slinear;
	short allow_ulaw;
	short allow_alaw;
	short allow_gsm;
	
	short consolelanguage;                       //console language
} dgv_conf;

typedef struct {
	int nALARM_RSLIP;
    int nALARM_RAIS;
    int nALARM_AISS;
    int nALARM_AIS16S;
    int nALARM_LOSS;
    int nALARM_CRC4SYNC;
    int nALARM_MFSYNC;
    int nALARM_SYNC;
} tddgv_e1alarms;

tddgv_e1alarms dgv_e1alarms[MAX_CARDS][2];

static struct {
	char fxs_atxfer[5];                          //atxfer
} features_conf;		

//static char context[AST_MAX_EXTENSION] = "default";

/* functions */


static char *GetCardName(short nCard);
void dgv_cleanup(void);
static struct dgv_pvt *find_ch_port(short port);
static struct dgv_pvt *find_free_port(short group, short backwards, short dc);
static struct dgv_pvt *initialize_ports(short port);
void start_dial(struct dgv_pvt *p);
static struct ast_channel *dgv_new(struct dgv_pvt *i, int state, char *context, short startpbx, int transfercapability);
static struct ast_channel *dgv_request(const char *type, int format, void *data, int *cause);

#ifdef CC_AST_HAS_VERSION_1_4
	static int dgv_digit_begin(struct ast_channel *ast, char digit);
#endif//#ifdef CC_AST_HAS_VERSION_1_4

#if defined(CC_AST_HAS_VERSION_1_4) && defined(CC_AST_HAS_SEND_DIGIT_END_DURATION)
	static int dgv_digit(struct ast_channel *ast, char digit, unsigned int duration);
#else//#if defined(CC_AST_HAS_VERSION_1_4) && defined(CC_AST_HAS_SEND_DIGIT_END_DURATION)
	static int dgv_digit(struct ast_channel *ast, char digit);
#endif//#if defined(CC_AST_HAS_VERSION_1_4) && defined(CC_AST_HAS_SEND_DIGIT_END_DURATION)

static int dgv_call(struct ast_channel *ast, char *dest, int timeout);
static int dgv_hangup(struct ast_channel *ast);
static int dgv_answer(struct ast_channel *ast);

#ifdef CC_AST_HAS_INDICATE_DATA
	static int dgv_indicate(struct ast_channel *ast, int condition, const void *data, size_t datalen);
#else//#ifdef CC_AST_HAS_INDICATE_DATA
	static int dgv_indicate(struct ast_channel *ast, int condition);
#endif//#ifdef CC_AST_HAS_INDICATE_DATA

static int dgv_write(struct ast_channel *ast, struct ast_frame *frame);
static int dgv_fixup(struct ast_channel *oldchan, struct ast_channel *newchan);
static struct ast_frame *dgv_read(struct ast_channel *ast);

short set_card_config(char *s,short sync, short use_crc4);
short set_port_config(short reload, char *s, char *port_id, char *context, short category, short min_rx,
					  short max_rx,  short interdigit_timeout, short txgain, short rxgain, short cid_pos, 
                      short echo_can, char *accountcode, short signalling, short b_subscriber_hangup, short group_b_value, 
                      short silence_threshold, char *musicclass, short pickupgroup, short gen_dtmf, short gen_ringback, char *callprogress, 
                      char *language, short dtmf_to_asterisk, short dialtone_timeout, short wait_dialtone, short gen_busy,
                      short drop_tx_samples_factor, short relaxdtmf, short echo_training, 
                      short ring_type, short flash_min_time, short flash_max_time, short immediate, char *mailbox,
                      short fax_detection);
short set_port_group(short group, char *s);
short tone_lookup(char *tone_str);
unsigned int dgv_gettick(void);
int __reload(short reset);
static int __unload_module(int reload);
int reload(void);
int unload_module(void);
int load_module(void);
static int digit_timeout_callback(const void *data);
static int dialtone_timeout_callback(const void *data);
static int ring_timeout_callback(const void *data);
static int all_echotraining_callback(const void *data);

static void *fv_thread(void *data); /*Send frametype as AST_FRAME_VOICE; ***zt_read*/
static void *fxss_thread(void *data);//thread used to outgoing call(fxs card)

#ifdef USE_LIBPRI
	static void *ss_thread(void *data);

	int dg_pri_findindex(short card, short span);
	int dg_pri_findpri(struct pri *pri);
	int dg_pri_findfds(int fds);
	int dg_pri_to_port(int pri_channel);
	int port_to_pri(int port);

	void ccs_handler(short card, short span, short size, void *SampleData);

	static pthread_t thd_sock;
	static pthread_t thd_pri;

	static char internationalprefix[10] = "";
	static char nationalprefix[10] = "";
	static char localprefix[20] = "";
	static char privateprefix[20] = "";
	static char unknownprefix[20] = "";
#endif//#ifdef USE_LIBPRI

static int use_callerid = 1;
static char customcascfg[TAMSTR] = "avaya-definity-line-side.cfg";
static char featurescfg[TAMSTR] = "features.conf";

//static int e1_category_timeout(void *data);

static void *ToneThread(void *param);
static void *StutterToneThread(void *param);

void save_log(short port, char *msg);

void QueueFrame(short port, struct ast_frame *fp);

//Callbacks
void event_handler(void *context_data);
void read_samples(short port, short size, void *SampleData);

//voicerlib definitions & functions
dg_event_data_structure event_context;
int setup_dgv(int reload);

static pthread_t timer_thread = AST_PTHREADT_NULL;

static struct ast_channel_tech dgv_tech = {
	.type = type,
	.description = tdesc,
	.capabilities = AST_FORMAT_GSM,
	.requester = dgv_request,
	.call = dgv_call,
	.hangup = dgv_hangup,
	.answer = dgv_answer,
	.read = dgv_read,
	.write = dgv_write,
	.bridged_channel = NULL,	
	.bridge = NULL,
    .fixup = dgv_fixup,    
#ifdef CC_AST_HAS_VERSION_1_4
		.send_digit_begin = dgv_digit_begin,
		.send_digit_end = dgv_digit,   
#else//#ifdef CC_AST_HAS_VERSION_1_4
    	.send_digit = dgv_digit,
#endif//#ifdef CC_AST_HAS_VERSION_1_4
  .indicate = dgv_indicate
};

#ifdef USE_LIBPRI
	//message handler
	static void primsg(struct pri *pri, char *s)
	{
#ifdef DEBUG_PRI
			FILE *f;
    		char szTemp[TAMSTR];

    		struct tm *now_dbg;
    		time_t curSecs;
    		struct timeval tv;

	    	gettimeofday(&tv, NULL);
	    	u32 timenow =  tv.tv_sec*1000L+tv.tv_usec/1000L;

    		curSecs = time(NULL);				
    		now_dbg = localtime(&curSecs);	
    
    		sprintf(szTemp,"*%02d:%02d:%02d.%06d*%s" ,now_dbg->tm_hour, now_dbg->tm_min, now_dbg->tm_sec,timenow,s);
    
	    	f = fopen("/var/log/voicerlib/digi_pri", "a+");
    		fprintf(f,"%s",szTemp);
    		fflush(f);
    		fclose(f);
	
    		last_time_pri = timenow;
#endif//#ifdef DEBUG_PRI
	}

	//error message handler
	static void prierr(struct pri *pri, char *s)
	{
    	char szTemp[TAMSTR];
    	sprintf(szTemp, "dgv_pri_msg: %s", s);
    
	    if (option_verbose >= 3)
    		ast_log(LOG_WARNING, szTemp);
	}
#endif//#ifdef USE_LIBPRI

static char *GetCardName(short nCard)
{
	switch(dg_GetCardType(nCard))
	{
		case VBE13060PCI:
			return "E13060 rv.1\0";
	    case VBE13030PCI:
    	    return "E13030\0";
		case VB0408PCI:
			return "0408PCI\0";
		case VBE13060PCI_R:
			return "E13060 rv.2\0";
		case VBE16060PCI:
		case VBE16060PCI_R:
			return "E16060 slim\0";
		case VB0404FX:
			return "0404FX\0";
		default:
			return "Unknown type";
	}
}

//------------------------------------------------------------
// Convert audio format constants - return string description
//------------------------------------------------------------
static inline char * ast2dgvformatname(int ast_format)
{
	switch(ast_format) {
		case AST_FORMAT_ALAW:
			return "AST_FORMAT_ALAW:ffWaveALaw";
		case AST_FORMAT_SLINEAR:
			return "AST_FORMAT_SLINEAR:ffWavePCM";
		case AST_FORMAT_ULAW:
			return "AST_FORMAT_ULAW:ffWaveULaw";
		case AST_FORMAT_GSM:
			return "AST_FORMAT_GSM:ffGsm610";
		default:
			return "UNKN:UNKN";
	}
}
//------------------------------------------------------------
// Convert signalling constants - return string description
//------------------------------------------------------------
static inline char *sig2str(int sig)
{
    switch (sig) {
        case DG_SIG_FXO:
            return "FXO";
        case DG_SIG_R2_NDIS_FIXED:
            return "E1 R2D/MFC - NDIS Fixed";
        case DG_SIG_R2_NDIS_VARIABLE:
            return "E1 R2D/MFC - NDIS Variable";
        case DG_SIG_PRI:
            return "E1 ISDN/PRI";
        case DG_SIG_CUSTOMCAS:
            return "E1 CustomCAS";
        case DG_SIG_FXS:
            return "FXS";        	
        default:
            return "Unknown";
   }
}

//linetone timeout - hangup and free channel
static int dialtone_timeout_callback(const void *data)
{
	struct dgv_pvt *p = (struct dgv_pvt *)data;

	if (p) 
    {
        if (p->call_state == DG_AST_WAITFORDIALTONE)
        {
            //no dial tone
            dg_DisableCallProgress(p->port);

            p->linetone_timeout_sched = -1;

            switch(p->sig)
            {    
                case DG_SIG_PRI:
                	//pri never waits for dialtone
                	p->wait_dialtone = 0;
                    break;
    			default:
    				break;
            }

            if (p->wait_dialtone) {
                if (p->verbosity==50)
                    ast_verbose(VERBOSE_PREFIX_3 "Dgv/%d: No dial tone.\n", p->port);

                if (p->ast_ch!=NULL)
                    ast_softhangup(p->ast_ch, AST_SOFTHANGUP_TIMEOUT);
            }
            else
            {
                //dial anyway
                start_dial(p);
            }
        }
    }
    return 0;
}

//ring timeout - hangup and free channel
static int ring_timeout_callback(const void *data)
{
	struct dgv_pvt *p = (struct dgv_pvt *)data;

	if (p) 
    {
    	if (p->verbosity==50)
        	ast_log(LOG_NOTICE, "ring_timeout_callback on port %d \n", p->port);
        	
     	p->ring_timeout_sched = -1;
    	
        if (p->call_state == DG_AST_RINGING)
        {
            switch(p->sig)
            { 
                case DG_SIG_FXO:
                	if (p->ast_ch!=NULL)
                    	ast_softhangup(p->ast_ch, AST_SOFTHANGUP_TIMEOUT);
                    	
                    break;
            }
        }
    }
    
    return 0;
}

//timeout function
static int digit_timeout_callback(const void *data)
{
	struct dgv_pvt *p = (struct dgv_pvt *)data;

	if (p) 
	{
		if (p->sig != DG_SIG_FXS)
		{	
	        if (p->verbosity==50)
    	        ast_log(LOG_WARNING,"********** Port %d receives schedule - id %x!!!!\n",p->port,p->digit_timeout_sched);
        	p->digit_timeout_sched = -1;
	        if (ast_exists_extension(NULL, p->context, p->exten, 1, p->cid_num))
    	    {
        	    //extension found! accept call
            	if (p->verbosity==50)
                	ast_log(LOG_NOTICE,"digit_timeout: Extension (%s@%s) found!!!.\n", p->exten, p->context);
	            //when send groupB you're saying that you receive enough numbers
    	        dg_R2AskForGroupII(p->port, DG_A3_PULSED);
        	}
	        else
    	    {
        	    if (p->verbosity==50)
            	    ast_log(LOG_NOTICE,"digit_timeout (%d): No more possible extensions. Dropping call.\n", p->port);
            	dg_HangUp(p->port);
	            if ((p->sig==DG_SIG_FXO) || (p->sig==DG_SIG_CUSTOMCAS) || (p->sig==DG_SIG_FXS))
    	            p->reserved=0;
        	    else
            	    p->reserved=2;      //wait receiving idle
	        }
		}
	}
	
	return 0;
}

//all_echotraining_callback function
static int all_echotraining_callback(const void *data)
{
	struct dgv_pvt *p = (struct dgv_pvt *)data;

	if (p) 
	{
		if (p->call_state == DG_AST_CONVERSATION)
		{
			if (p->verbosity==50)
       		  	ast_log(LOG_NOTICE, "all_echotraining (%d): Starting dg_SetEchoCancelationAdapt(4) \n", p->port);

       		dg_SetEchoCancelationAdapt(p->port, 4);
		}
	    p->all_echotraining_sched = -1;
	}
	return 0;
}

//---------------------------------------------------------------------------------
// Save signalling into a file
//---------------------------------------------------------------------------------
void save_log(short port, char *msg)
{
    FILE *f;
    struct dgv_pvt *p=NULL;

    char file_name[TAMSTR];

    p = find_ch_port(port);
    if (p) {
        ast_mutex_lock(&p->lock); 
        {
            sprintf(file_name,"/var/log/voicerlib/dgvchannel-%d",port);
            f=fopen(file_name,"a+");
            if (f!=NULL) {
                fprintf(f,"%s\n",msg);
                fclose(f);
            }
        } 
        ast_mutex_unlock(&p->lock);
    }
}

//--------- 
/*static int e1_category_timeout(void *data)
{
	struct dgv_pvt *p = data;

	if (p) {

		ast_log(LOG_WARNING,"********** Porta %d category timeout - id %x!!!!\n",p->port,p->e1_timeout_sched);
		dg_R2SendGroupB(p->port, B_FREE_CALLING);
		ast_log(LOG_NOTICE, "e1_category_timeout: B_FREE_CALLING\n");

	}
	
	return 0;
}*/

//-------------------------------------------------------------------------------------------
// Console Functions
//-------------------------------------------------------------------------------------------

//enables/disables audio debugging messages
static int dgv_debug_audio(int fd, int argc, char **argv)
{
	short port;
	short enable=0;
	struct dgv_pvt *p;

		

	if (argc != 5)
		return RESULT_SHOWUSAGE;

	//gets port number
	port = atoi(argv[3]);

	if (port<0 || port>=dg_GetPortsCount())
    {
        ast_cli(fd, "ERROR for port %d \n", port);
		return RESULT_SHOWUSAGE;
    }

	if (!strncmp(argv[4], "on", 2))
		enable = 1;
	else
		enable = 0;

	p = find_ch_port(port);
	if (p) 
	{
		p->audio_debug = enable;
		ast_cli(fd, "\n       %s messages for audio debugging on port %d \n\n", ((enable) ? "Enabling":"Disabling"), port);
	}
	else
		return RESULT_SHOWUSAGE;

	return RESULT_SUCCESS;
}

//TAPS parameter must be: 
/*	ECHO OFF = 0,
	ECHO_TAPS_32 = 1,
	ECHO_TAPS_64 = 2,
	ECHO_TAPS_128 = 3,
	ECHO_TAPS_256 = 4,
	ECHO_TAPS_512 = 5
*/
static int dgv_echo_can(int fd, int argc, char **argv)
{
    short port;
    short taps;
    short ret;


    if (argc != 4)
        return RESULT_SHOWUSAGE;

    port = atoi(argv[2]);
    taps = atoi(argv[3]);


    if (taps==0) {
        ast_cli(fd, "Turning off echo cancelation for port %d\n", port);
        dg_DisableEchoCancelation(port);
    }
    else
    {
        if (taps<=5) 
        {
            ret = dg_EnableEchoCancelation(port,taps,800);
            if (ret!=DG_EXIT_SUCCESS) {
                ast_cli(fd, "Error starting echo cancelation for port %d - code %x\n", port,ret);
            }
            else
                ast_cli(fd, "Turning on DSP Echo Cancelation for port %d - taps: %d - (%s)\n", port, taps, taps_desc[taps]);
        }
        else
        {
        	switch (taps)
        	{
        		case 6:
					dg_SetEchoCancelationAdapt(port, 0);
					break;
            	case 7:                 
                    dg_SetEchoCancelationAdapt(port, 1);
                    break;
                case 8:
                    dg_EnableEchoCancelation(port,3,888);
                    break;
                case 9:
                    dg_SetEchoCancelationAdapt(port, 3);
                    break;
				case 10:
                    dg_SetEchoCancelationAdapt(port, 4);
                    break;                    
        	}
        }
    }

    return RESULT_SUCCESS;
}


//dgv log r2 <on/off>
static int ast_r2_log(int fd, int argc, char **argv)
{
    if (argc != 4)
        return RESULT_SHOWUSAGE;

    if (!strncmp(argv[3],"on",2))
    {
        dg_EnableDebug(1);
        ast_cli(fd, "All R2/MFC signalling will be saved in /var/log/voicerlib/debug-<ch>\n");
    }
    else
    {
        dg_DisableDebug();
        ast_cli(fd, " R2/MFC signalling log disabled.\n");
    }

    return RESULT_SUCCESS;
}

//dgv verbose <port> <level>
static int ast_verbose_channel(int fd, int argc, char **argv)
{
	short port;
	struct dgv_pvt *p;
		
	if (argc != 4)
		return RESULT_SHOWUSAGE;

	//gets port number
    if (!strncmp(argv[2], "all", 3))
    {
        //all ports
        for (port = 1; port <= dg_GetPortsCount(); port++ ) {
            p = find_ch_port(port);
            if (p) {
                p->verbosity = atoi(argv[3]);

            }
            else
                return RESULT_SHOWUSAGE;
        }

		switch (dgv_conf.consolelanguage)
		{
			case 0://English		
        		ast_cli(fd, "\n       Set verbosity to %d on ALL ports. \n\n", p->verbosity);
				break;
			case 1://Portuguese
				ast_cli(fd, "\n       Verbosidade setada para %d em TODAS as portas. \n\n", p->verbosity);
				break;
			default:
				ast_cli(fd, "\n       Set verbosity to %d on ALL ports. \n\n", p->verbosity);
				break;		
		}
	}
    else
    {
        port = atoi(argv[2]);

    	if (port < 0 || port >= dg_GetPortsCount())
        {            
			switch (dgv_conf.consolelanguage)
			{
				case 0://English		
        			ast_cli(fd, "       ERROR for port %d. \n", port);
					break;
				case 1://Portuguese
					ast_cli(fd, "       ERRO para porta %d. \n", port);
					break;
				default:
					ast_cli(fd, "       ERROR for port %d. \n", port);
					break;		
			}            
            
    		return RESULT_SHOWUSAGE;
        }
        p = find_ch_port(port);
        if (p) {
            p->verbosity = atoi(argv[3]);
            
			switch (dgv_conf.consolelanguage)
			{
				case 0://English		
        			ast_cli(fd, "\n       Set verbosity to %d on port %d. \n\n", p->verbosity, port);
					break;
				case 1://Portuguese
					ast_cli(fd, "\n       Verbosidade setada para %d na porta %d. \n\n", p->verbosity, port);
					break;
				default:
					ast_cli(fd, "\n       Set verbosity to %d on port %d. \n\n", p->verbosity, port);
					break;		
			}
        }
        else
            return RESULT_SHOWUSAGE;
    }

	return RESULT_SUCCESS;
}



static int dgv_show_channels(int fd, int argc, char **argv)
{
#define FORMAT "%7s %7s %-10.10s %-15.15s %-16.16s %-12.12s %-9.9s\n"
	int group=0;
	struct dgv_pvt *tmp = NULL;
	char tmps[20] = "";
	char grp_tmps[20] = "";
	ast_mutex_t *lock;
	struct dgv_pvt *start;
	int lin=20;
	int show=0;

	if (argc == 4)
	{
		//group selected
		group = atoi(argv[3]);
	}
	ast_cli(fd, "List of available DigiVoice Channels");
	if (group!=0)
		ast_cli(fd, " of group %d\n",group);
	else
		ast_cli(fd, "\n");

	lock = &iflock;
	start = iffirst;
	
	tmp = start;
	while (tmp) {
		if (lin>=20)
		{
			ast_cli(fd, "\n");
			ast_cli(fd, FORMAT, "Chan", "Group", "Extension", "Context", "PortId", "CardType", "Intrf.");
			ast_cli(fd, "\n");
			lin=0;
		}
		
		if (tmp->port > 0) {
			snprintf(tmps, sizeof(tmps), "%d", tmp->port);
		} else
			ast_copy_string(tmps, "pseudo", sizeof(tmps));

		snprintf(grp_tmps, sizeof(grp_tmps), "%d", tmp->group);


		if (group==0)
			show=1;
		else
			if (group == tmp->group)
				show=1;
			else
				show=0;

		if (show) 
		{
			switch (dg_GetCardInterface(tmp->board))
			{ 
				case DG_DIGITAL_INTERFACE:
					ast_cli(fd, FORMAT, tmps, grp_tmps, tmp->exten, tmp->context, 
				            tmp->szPortID, GetCardName(tmp->board), "R2/ISDN");
				 	break;				            					
				case DG_FXO_INTERFACE:
					ast_cli(fd, FORMAT, tmps, grp_tmps, tmp->exten, tmp->context, 
				            tmp->szPortID, GetCardName(tmp->board), "FXO");
				    break;
				case DG_FXS_INTERFACE:
					ast_cli(fd, FORMAT, tmps, grp_tmps, tmp->exten, tmp->context, 
				            tmp->szPortID, GetCardName(tmp->board), "FXS");
				    break;        		
			}
			lin++;	
		}

		tmp = tmp->next;
	}
	ast_mutex_unlock(lock);
	return RESULT_SUCCESS;
}

//---- reset all channels -  it's a kind of reload with recreation of all control threads
static int dgv_reset(int fd, int argc, char **argv)
{
	short ch,ret;
	struct dgv_pvt *p;

	if (argc != 2)
		return RESULT_SHOWUSAGE;

	for (ch=1; ch<=dg_GetPortsCount();ch++)
	{
		p = find_ch_port(ch);
		if (p!=NULL)
		{
			if (p->ast_ch!=NULL)
				ast_softhangup(p->ast_ch, AST_SOFTHANGUP_APPUNLOAD);

            p->reserved=0;
		}
		//if ((dg_GetPortInterface(ch) == DG_DIGITAL_INTERFACE) && p->sig!=DG_SIG_PRI)
        switch(p->sig)
		{

            case DG_SIG_R2_NDIS_FIXED:
            case DG_SIG_R2_NDIS_VARIABLE:
    			ret=dg_DestroyE1Thread(ch);
    			if (p->verbosity==90)
    				ast_log(LOG_NOTICE, "Destroying E1 thread for port %d - cod %x\n",ch,ret);		
			
                break;
			case DG_SIG_CUSTOMCAS:
    			ret=dg_DestroyCustomCAS(ch);
    			if (p->verbosity==90)
    				ast_log(LOG_NOTICE, "Destroying Custom CAS thread for port %d - cod %x\n",ch,ret);
    				
				break;                
		}
		//calls all destructor functions
		dg_DisableInputBuffer(ch);
		dg_StopPlayBuffer(ch);
		dg_DestroyCallProgress(ch);
	}
	setup_dgv(1);	//
	return RESULT_SUCCESS;
}

//---- Shows channel version
static int dgv_show_version(int fd, int argc, char **argv)
{
    int p;

    ast_cli(fd, "\nChannel Driver Version: %s \n", CHANDGV_VERSION);
    ast_cli(fd, "Firmware Versions: \n");
    for (p=1;p<=dg_GetCardsCount();p++) 
    {
        ast_cli(fd, "     |_ Card %d - Version %xh \n",p, dg_GetVersion(p));
    }
    
    ast_cli(fd, "\n");    
    
    system("ls -gGo /usr/lib/libdigivoice.so");

    return RESULT_SUCCESS;
}

//---- Shows specific channel information
static int dgv_show_channel(int fd, int argc, char **argv)
{
	struct dgv_pvt *tmp;
	int channel;

    if (argc != 4)
        return RESULT_SHOWUSAGE;

	channel = atoi(argv[3]);

	tmp = find_ch_port(channel);
	if (tmp!=NULL)
	{
		ast_cli(fd, "\n");
		ast_cli(fd, " Channel: %d\n", tmp->port);
		ast_cli(fd, " Group: %d\n", tmp->group);
		ast_cli(fd, " Card number: %d\n", tmp->board);
		ast_cli(fd, " Extension: %s\n", tmp->exten);
		ast_cli(fd, " Context: %s\n", tmp->context);
		ast_cli(fd, " Owner: %s\n", (tmp->ast_ch)  ?  tmp->ast_ch->name  :  "<None>");
		ast_cli(fd, " Account Code: %s\n", (strlen(tmp->accountcode)>0)  ?  tmp->accountcode :  "<None>");
		ast_cli(fd, " Card model: %s\n", GetCardName(tmp->board));
		ast_cli(fd, " RX gain: %d\n", tmp->rx_gain);
		ast_cli(fd, " TX gain: %d\n", tmp->tx_gain);
		ast_cli(fd, " Reserved: %x\n", tmp->reserved);
        ast_cli(fd, " Silence Threshold: %d\n", tmp->silence_threshold);
        ast_cli(fd, " Signalling: %s\n", sig2str(tmp->sig));
        ast_cli(fd, " Pickup Group: %d\n", (int)tmp->pickupgroup);
        ast_cli(fd, " DialTone Timeout: %d\n", (int)tmp->dialtone_timeout);
        ast_cli(fd, " MusicOnHold Class: %s\n", (ast_strlen_zero(tmp->musicclass)) ? "<default>\0" : tmp->musicclass);
		ast_cli(fd, " Using DSP echo cancelation: %s\n", tmp->echo_can ? "yes":"no");
        if (tmp->echo_can>0 && tmp->echo_can<6)
             ast_cli(fd, " Using echo cancelation taps = %s\n", taps_desc[tmp->echo_can]);
        ast_cli(fd, " Using echo cancelation training after %d ms\n", tmp->echo_training);
        ast_cli(fd, " Drop tx samples factor: %d\n", (int)tmp->drop_tx_samples_factor);
        ast_cli(fd, " Relax DTMF level: %d\n", (int)tmp->relaxdtmf);
        ast_cli(fd, " Generate ringback tone: %s\n", tmp->gen_ringback ? "yes":"no");
        ast_cli(fd, " Generate busy tone: %s\n", tmp->gen_busy ? "yes":"no");
        ast_cli(fd, " Regenerate detected DTMF: %s\n", tmp->gen_dtmf  ? "yes":"no");
        ast_cli(fd, " Detects DTMF and pass it to Asterisk: %s\n", tmp->dtmf_to_asterisk ? "yes":"no");
        ast_cli(fd, " Callprogress configuration file: %s\n", tmp->callprogress);
        ast_cli(fd, " Language: %s\n", tmp->language);
        ast_cli(fd, " Current Rx codec: %s\n", ast2dgvformatname(tmp->read_last_format));
		ast_cli(fd, " Current Tx codec: %s\n", ast2dgvformatname(tmp->write_last_format));
		ast_cli(fd, " Fax detection: %s\n", tmp->fax_detection ? "yes":"no");
		switch (dg_GetCardInterface(tmp->board))
		{ 
			case DG_DIGITAL_INTERFACE:
				ast_cli(fd, " Interface Type: R2/ISDN \n");
				break;				            					
			case DG_FXO_INTERFACE:
				ast_cli(fd, " Interface Type: FXO \n");
				break;				
			case DG_FXS_INTERFACE:
				ast_cli(fd, " Interface Type: FXS \n");
				break;		
		}		

        switch(tmp->sig)
        {

            case DG_SIG_R2_NDIS_FIXED:
            case DG_SIG_R2_NDIS_VARIABLE:
                ast_cli(fd, "   |_ PortId: %s\n", tmp->szPortID);
                ast_cli(fd, "   |_ Category: %d\n", tmp->category);
                //ast_cli(fd, "   |_ R2D/MFC mode: %s\n", (tmp->auto_r2dmfc) ? "Auto":"Variable NDIS");
                if (tmp->sig==DG_SIG_R2_NDIS_VARIABLE)
                {
                    ast_cli(fd, "   |_ Min rx digits: %d\n", tmp->min_digits_rx);
                    ast_cli(fd, "   |_ Interdigit timeout: %d\n", tmp->interdigit_timeout);

                }
                ast_cli(fd, "   |_ Max rx digits for incoming calls: %d\n", tmp->max_digits_rx);
                ast_cli(fd, "   |_ Ask for CID position: %d\n", tmp->ask_for_cid_position);
                ast_cli(fd, "   |_ B Subscriber hangup: %s\n", tmp->b_subscriber_hangup ? "yes":"no");
                break;
#ifdef USE_LIBPRI                
            case DG_SIG_PRI:
            	//ISDN information
                ast_cli(fd, "   |_ Outgoing call: %s\n", tmp->outgoing ? "yes":"no");          	
                break;
#endif//#ifdef USE_LIBPRI                
            case DG_SIG_CUSTOMCAS:
                //CUSTOMCAS information
                ast_cli(fd, "   |_ Default Custom CAS configuration file: %s\n", tmp->customcascfg);
                ast_cli(fd, "   |_ Default Call Progress configuration file: %s\n", dgv_conf.szDefaultCallProgressFile);                	
                ast_cli(fd, "   |_ Wait for dialtone: %s\n", tmp->wait_dialtone ? "yes":"no");
                break;
            case DG_SIG_FXO:
                //FXO information
                ast_cli(fd, "   |_ Default Call Progress configuration file: %s\n", dgv_conf.szDefaultCallProgressFile);	
                ast_cli(fd, "   |_ Wait for dialtone: %s\n", tmp->wait_dialtone ? "yes":"no");
                break;
            case DG_SIG_FXS:
            	//FXS information
                ast_cli(fd, "   |_ Default Call Progress configuration file: %s\n", dgv_conf.szDefaultCallProgressFile);
                ast_cli(fd, "   |_ Flash Min Time: %d \n", (int)tmp->flash_min_time);
                ast_cli(fd, "   |_ Flash Max Time: %d \n", (int)tmp->flash_max_time);
                ast_cli(fd, "   |_ Immediate: %s \n", tmp->immediate ? "yes":"no");
                ast_cli(fd, "   |_ MailBox: %s \n", tmp->mailbox);
                ast_cli(fd, "   |_ PortId: %s\n", tmp->szPortID);                  
                ast_cli(fd, "   |_ Ring Type: %d \n", (int)tmp->ring_type);                              
                break;
        }
	}
	return RESULT_SUCCESS;
}

//only show the alarm status
static int dgv_show_alarms(int fd, int argc, char **argv)
{
	int c;
	short card=0;	
	
    if (argc != 4)
    {
        ast_cli(fd, "\n");
        
		switch (dgv_conf.consolelanguage)
		{
			case 0://English		
        		ast_cli(fd, "       Error: You passed only %d args. \n", argc);
				break;
			case 1://Portuguese
				ast_cli(fd, "       Erro: Voce passou somente %d parametros. \n", argc);
				break;
			default:
				ast_cli(fd, "       Error: You passed only %d args. \n", argc);
				break;		
		}         	    	
        return RESULT_SHOWUSAGE;
    }
	
	card = atoi(argv[3]);
	
	if (card == 99)
	{
		ast_cli(fd, "\n       Requesting alarm notifications... \n\n");

		for (c = 1; c <= dg_GetCardsCount(); c++)
			dg_GetAlarmStatus(c);	
			
		return RESULT_SUCCESS; 
	}
	
	switch (dg_GetCardType(card))
	{			    		
		case VB0408PCI:
		case VB0404FX:
			ast_cli(fd, "\n");
			
			switch (dgv_conf.consolelanguage)
			{
				case 0://English		
        			ast_cli(fd, "       Error: This card isn't a E1 Card. \n");
					break;
				case 1://Portuguese
					ast_cli(fd, "       Erro: Esta placa nao eh uma Placa E1. \n");
					break;
				default:
					ast_cli(fd, "       Error: This card isn't a E1 Card. \n");
					break;		
			}       	    	
			return RESULT_SHOWUSAGE;
						
	    case VBE13030PCI:
		case VBE13060PCI:			    		
	    case VBE13060PCI_R:	    
		case VBE16060PCI:
		case VBE16060PCI_R:
			break;
		default:
			ast_cli(fd, "\n");
			
			switch (dgv_conf.consolelanguage)
			{
				case 0://English		
        			ast_cli(fd, "       Error: This card doesn't exist... \n");
					break;
				case 1://Portuguese
					ast_cli(fd, "       Erro: Esta placa nao existe... \n");
					break;
				default:
					ast_cli(fd, "       Error: This card doesn't exist... \n");
					break;		
			}       	    	
			return RESULT_SHOWUSAGE;			
	}
				
	switch (dgv_conf.consolelanguage)
	{
		case 0://English		
			ast_cli(fd, "\n       DigiVoice - Alarms from E1 Card: %d \n\n", card);
			break;
		case 1://Portuguese
			ast_cli(fd, "\n       DigiVoice - Alarmes da Placa E1: %d \n\n", card);			
			break;
		default:
			ast_cli(fd, "\n       DigiVoice - Alarms from E1 Card: %d \n\n", card);		
			break;		
	}		
		
	if (dg_GetCardPortsCount(card) == 60)
	{		
		ast_cli(fd, "       +---[ Span: 1 ]----------+---[ Span: 2 ]----------+ \n");
		ast_cli(fd, "       |                        |                        | \n");
		ast_cli(fd, "       |  ALARM_LOSS       %s  |  ALARM_LOSS       %s  | \n", showalarmstatus_desc[dgv_e1alarms[card - 1][0].nALARM_LOSS]    , showalarmstatus_desc[dgv_e1alarms[card - 1][1].nALARM_LOSS]    );
		ast_cli(fd, "       |  ALARM_AISS       %s  |  ALARM_AISS       %s  | \n", showalarmstatus_desc[dgv_e1alarms[card - 1][0].nALARM_AISS]    , showalarmstatus_desc[dgv_e1alarms[card - 1][1].nALARM_AISS]    );
		ast_cli(fd, "       |  ALARM_AIS16S     %s  |  ALARM_AIS16S     %s  | \n", showalarmstatus_desc[dgv_e1alarms[card - 1][0].nALARM_AIS16S]  , showalarmstatus_desc[dgv_e1alarms[card - 1][1].nALARM_AIS16S]  );
		ast_cli(fd, "       |  ALARM_SYNC       %s  |  ALARM_SYNC       %s  | \n", showalarmstatus_desc[dgv_e1alarms[card - 1][0].nALARM_SYNC]    , showalarmstatus_desc[dgv_e1alarms[card - 1][1].nALARM_SYNC]    );
		ast_cli(fd, "       |  ALARM_MFSYNC     %s  |  ALARM_MFSYNC     %s  | \n", showalarmstatus_desc[dgv_e1alarms[card - 1][0].nALARM_MFSYNC]  , showalarmstatus_desc[dgv_e1alarms[card - 1][1].nALARM_MFSYNC]  );
		ast_cli(fd, "       |  ALARM_RAIS       %s  |  ALARM_RAIS       %s  | \n", showalarmstatus_desc[dgv_e1alarms[card - 1][0].nALARM_RAIS]    , showalarmstatus_desc[dgv_e1alarms[card - 1][1].nALARM_RAIS]    );
		ast_cli(fd, "       |  ALARM_CRC4SYNC   %s  |  ALARM_CRC4SYNC   %s  | \n", showalarmstatus_desc[dgv_e1alarms[card - 1][0].nALARM_CRC4SYNC], showalarmstatus_desc[dgv_e1alarms[card - 1][1].nALARM_CRC4SYNC]);
		ast_cli(fd, "       |  ALARM_RSLIP      %03d  |  ALARM_RSLIP      %03d  | \n",                  dgv_e1alarms[card - 1][0].nALARM_RSLIP    ,                      dgv_e1alarms[card - 1][1].nALARM_RSLIP    );
		ast_cli(fd, "       |                        |                        | \n");
		ast_cli(fd, "       +------------------------+------------------------+ \n\n");
	}
	else
	{
		ast_cli(fd, "       +---[ Span: 1 ]----------+ \n");
		ast_cli(fd, "       |                        | \n");
		ast_cli(fd, "       |  ALARM_LOSS       %s  | \n", showalarmstatus_desc[dgv_e1alarms[card - 1][0].nALARM_LOSS]    );
		ast_cli(fd, "       |  ALARM_AISS       %s  | \n", showalarmstatus_desc[dgv_e1alarms[card - 1][0].nALARM_AISS]    );
		ast_cli(fd, "       |  ALARM_AIS16S     %s  | \n", showalarmstatus_desc[dgv_e1alarms[card - 1][0].nALARM_AIS16S]  );
		ast_cli(fd, "       |  ALARM_SYNC       %s  | \n", showalarmstatus_desc[dgv_e1alarms[card - 1][0].nALARM_SYNC]    );
		ast_cli(fd, "       |  ALARM_MFSYNC     %s  | \n", showalarmstatus_desc[dgv_e1alarms[card - 1][0].nALARM_MFSYNC]  );
		ast_cli(fd, "       |  ALARM_RAIS       %s  | \n", showalarmstatus_desc[dgv_e1alarms[card - 1][0].nALARM_RAIS]    );
		ast_cli(fd, "       |  ALARM_CRC4SYNC   %s  | \n", showalarmstatus_desc[dgv_e1alarms[card - 1][0].nALARM_CRC4SYNC]);
		ast_cli(fd, "       |  ALARM_RSLIP      %03d  | \n",                    dgv_e1alarms[card - 1][0].nALARM_RSLIP      );
		ast_cli(fd, "       |                        | \n");
		ast_cli(fd, "       +------------------------+ \n\n");
	}

	switch (dgv_conf.consolelanguage)
	{
		case 0://English		
			ast_cli(fd, "       Alarms Descriptions: \n");
			ast_cli(fd, "        | \n");
			ast_cli(fd, "        |_ ALARM_LOSS     - Loss of Signal Status Indication. \n");			
			ast_cli(fd, "        |_ ALARM_AISS     - Alarm Indication Status Signal. \n");
			ast_cli(fd, "        |_ ALARM_AIS16S   - Alarm Indication Signal 16 Status. \n");
			ast_cli(fd, "        |_ ALARM_SYNC     - Basic Frame Alignment. \n");
			ast_cli(fd, "        |_ ALARM_MFSYNC   - Multiframe Alignment. \n");
			ast_cli(fd, "        |_ ALARM_RAIS     - Remote Alarm Indication Status. \n");						
			ast_cli(fd, "        |_ ALARM_CRC4SYNC - CRC-4 Synchronization. \n");						
			ast_cli(fd, "        |_ ALARM_RSLIP    - Frame Slip. \n\n");			
			break;
		case 1://Portuguese
			ast_cli(fd, "       Descricoes dos Alarmes: \n");
			ast_cli(fd, "        | \n");			
			ast_cli(fd, "        |_ ALARM_LOSS     - Indicacao de Perda de Sinal. \n");			
			ast_cli(fd, "        |_ ALARM_AISS     - Status de Indicacao de Alarme. \n");
			ast_cli(fd, "        |_ ALARM_AIS16S   - Indicacao de Alarme no Canal 16. \n");
			ast_cli(fd, "        |_ ALARM_SYNC     - Perda de Sincronismo de Quadro. \n");
			ast_cli(fd, "        |_ ALARM_MFSYNC   - Perda de Sincronismo de Multiquadro. \n");
			ast_cli(fd, "        |_ ALARM_RAIS     - Indicacao de Alarme Remoto. \n");						
			ast_cli(fd, "        |_ ALARM_CRC4SYNC - Erro de CRC-4. \n");						
			ast_cli(fd, "        |_ ALARM_RSLIP    - Escorregamento. \n\n");		
			break;
		default:
			ast_cli(fd, "       Alarms Descriptions: \n");
			ast_cli(fd, "        | \n");
			ast_cli(fd, "        |_ ALARM_LOSS     - Loss of Signal Status Indication. \n");			
			ast_cli(fd, "        |_ ALARM_AISS     - Alarm Indication Status Signal. \n");
			ast_cli(fd, "        |_ ALARM_AIS16S   - Alarm Indication Signal 16 Status. \n");
			ast_cli(fd, "        |_ ALARM_SYNC     - Basic Frame Alignment. \n");
			ast_cli(fd, "        |_ ALARM_MFSYNC   - Multiframe Alignment. \n");
			ast_cli(fd, "        |_ ALARM_RAIS     - Remote Alarm Indication Status. \n");						
			ast_cli(fd, "        |_ ALARM_CRC4SYNC - CRC-4 Synchronization. \n");						
			ast_cli(fd, "        |_ ALARM_RSLIP    - Frame Slip. \n\n");			
			break;		
	}
	
	return RESULT_SUCCESS;
}

/* debug using */
static int dgv_drop(int fd, int argc, char **argv)
{
    short port;
    short times=0;
    short txrx=0;  //0-tx 1 - rx
    struct dgv_pvt *tmp;


    if (argc != 5)
    {
        ast_cli(fd, "\n");    	    	
        ast_cli(fd, "Error: You passed only %d args \n", argc);    
        return RESULT_SHOWUSAGE;
    }


    if (!strncmp(argv[2],"tx",2))
        txrx=0;
    else
        txrx=1;

    port = atoi(argv[3]);
    times = atoi(argv[4]);

    tmp = find_ch_port(port);
    if (tmp!=NULL)
    {
        ast_cli(fd, "Forcing dropping %s samples %d times\n", (txrx ? "rx":"tx"), times);
        if (txrx)
            tmp->force_drop_rx = times;
        else
            tmp->force_drop_tx = times;
    }

    return RESULT_SUCCESS;
}

static int dgv_record(int fd, int argc, char **argv)
{
	short record_ch=0;
	
    if (argc != 3)
    {
        ast_cli(fd, "\n");
        
		switch (dgv_conf.consolelanguage)
		{
			case 0://English		
        		ast_cli(fd, "       Error: You passed only %d args. \n", argc);
				break;
			case 1://Portuguese
				ast_cli(fd, "       Erro: Voce passou somente %d parametros. \n", argc);
				break;
			default:
				ast_cli(fd, "       Error: You passed only %d args. \n", argc);
				break;		
		}         	    	
        return RESULT_SHOWUSAGE;
    }	
	
	record_ch = atoi(argv[2]);
	
	/*
	if (dg_IsRecording(record_ch))
		dg_StopRecordFile(record_ch);
	else	
		dg_RecordFile(record_ch, "/var/spool/voicerlib/teste.gsm", "");
    */
    
	return RESULT_SUCCESS;
}

//-------------------------------------------------------------------------------------------
// Console Commands Help
//-------------------------------------------------------------------------------------------
static char show_channels_usage_en[] =
	"\n"
	"       Shows a list of available dgv channels. \n       If group is specified, show channels of this group. \n"
	"        |_ Usage: dgv show channels [group] \n"	
	"\n";

static char show_channels_usage_br[] =
	"\n"
	"       Mostra a lista de canais disponiveis. \n       Caso especifique o grupo, serao mostrados os canais deste grupo. \n"
	"        |_ Uso: dgv show channels [group] \n"	
	"\n";

static char show_channel_usage_en[] =
	"\n"
	"       Detailed information about a given channel. \n"
	"        |_ Usage: dgv show channel <chan num> \n"	
	"\n";
	
static char show_channel_usage_br[] =
	"\n"
	"       Informacoes detalhadas de um determinado canal. \n"	
	"        |_ Uso: dgv show channel <chan num> \n"	
	"\n";
	
static char reset_usage_en[] =
	"\n"
	"       Resets all DigiVoice channels, hangup and restart all controls. \n"
	"        |_ Usage: dgv reset \n"	
	"\n";
	
static char reset_usage_br[] =
	"\n"
	"       Reseta todos os canais DigiVoice, desliga e restarta todos os controles. \n"
	"        |_ Uso: dgv reset \n"	
	"\n";
	
static char show_version_usage_en[] =
	"\n"
	"       Shows chan_dgv current version. \n"
	"        |_ Usage: dgv show version \n"	
	"\n";
	
static char show_version_usage_br[] =
	"\n"
	"       Mostra a versao corrente do chan_dgv. \n"
	"        |_ Uso: dgv show version \n"	
	"\n";
	
static char show_alarms_usage_en[] =
	"\n"
	"       Shows alams from specific E1 Card installed. \n"
	"        |_ Usage: dgv show alarms <card> \n"	
	"\n";
	
static char show_alarms_usage_br[] =
	"\n"	
	"       Mostra alarmes de uma especifica Placa E1 instalada. \n"
	"        |_ Uso: dgv show alarms <cards> \n"	
	"\n";

//for testing purposes
static char debug_audio_usage_en[] =
	"\n"
	"       Enables or disables messages (for testing purposes). \n"
	"        |_ Usage: dgv debug audio <ch> <on/off> \n"	
	"\n";
	
static char debug_audio_usage_br[] =
	"\n"
	"       Habilita/Desabilita debug de audio (para fins de teste). \n"	
	"        |_ Uso: dgv debug audio <ch> <on/off> \n"	
	"\n";
	
static char verbose_usage_en[] =
	"\n"
	"       Sets verbosity level to specific channel. <ch>=1..n or 'all' , <level>=0..n (for testing purposes). \n"
	"        |_ Usage: dgv verbose <ch> <level> \n\n"
	"       Verbosity level Descriptions: \n"
	"        | \n"
	"        |_ Level / Description \n"
	"        |_    0  - Disable dgv verbose. \n"	
	"        |_   10  - start_dial + dgv_new + dgv_request + dgv_call + dgv_hangup + dgv_fixup. \n"
	"        |_   20  - set_port_group + find_free_port. \n"
	"        |_   30  - ISDN events. \n"
	"        |_   40  - ToneThread + StutterToneThread. \n"
	"        |_   50  - DigiVoice Hardware events. \n"
	"        |_   60  - read_samples. \n"
	"        |_   70  - QueueFrame. \n"
	"        |_   80  - dgv_write. \n"
	"        |_   90  - dgv_reset + unload_module. \n"
	"        |_  100  - dgv_indicate. \n"
	"\n";
	
static char verbose_usage_br[] =
	"\n"
	"       Seta um nivel de log para um canal especifico. <ch>=1..n or 'all', <level>=0..n (para fins de teste). \n"	
	"        |_ Uso: dgv verbose <ch> <level> \n\n"
	"       Descricao dos niveis de log: \n"
	"        | \n"
	"        |_ Nivel / Descricao \n"
	"        |_    0  - Desabilita dgv verbose. \n"	
	"        |_   10  - start_dial + dgv_new + dgv_request + dgv_call + dgv_hangup + dgv_fixup. \n"
	"        |_   20  - set_port_group + find_free_port. \n"
	"        |_   30  - Eventos ISDN. \n"
	"        |_   40  - ToneThread + StutterToneThread. \n"
	"        |_   50  - Eventos Hardware DigiVoice. \n"
	"        |_   60  - read_samples. \n"
	"        |_   70  - QueueFrame. \n"
	"        |_   80  - dgv_write. \n"
	"        |_   90  - dgv_reset + unload_module. \n"
	"        |_  100  - dgv_indicate. \n"
	"\n";
	
static char r2_log_usage_en[] =
	"\n"
	"       Turns on/off r2d/mfc logs in /var/log/voicerlib/debug-<ch> (for testing purposes). \n"
	"        |_ Usage: dgv r2 log <on/off> \n"	
	"\n";
	
static char r2_log_usage_br[] =
	"\n"
	"       Habilita/Desabilita logs de r2d/mfc em /var/log/voicerlib/debug-<ch> (para fins de teste). \n"	
	"        |_ Uso: dgv r2 log <on/off> \n"	
	"\n";
	
static char echo_can_usage_en[] =
	"\n"
	"       Set echo cancelation manually with TAPS (0 - off, 1-32, 2-64, 3-128, 4-256, 5-512) (for testing purposes). \n"
	"        |_ Usage: dgv echo <ch> <taps> \n"	
	"\n";
	
static char echo_can_usage_br[] =
	"\n"
	"       Seta cancelamento de eco manualmente com TAPS (0 - off, 1-32, 2-64, 3-128, 4-256, 5-512) (para fins de teste). \n"
	"        |_ Uso: dgv echo <ch> <taps> \n"	
	"\n";
	
static char drop_txrx_usage_en[] =
	"\n"
	"       Force dropping TX/RX samples. <times> times on <ch> channel (for testing purposes). \n"
	"        |_ Usage: dgv drop tx/rx <ch> <times> \n"	
	"\n";
	
static char drop_txrx_usage_br[] =
	"\n"
	"       Remove amostras de TX/RX. <times> vezes em um <ch> canal (para fins de teste). \n"	
	"        |_ Uso: dgv drop tx/rx <ch> <times> \n"	
	"\n";

static char record_file_en[] =
	"\n"
	"       Please, don't use this function. \n"
	"        |_ Usage: dgv record <ch> \n"	
	"\n";
	
static char record_file_br[] =
	"\n"	
	"       Nao utilize esta funcao. \n"
	"        |_ Uso: dgv record <ch> \n"	
	"\n";
	
//-------------------------------------------------------------------------------------------
//Console Definitions -> English 
//-------------------------------------------------------------------------------------------
static struct ast_cli_entry dgv_cli_en[] = {
	{ {"dgv", "show", "channels", NULL}, dgv_show_channels,
	  "Show active DigiVoice channels", show_channels_usage_en },
	  
	{ {"dgv", "show", "channel", NULL}, dgv_show_channel,
	  "Show detailed information about a given channel", show_channel_usage_en },
	  
	{ {"dgv", "reset", NULL, NULL}, dgv_reset,
	  "Reset all DigiVoice channels", reset_usage_en },
	  
	{ {"dgv", "show", "version", NULL}, dgv_show_version,
	  "Show chan_dgv current version", show_version_usage_en },
	  
	{ {"dgv", "show", "alarms", NULL}, dgv_show_alarms,
	  "Shows alams from specific E1 card installed", show_alarms_usage_en },
  
	//for testing purposes  
	{ {"dgv", "debug", "audio", NULL}, dgv_debug_audio,
	  "Enable/Disable audio debug (for testing purposes)", debug_audio_usage_en },
	  
    { {"dgv", "verbose", NULL, NULL}, ast_verbose_channel,
      "Sets verbosity level to specific channel (for testing purposes)", verbose_usage_en },
      
    { {"dgv", "r2", "log", NULL}, ast_r2_log,
      "Logs all R2D/MFC signalling into a file (for testing purposes)", r2_log_usage_en },
      
    { {"dgv", "echo", NULL, NULL}, dgv_echo_can,
      "Turns echo cancelation on/off (for testing purposes)", echo_can_usage_en },
      
    { {"dgv", "drop", NULL, NULL, NULL}, dgv_drop,
      "Dropping samples utility (for testing purposes)", drop_txrx_usage_en },
      
    { {"dgv", "record", NULL}, dgv_record,
      "Please, don't use this function (for testing purposes)", record_file_en },
};

//-------------------------------------------------------------------------------------------
//Console Definitions -> Portuguese 
//-------------------------------------------------------------------------------------------
static struct ast_cli_entry dgv_cli_br[] = {
	{ {"dgv", "show", "channels", NULL}, dgv_show_channels,
	  "Mostra os canais DigiVoice ativos", show_channels_usage_br },
	  
	{ {"dgv", "show", "channel", NULL}, dgv_show_channel,
	  "Mostra informacoes detalhadas de um determinado canal", show_channel_usage_br },
	  
	{ {"dgv", "reset", NULL, NULL}, dgv_reset,
	  "Reseta todos os canais DigiVoice", reset_usage_br },
	  
	{ {"dgv", "show", "version", NULL}, dgv_show_version,
	  "Mostra a versao corrente do chan_dgv", show_version_usage_br },
	  
	{ {"dgv", "show", "alarms", NULL}, dgv_show_alarms,
	  "Mostra alarmes de uma especifica placa E1 instalada", show_alarms_usage_br },
	  
	//para fins de teste 
	{ {"dgv", "debug", "audio", NULL}, dgv_debug_audio,
	  "Habilita/Desabilita debug de audio (para fins de teste)", debug_audio_usage_br },
	  
    { {"dgv", "verbose", NULL, NULL}, ast_verbose_channel,
      "Seta um nivel de log para um canal especifico (para fins de teste)", verbose_usage_br },
      
    { {"dgv", "r2", "log", NULL}, ast_r2_log,
      "Gera um arquivo de log com toda sinalizacao R2D/MFC (para fins de teste)", r2_log_usage_br },
      
    { {"dgv", "echo", NULL, NULL}, dgv_echo_can,
      "Habilita/Desabilita cancelamento de eco para efeito de testes (para fins de teste)", echo_can_usage_br },
      
    { {"dgv", "drop", NULL, NULL, NULL}, dgv_drop,
      "Utilitario para remover amostras (para fins de teste)", drop_txrx_usage_br },
      
    { {"dgv", "record", NULL}, dgv_record,
      "Nao utilize esta funcao (para fins de teste)", record_file_br },           
};

static void *do_timer(void *data)
{
	struct timespec timeout;
	int schedms;
	int numevents = 0;
	for(;;) {
		struct timeval now = ast_tvnow();
		schedms = ast_sched_wait(sched);
		/* this shouldn't happen, but provide a 1 second default just in case */
		if (schedms <= 0)
			schedms = 1000;
		timeout.tv_sec = now.tv_sec + (schedms / 1000);
		timeout.tv_nsec = (now.tv_usec * 1000) + ((schedms % 1000) * 1000);
		/* prevent stuff from clobbering cdr_pending_cond, then wait on signals sent to it until the timeout expires */
		ast_mutex_lock(&timer_pending_lock);
		{
			ast_cond_timedwait(&timer_pending_cond, &timer_pending_lock, &timeout);
			numevents = ast_sched_runq(sched);
		}
		ast_mutex_unlock(&timer_pending_lock);
	}

	return NULL;
}

//isdn support functions
#ifdef USE_LIBPRI
	//---------------------------------------------
	// Find dchans index based on card and span
	//---------------------------------------------
	int dg_pri_findindex(short card, short span)
	{
	    int i= 0;
    
    	for (i=0; i< dchans_count; i++) 
    	{
        	if (dchans[i].card==card && dchans[i].span==span) 
        	{
            	//found!
            	return i;
        	}
    	}
    	//not found
    	return -1;
	}

	//---------------------------------------------
	// Find dchans index based on pri handle
	//---------------------------------------------
	int dg_pri_findpri(struct pri *pri)
	{
    	int i= 0;
    
	    for (i=0; i< dchans_count; i++) 
    	{
        	if (dchans[i].pri==pri) 
        	{
	            //found!
            	return i;
        	}
    	}
    	//not found
    	return -1;
	}
	
	//---------------------------------------------
	// Find dchans index based on fds[1]
	//---------------------------------------------
	int dg_pri_findfds(int fds)
	{
	    int i= 0;
	    
	    for (i=0; i< dchans_count; i++) 
	    {
        	if (dchans[i].fds[TX]==fds) 
        	{
	            //found!
    	        return i;
        	}
    	}
    	//not found
    	return -1;
	}

	//------------------------------------------------
	// dg_pri_to_port
	// converts pri channel (1-15,17-31,32-46,48-62) to voicerlib
	// defaults (1-30,31-60)
	//------------------------------------------------
	int dg_pri_to_port(int pri_channel)
	{
		if (pri_channel <= 15)
			return pri_channel;
		else
			if ((pri_channel >= 17) && (pri_channel <= 31))
				return (pri_channel - 1);
			else
				if ((pri_channel >= 32) && (pri_channel <= 46))
					return (pri_channel - 31);
				else
					if ((pri_channel >= 48) && (pri_channel <= 62))
						return (pri_channel - 32);
					else
						return 0;
	}
	
	//------------------------------------------------
	// port_to_pri
	// converts voicerlib port (1-30,31-60) to pri 
	// channel (1-15,17-31) 
	//------------------------------------------------
	int port_to_pri(int port)
	{
		if (port <= 15)
			return port;
		else
			if ((port >= 16) && (port <= 30))
				return (port + 1);
			else
			{
				port = port - 30;

				if (port <= 15)
					return port;
				else
					return (port + 1);
			}

		// channel (1-15,17-31,32-46,48-62)
		/*if (port <= 15)
			return port;
		else
			if ((port >= 16) && (port <= 45))
				return (port + 1);
			else
				if ((port >= 46) && (port <= 60))
					return (port + 2);
				else
					return 0;*/			
	}

	//--------------------------------------
	// Simple switch to isdn - overlap
	//--------------------------------------
	static void *ss_thread(void *data)
	{
	    struct dgv_pvt *p = data;
    	char exten[AST_MAX_EXTENSION]="";
    	int len = 0;
    	int res;
    	int timeout;
    	char ss_name[256];

    	pthread_t threadid;
    	pthread_attr_t attr;

    	if(p->ast_ch)
    	{   	
	        if (option_verbose >= 3) 
    	        ast_verbose(VERBOSE_PREFIX_3 "Starting simple switch on '%s'\n", p->ast_ch->name);

			strncpy(ss_name, p->ast_ch->name, 100);

    	    /* Now loop looking for an extension */
        	ast_copy_string(exten, p->exten, sizeof(exten));
        	len = strlen(exten);
        	res = 0;       
      
        	while((len < AST_MAX_EXTENSION-1) && ast_matchmore_extension(p->ast_ch, p->ast_ch->context, exten, 1, p->cid_num)) 
        	{
            	if (len && !ast_ignore_pattern(p->ast_ch->context, exten))
                	dg_GenerateMF(p->port,GENERATE_OFF,0);
            	else
            	{
                	dg_GenerateMF(p->port,GENERATE_TONE1,0);
                
					if (p->verbosity==30)
					ast_log(LOG_NOTICE, "ss_thread GENERATE_TONE1 on %d \n", p->port);                                
    	        }    

        	    if (ast_exists_extension(p->ast_ch, p->ast_ch->context, exten, 1, p->cid_num))
        		{
	                timeout = matchdigittimeout;                
    	    	}    
        	    else
            	{
                	timeout = gendigittimeout;                
            	}
            
            	/*chan_zap.c send a AST_FRAME_VOICE in zt_read, here we call a thread to send it*/
    			pthread_attr_init(&attr);
    			pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);         
    	        if (ast_pthread_create(&threadid, &attr, fv_thread, p)>=0) 
        	    {           	
    				if (option_verbose >= 3)
                		ast_log(LOG_NOTICE,"fv_thread Created - port %s \n", p->ast_ch->name);	
	            }
    	        else                                
        	    {            	
            		ast_log(LOG_ERROR, "Error creating fv_thread - port %s \n", p->ast_ch->name);            
            	}
            	pthread_attr_destroy(&attr);	

	            res = ast_waitfordigit(p->ast_ch, timeout);
	                                    
    	        if (res < 0) {
        	        ast_log(LOG_DEBUG, "ss_thread: waitfordigit returned < 0...\n");               
            	    ast_hangup(p->ast_ch);                
                	return NULL;
            	} else if (res) {
	                exten[len++] = res;
    	            exten[len] = '\0';
        	    } else
            	{            	
        	  		if (option_verbose >= 3)	         	
                		ast_verbose(VERBOSE_PREFIX_3 "ss_thread: Timeout on '%s'\n", p->ast_ch->name);
	                break;
    	        }
        	    usleep(1);
        	}//while

        	dg_GenerateMF(p->port,GENERATE_OFF,0);
        
        	/* if no extension was received ('unspecified') or invalid extension on overlap call, use the 's' extension */
        	if ((ast_strlen_zero(exten)) || (!ast_exists_extension(p->ast_ch, p->ast_ch->context, exten, 1, p->cid_num))) 
        	{
	            if (option_verbose >= 3)
    	            ast_verbose(VERBOSE_PREFIX_3 "Going to extension s|1 because of empty or invalid extension received on overlap call\n");
        	    exten[0] = 's';
            	exten[1] = '\0';
        	}   
                    
        	if (ast_exists_extension(p->ast_ch, p->ast_ch->context, exten, 1, p->cid_num)) 
        	{
            	/* Start the real PBX */           
            	ast_copy_string(p->ast_ch->exten, exten, sizeof(p->ast_ch->exten));
            	ast_setstate(p->ast_ch, AST_STATE_RING);
            	p->call_state=DG_AST_RINGING;
            	res = ast_pbx_run(p->ast_ch);
            	if (res) 
                	ast_log(LOG_WARNING, "PBX exited non-zero!\n");
        	} 
        	else 
        	{
            	ast_log(LOG_WARNING, "No such possible extension '%s' in context '%s'\n", exten, p->ast_ch->context);
            	p->ast_ch->hangupcause = AST_CAUSE_UNALLOCATED;
            	ast_hangup(p->ast_ch);
            	p->exten[0] = '\0';
            	/* Since we send release complete here, we won't get one */
            	p->call = NULL;
	        }
    	}
    
    	if (option_verbose >= 3) 
        	ast_verbose( VERBOSE_PREFIX_3 "Stopping simple switch on '%s'\n", ss_name);

    	return NULL;
	}

	//--------------------------------------
	// Handles CCS Information
	//--------------------------------------
	void ccs_handler(short card, short span, short size, void *SampleData)
	{
    	char data[255];
    	char data2[1000];
    	unsigned char ccs_data[260];
    	short i,ret;

	    if (size>0)
    	{
        	ast_mutex_lock(&ccs_lock);
        
        	sprintf(data,"Entrou no CCS_Handler card=%d span=%d size=%d\n",card,span,size);
        	primsg(NULL, data);

        	//shows (card/framer/size)
        	sprintf(data2," RX (c: %d /spn: %d /size: %d): ", card, span, size);
        
        	for (i=0; i< size; i++) 
        	{
	            //sprintf(data,"%02x ",ccs_data[i]);
    	        ccs_data[i] = *((u8 *)SampleData + i);
        	    sprintf(data,"%02x ",ccs_data[i]);
            	strcat(data2,data);
        	}
        	//append two more bytes to pass to libpri
        	ccs_data[size]=0;
        	ccs_data[size+1]=0;
    
        	//send data to lippri
        	i= dg_pri_findindex(card,span);
        	if (i>-1) {
				if (option_debug == 30)
                	ast_log(LOG_NOTICE, "%s\n", data2);
#ifdef DEBUG
		            write_log(data2);
#endif//#ifdef DEBUG
            	ret = write(dchans[i].fds[TX], ccs_data, size+2);   //need to pass two more bytes to libpri
        
	            if (ret<1)
	                ast_log(LOG_ERROR, "\nERROR WRITE %d\n",ret);
    	        //else
        	    //    printf("\nWriting to libpri %d ret = %d\n",dchans[i].fds[TX],ret);
	        }
        	//sprintf(data,"Saiu do CCS_Handler card=%d span=%d size=%d\n",card,span,size);
        	//primsg(NULL, data);
        	ast_mutex_unlock(&ccs_lock);      
	    }
	}

	//Socket monitoring routine
	static void *dgfifo_thread(void *data)
	{
	    unsigned char ccs_data[255];
    	short i,j;
    	short k;
    	int pos;
    	short ret;
    	struct timeval tv;

    	fd_set read_fds;
    	fd_set master_fds;

#ifdef DEBUG_PRI
		    char data2[1000];
    		char data1[255];
#endif//#ifdef DEBUG_PRI
    
	    FD_ZERO(&master_fds);

	    for (i=0; i<dchans_count; i++) 
    	{   
	        if (dchans[i].fds[TX]>-1)
    	        FD_SET(dchans[i].fds[TX] , &master_fds);
    	}

    	ast_log(LOG_NOTICE, "Starting socket thread - fdmax=%d\n",fdmax);
    	thd_socket_running = 1;
    	while (thd_socket_running)
    	{
	        /* configure select() timeout to 100ms */
    	    tv.tv_sec = 0;
        	tv.tv_usec = 100000;
        
        	//copy to local copy since read_fs will be modified by select()
        	FD_ZERO(&read_fds);
        	read_fds = master_fds;
        	//waiting for ever
        	ret = select(fdmax+1, &read_fds, NULL, NULL, /*NULL*/ &tv );
        	if (ret == -1) {
	            perror("select");
    	    }

	        //printf("socket_thread: Saiu do select - ret=%d\n",ret);
    	    for (i=0; i<= fdmax; i++) 
        	{
            	if (FD_ISSET(i, &read_fds)) 
            	{ // we got one!!
                	//printf("Preparando para dar READ\n",ret);
                	//ast_mutex_lock(&ccs_lock);
                	if ( (ret=read(i, ccs_data, sizeof(ccs_data))) > 0)
	                {
    	                pos = 0;

	                    //send data to ccs
    	                k=dg_pri_findfds(i); //k receive array index
        	            if (k>-1) {
            	            ret-=2; //remove 2 bytes of the package (crc will be generated by the chip)
#ifdef DEBUG_PRI
		                        sprintf(data1,"dgfifo_thread: (%d) vai enviar CCS placa=%d span=%d size=%d\n",i, dchans[k].card, dchans[k].span,ret);
    		                    primsg(NULL, data1);
        	                	sprintf(data2," TX (c: %d /spn: %d /fds : %d): ", dchans[k].card,dchans[k].span,i);
            	            	for (j=0; j< ret; j++) 
                	        	{
                    	        	sprintf(data1,"%02x ",ccs_data[j]);
                        	    	strcat(data2,data1);
	                        	}
	    	                    if (option_debug == 30)
    	    	                	ast_log(LOG_NOTICE,"%s\n", data2);
        	    	            strcat(data2,"\n");
            	    	        primsg(NULL, data2);
#endif//#ifdef DEBUG_PRI
	                        j=dg_SendCCSBuffer(dchans[k].card, dchans[k].span, ccs_data, ret,NULL);
                        
	                        if  (j!=DG_EXIT_SUCCESS) {
    	                        ast_log(LOG_NOTICE, "Erro SendCCSBuffer cod %x\n",j);
        	                }
#ifdef DEBUG_PRI
                        		sprintf(data1,"dgfifo_thread (%d): ENVIOU CCS placa=%d span=%d size=%d\n",i, dchans[k].card, dchans[k].span,ret);
                        		primsg(NULL, data1);
#endif//#ifdef DEBUG_PRI
                    	}
                    	else
                    	{
                        	ast_log(LOG_NOTICE, "socket ERRO %d - k = %d\n", i, k);
                    	}
                	}
                	else
                    	ast_log(LOG_NOTICE, "socket READ ERROR\n");

                	//ast_mutex_unlock(&ccs_lock);
	            }
    	    }
	    }
    	return 0;
	}

	static void apply_plan_to_number(char *buf, size_t size, const struct dgv_pvt *p, const char *number, const int plan)
	{
		switch (plan) {
			case PRI_INTERNATIONAL_ISDN:		/* Q.931 dialplan == 0x11 international dialplan => prepend international prefix digits */
				snprintf(buf, size, "%s%s", p->internationalprefix, number);
				break;
			case PRI_NATIONAL_ISDN:			/* Q.931 dialplan == 0x21 national dialplan => prepend national prefix digits */
				snprintf(buf, size, "%s%s", p->nationalprefix, number);
				break;
			case PRI_LOCAL_ISDN:			/* Q.931 dialplan == 0x41 local dialplan => prepend local prefix digits */
				snprintf(buf, size, "%s%s", p->localprefix, number);
				break;
			case PRI_PRIVATE:			/* Q.931 dialplan == 0x49 private dialplan => prepend private prefix digits */
				snprintf(buf, size, "%s%s", p->privateprefix, number);
				break;
			case PRI_UNKNOWN:			/* Q.931 dialplan == 0x00 unknown dialplan => prepend unknown prefix digits */
				snprintf(buf, size, "%s%s", p->unknownprefix, number);
				break;
			default:				/* other Q.931 dialplan => don't twiddle with callingnum */
				snprintf(buf, size, "%s", number);
				break;
		}
	}

	//******************************************************************************************
	// event handler pri
	// Each event must be interpreted and translated to voicerlib events
	// 
	//******************************************************************************************
	static void pri_event_handle(struct pri *pri, pri_event *e, int dc)
	{
	    dg_event_data_structure EventContext;
    	short port,span;
    	char data[256];
    
		char nothing[256];    

	    static struct dgv_pvt *p;
	    char plancallingnum[256];

    	pthread_t threadid;
    	pthread_attr_t attr;

    	switch(e->gen.e) {
        	case PRI_EVENT_PROCEEDING:
            	EventContext.port = dg_GetAbsolutePortNumber(dchans[dc].card, 
                	                dg_pri_to_port(e->proceeding.channel) + ((dchans[dc].span-1)*30) );
                        
            	p = find_ch_port(EventContext.port);
            	//check if port is free
            	if (p) {
                	if (p->verbosity==30)
                    	ast_log(LOG_NOTICE, ">>>>>--> (idx %d) PRI_EVENT_PROCEEDING: %s (%d) - ch=%d\n", dc, pri_event2str(e->gen.e), e->gen.e, e->proceeding.channel);
            	}
            	break;
        	case PRI_EVENT_SETUP_ACK:
				if (p) {        
                	if (p->verbosity==30)
                    	ast_log(LOG_NOTICE, ">>>>>--> (idx %d) PRI_EVENT_SETUP_ACK: %s (%d) - ch=%d\n", dc, pri_event2str(e->gen.e), e->gen.e, e->ring.channel);        
				}
            	break;
        	case PRI_EVENT_INFO_RECEIVED:
            	    EventContext.port = dg_GetAbsolutePortNumber(dchans[dc].card, 
                                        dg_pri_to_port(e->ring.channel) + ((dchans[dc].span-1)*30) );
                            
                	p = find_ch_port(EventContext.port);
                	if (p)
                	{
                		if (p->verbosity==30)
                    		ast_log(LOG_NOTICE, "RECEBEU PRI_EVENT_INFO_RECEIVED-no ch=%d!!!!!!!\n",e->ring.channel);                	
                	
                    	ast_mutex_lock(&p->lock);
                    	/* queue DTMF frame if the PBX for this call was already started (we're forwarding INFORMATION further on */
                    	if (dchans[dc].overlapdial && p->call==e->ring.call && p->ast_ch) 
                    	{
                        	/* how to do that */
                        
                        	int digitlen = strlen(e->ring.callednum);
                        	char digit;
                        	int i;

                			if (p->verbosity==30)
                    			ast_log(LOG_NOTICE, "PRI_EVENT_INFO_RECEIVED: ast_set_callerid on port %d \n", EventContext.port);
                    		
							ast_set_callerid(p->ast_ch, p->cid_num, p->cid_name, NULL);

                        	for (i=0; i<digitlen; i++)
                        	{                       	
                            	digit = e->ring.callednum[i];                      
                            	{
                            		struct ast_frame f = { AST_FRAME_DTMF, digit, };
                                	ast_mutex_lock(&p->lock);
                                	{
                                    	QueueFrame(p->port, &f);
                                	}
                                	ast_mutex_unlock(&p->lock);
                            	}
                        	}
                    	}
                    	ast_mutex_unlock(&p->lock);
                	}
					break;
        	case PRI_EVENT_RING:
	            if (e->ring.channel != -1) 
    	        {
        	        //calls voicerlib event handler
            	    //abs_port must consider the card/span
                
                	EventContext.port = dg_GetAbsolutePortNumber(dchans[dc].card, 
                	     		        dg_pri_to_port(e->ring.channel) + ((dchans[dc].span-1)*30) );
                			        
                	p = find_ch_port(EventContext.port);                
                	if (p)
						if (p->verbosity==30)
						{
							ast_log(LOG_NOTICE, "PRI_EVENT_RING 1 - dchans[dc].card: %d \n", dchans[dc].card);						
							ast_log(LOG_NOTICE, "PRI_EVENT_RING 1 - e->ring.channel: %d \n", e->ring.channel);
							ast_log(LOG_NOTICE, "PRI_EVENT_RING 1 - dchans[dc].span: %d \n", dchans[dc].span);				
            				ast_log(LOG_NOTICE, "PRI_EVENT_RING 1 - e->ring.flexible: %d \n", e->ring.flexible);
            				ast_log(LOG_NOTICE, "PRI_EVENT_RING 1 - EventContext.port: %d \n", EventContext.port);            			
						}           
            	}
            	else
            	{
                	if (e->ring.flexible) {
                    	ast_mutex_lock(&iflock); {
                        	p = find_free_port(0,0, dc);	//dc indicates that function must search only inside the correct span
                    	} ast_mutex_unlock(&iflock);

                    	if (p)
                    	{ 
							EventContext.port = p->port;
						                    	
    	                    e->ring.channel = port_to_pri(p->ch_board);
        	        		if (p->verbosity==30)
            					ast_log(LOG_NOTICE, "PRI_EVENT_RING 2 >>>>>--> flex=%d ch=%d\n", e->ring.flexible, e->ring.channel);   
                    	}    
                	}
                	else
                    	ast_log(LOG_WARNING, "SETUP flexible received, but there is no port available!\n");
	            }

    	        //check if port is free
        	    if (p) {
            	    if (p->verbosity==30)
                	    ast_log(LOG_NOTICE, "pri_event_handle >>>>>--> (Card: %d Span: %d / pri=%x) PRI 2: %s (%d) - ch=%d\n", dchans[dc].card, dchans[dc].span ,(unsigned int)pri, pri_event2str(e->gen.e), e->gen.e, e->ring.channel);
                
                	if (p->ast_ch) {
                    	//port already has a channel owned
                   		if (p->verbosity==30)
                        	ast_log(LOG_WARNING, "PRI RING: Port %d is already taken\n",p->port);

						pri_hangup(pri, e->ring.call, PRI_CAUSE_REQUESTED_CHAN_UNAVAIL);
                	}
                	else
                	{
                    	ast_mutex_lock(&p->lock); 
                    	{
                        	if (p->call == e->ring.call) 
                        	{
                            	ast_log(LOG_WARNING, "Duplicate setup requested on port %d\n",p->port);
                            	ast_mutex_unlock(&p->lock);
                            	break;
                        	}
                        	p->reserved = 1;
                        	p->pri = pri;   //copy pri structure
                        	p->call = e->ring.call;
	                    } ast_mutex_unlock(&p->lock);

    	                if (p->verbosity==30)
        	            	ast_log(LOG_WARNING, "PRI_RING: callingnum: %s \n", e->ring.callingnum);

            	        apply_plan_to_number(plancallingnum, sizeof(plancallingnum), p, e->ring.callingnum, e->ring.callingplan);
                    
	                    if (p->verbosity==30)
    	                	ast_log(LOG_WARNING, "PRI_RING: number: %s %d %s\n", plancallingnum, e->ring.callingplan, e->ring.callednum);

						if (p->use_callerid) 
						{
							ast_shrink_phone_number(plancallingnum);
							ast_copy_string(p->cid_num, plancallingnum, sizeof(p->cid_num));
							
							ast_copy_string(p->cid_name, e->ring.callingname, sizeof(p->cid_name));
							p->cid_ton = e->ring.callingplan; /* this is the callingplan (TON/NPI), e->ring.callingplan>>4 would be the TON */
						
							if (p->verbosity==30)
            	            	ast_log(LOG_WARNING, "--- PRI_RING: cid_num: %s cid_name:%s\n", p->cid_num, p->cid_name);
						}
						else 
						{
							p->cid_num[0] = '\0';
							p->cid_name[0] = '\0';
							p->cid_ton = 0;
						
							if (p->verbosity==30)
                        		ast_log(LOG_WARNING, "--- PRI_RING: No CallerID information. \n");						
						}
                    
						apply_plan_to_number(p->rdnis, sizeof(p->rdnis), p,
											 e->ring.redirectingnum, e->ring.callingplanrdnis);

						/* If immediate=yes go to s|1 */
						if (p->immediate) 
						{
							if (option_verbose >= 3)
								ast_verbose(VERBOSE_PREFIX_3 "Going to extension s|1 because of immediate=yes\n");
							p->exten[0] = 's';
							p->exten[1] = '\0';
						}
						/* Get called number */
						else if (!ast_strlen_zero(e->ring.callednum)) {
							if (option_verbose >= 3)						
		                        ast_verbose(VERBOSE_PREFIX_3 "Getting extension %s  %s\n", e->ring.callednum, p->exten);
							ast_copy_string(p->exten, e->ring.callednum, sizeof(p->exten));
							ast_copy_string(p->dnid, e->ring.callednum, sizeof(p->dnid));
						} else
							p->exten[0] = '\0';
						
						/* Set DNID on all incoming calls -- even immediate */
						if (!ast_strlen_zero(e->ring.callednum))
							ast_copy_string(p->dnid, e->ring.callednum, sizeof(p->dnid));

						//------------------------------ 
						/* No number yet, but received "sending complete"? */
						if (e->ring.complete && (ast_strlen_zero(e->ring.callednum))) {
							if (option_verbose >= 3)
								ast_verbose(VERBOSE_PREFIX_3 "Going to extension s|1 because of Complete received\n");
							p->exten[0] = 's';
							p->exten[1] = '\0';
						}
						//----------------------------------- 
					
						if ( (!dchans[dc].overlapdial) &&
						     (!ast_strlen_zero(p->exten)) &&
					    	 (!ast_exists_extension(NULL, p->context, p->exten, 1, p->cid_num)) )
                		{
							if (option_verbose >= 3)
								ast_verbose(VERBOSE_PREFIX_3 "Going to extension s|1 because of invalid extension received\n");

							p->exten[0] = 's';
							p->exten[1] = '\0';
            	    	}
						
	                    /* Make sure extension exists (or in overlap dial mode, can exist) */
    	                if ((dchans[dc].overlapdial && 
        	                 ast_canmatch_extension(NULL, p->context, p->exten, 1, p->cid_num)) ||
            	            ast_exists_extension(NULL, p->context, p->exten, 1, p->cid_num) ) 
                	    {
    						p->pri_channel = e->ring.channel;
                        	if (e->ring.complete || !dchans[dc].overlapdial)
	                        {                            	
    	                        pri_proceeding(pri, e->ring.call, e->ring.channel, 0);
                            
    							if (option_verbose >= 3)                        	
            	                	ast_log(LOG_NOTICE, "Sending pri_proceeding - %d \n", e->ring.channel);                            
                	        }
                    	    else
                        	{                            
                            	pri_need_more_info(pri, e->ring.call, e->ring.channel, 1);
                            
    							if (option_verbose >= 3)                            
                            		ast_log(LOG_NOTICE, "Sending pri_need_more_info %d \n", e->ring.channel);
                        	}

	                        if (dchans[dc].overlapdial && ast_matchmore_extension(NULL, p->context, p->exten, 1, p->cid_num)) 
    	                    {                            
        	                    //overlapdial
            	                if (dgv_new(p,AST_STATE_RESERVED, p->context,DONT_START_PBX,0)!=NULL) 
                	            {                         
    								pthread_attr_init(&attr);
    								pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);                	            	       
                    	            if (ast_pthread_create(&threadid, &attr, ss_thread, p)>=0) 
                        	        {
    									if (option_verbose >= 3)
                                	    	ast_log(LOG_NOTICE,"ss_thread Created - port %d \n", e->ring.channel);                                    	
	                                }
    	                            else                                
        	                            ast_log(LOG_ERROR, "Error creating ss_thread - port %d \n", e->ring.channel);
     	                            pthread_attr_destroy(&attr);
                                    
            	                }
                	            else
                    	        {
                        	        ast_log(LOG_ERROR,"Error creating dgv_new to overlapdial - port %d \n", e->ring.channel);
                            	}
	                        }
    	                    else
        	                {
            	                EventContext.command = EV_RINGS;
                	            event_handler(&EventContext);
                    	    }
	                    }
    	                else
        	            {                 	
            	            if (option_verbose >= 3)
                	            ast_verbose(VERBOSE_PREFIX_3 "Extension '%s' in context '%s' from '%s' does not exist. Rejecting call on channel %d\n",
                    	                    p->exten, p->context, p->cid_num, p->port/*p->logicalspan, p->prioffset, pri->span*/);
	                        pri_hangup(pri, e->ring.call, PRI_CAUSE_UNALLOCATED);
    	                    p->call = NULL;
        	                p->exten[0] = '\0';
            	            p->reserved = 0;
                	    }
                	}
            	}
            	else
            	{
                	//there is no port available
	                //....
    	            ast_log(LOG_WARNING, "There is no port available.\n");
        	        pri_hangup(pri, e->ring.call, PRI_CAUSE_REQUESTED_CHAN_UNAVAIL);
	            }
    			break;
    		case PRI_EVENT_ANSWER:
    			//..pri_hangup(pri, e->answer.call, PRI_CAUSE_NORMAL_UNSPECIFIED);
            	EventContext.port = dg_GetAbsolutePortNumber(dchans[dc].card, 
            			    		dg_pri_to_port(e->ring.channel) + ((dchans[dc].span-1)*30) );
            	p = find_ch_port(EventContext.port);
            	//check if port is free
            	if (p) {
		    		if (p->verbosity==30)
		    			ast_log(LOG_NOTICE, ">>>>>--> PRI_EVENT_ANSWER (idx %d) PRI 2: %s (%d) (%d)\n", dc,pri_event2str(e->gen.e), e->gen.e, e->answer.channel);
            	
        	        if (p->ast_ch) 
            	    {
                		if (p->verbosity==30)
                   			ast_log(LOG_NOTICE, "PRI_EVENT_ANSWER: ast_set_callerid on port %d \n", EventContext.port);
                    		                	
	                	ast_set_callerid(p->ast_ch, p->cid_num, p->cid_name, NULL);
                		
						p->pri_channel = e->answer.channel;
        	            //pri_proceeding(pri, e->answer.call, e->answer.channel, 0);
            	        EventContext.command = EV_ANSWERED;
                	    event_handler(&EventContext);
                	}
            	}    		
	    		break;
    	    case PRI_EVENT_RINGING: //network side
        	    EventContext.port = dg_GetAbsolutePortNumber(dchans[dc].card, 
            			            dg_pri_to_port(e->ring.channel) + ((dchans[dc].span-1)*30) );
            	p = find_ch_port(EventContext.port);
	            //check if port is free	            
	            if (p) {    	
        	        if (p->verbosity==30)
                    	ast_log(LOG_NOTICE, ">>>>>--> PRI_EVENT_RINGING (idx %d) PRI 1: %s (%d)\n", dc,pri_event2str(e->gen.e), e->gen.e);
                	//..q931_facility(pri, e->ringing.call);
                	//..pri_answer(pri, e->ringing.call, e->ringing.channel, 0);
            	}
            	break;
        	case PRI_EVENT_HANGUP:
            	//abs_port must consider the card/span
            	EventContext.port = dg_GetAbsolutePortNumber(dchans[dc].card, 
            						dg_pri_to_port(e->hangup.channel) + ((dchans[dc].span-1)*30) );
            	//EventContext.port = dg_GetAbsolutePortNumber(dchans[dc].card, e->hangup.channel + ((dchans[dc].span-1)*30) );
            	p = find_ch_port(EventContext.port);
            	//check if port is free
            	if (p) 
            	{
                	if (p->verbosity==30) 
                    	ast_log(LOG_NOTICE, ">>>>>--> PRI_EVENT_HANGUP (idx %d) PRI 2: %s (%d) (%d)\n", dc,pri_event2str(e->gen.e), e->gen.e,p->port );

                	if (!p->alreadyhungup) 
                	{
                    	p->alreadyhungup = 1;
                    	if (p->ast_ch) 
                    	{
	                        p->ast_ch->hangupcause = e->hangup.cause;
                        
							if (p->verbosity==30)
    	    	            	ast_log(LOG_NOTICE, ">>>>>--> PRI_EVENT_HANGUP with cause %d on port %d \n", e->hangup.cause, p->port);                    
        	            
        	    	        int pri_send_ast_control = 0;
                        
                    	    switch(e->hangup.cause) 
                        	{                                                
                            	case PRI_CAUSE_USER_BUSY:
                                	if (p->verbosity==30)
                    			    	ast_log(LOG_NOTICE, ">>>>>--> PRI_EVENT_HANGUP with cause PRI_CAUSE_USER_BUSY on %d \n", p->port);                        	
									pri_send_ast_control = 1;
    	                        	break;
        	                	case PRI_CAUSE_CALL_REJECTED:
		    						if (p->verbosity==30)
                	        			ast_log(LOG_NOTICE, ">>>>>--> PRI_EVENT_HANGUP with cause PRI_CAUSE_CALL_REJECTED on %d \n", p->port);
									pri_send_ast_control = 2;
                    				break;
                        		case PRI_CAUSE_NETWORK_OUT_OF_ORDER:
					    			if (p->verbosity==30)
                    	    			ast_log(LOG_NOTICE, ">>>>>--> PRI_EVENT_HANGUP with cause PRI_CAUSE_NETWORK_OUT_OF_ORDER on %d \n", p->port);
	                    	        pri_send_ast_control = 2;
    	                			break;
        	                	case PRI_CAUSE_NORMAL_CIRCUIT_CONGESTION:
	    							if (p->verbosity==30)
                	        			ast_log(LOG_NOTICE, ">>>>>--> PRI_EVENT_HANGUP with cause PRI_CAUSE_NORMAL_CIRCUIT_CONGESTION on %d \n", p->port);
									pri_send_ast_control = 2;
                    				break;                        			                             
                        		case PRI_CAUSE_SWITCH_CONGESTION:
				    				if (p->verbosity==30)
                        				ast_log(LOG_NOTICE, ">>>>>--> PRI_EVENT_HANGUP with cause PRI_CAUSE_SWITCH_CONGESTION on %d \n", p->port);
									pri_send_ast_control = 2;
    	                			break;                        			                             
        	            	    case PRI_CAUSE_DESTINATION_OUT_OF_ORDER:
								    if (p->verbosity==30)
                	    			    ast_log(LOG_NOTICE, ">>>>>--> PRI_EVENT_HANGUP with cause PRI_CAUSE_DESTINATION_OUT_OF_ORDER on %d \n", p->port);
									pri_send_ast_control = 2;
                    				break;                    			                                 
                        		case PRI_CAUSE_NORMAL_TEMPORARY_FAILURE:
	    							if (p->verbosity==30)
                        				ast_log(LOG_NOTICE, ">>>>>--> PRI_EVENT_HANGUP with cause PRI_CAUSE_NORMAL_TEMPORARY_FAILURE on %d \n", p->port);
									pri_send_ast_control = 2;
    	                			break;                        			
        	        	        default:
									if (p->verbosity==30)
    	        	        			ast_log(LOG_NOTICE, ">>>>>--> PRI_EVENT_HANGUP with cause default on %d \n", p->port);                    	
        	        	    		pri_send_ast_control = 3;
            	        			break;                	        
	                        }
                        
	            	    	struct ast_frame f;
    	                	switch (pri_send_ast_control)
        	            	{
	        	            	case 1://case PRI_CAUSE_USER_BUSY
					        		if (p->verbosity==30)
        	    						ast_log(LOG_NOTICE, "Sending AST_CONTROL_BUSY into PRI_EVENT_HANGUP on '%d' \n", p->port);
            						                    	
                	    	        f.frametype = AST_FRAME_CONTROL;
                    	    	    f.subclass = AST_CONTROL_BUSY;
                        	    	ast_mutex_lock(&p->lock); 
                            		{
		                            	QueueFrame(p->port, &f);
    		                        } 
        		                    ast_mutex_unlock(&p->lock);          			
            		        		break;
                		    	case 2://case PRI_CAUSE_ALL
			        				if (p->verbosity==30)
            							ast_log(LOG_NOTICE, "Sending AST_CONTROL_CONGESTION into PRI_EVENT_HANGUP on '%d' \n", p->port);
            					
	                        	    f.frametype = AST_FRAME_CONTROL;
    	                        	f.subclass = AST_CONTROL_CONGESTION;
	        	                    ast_mutex_lock(&p->lock); 
    	        	                {
        	        	            	QueueFrame(p->port, &f);
            	        	        } 
                	        	    ast_mutex_unlock(&p->lock);            			                    	
                    				break;
	                    		case 3://case default
			        				if (p->verbosity==30)
			        					ast_log(LOG_NOTICE, "Sending AST_FRAME_VOICE into PRI_EVENT_HANGUP on '%d' \n", p->port);

	                	        	p->ast_ch->_softhangup |= AST_SOFTHANGUP_DEV;

			        				f.frametype = AST_FRAME_VOICE;
			        				f.subclass = p->ast_ch->rawreadformat;
			        				
			        				switch (p->ast_ch->rawreadformat)
			        				{
			        					case AST_FORMAT_SLINEAR:
					        				f.datalen = DATALEN * 2;
						        			f.samples = DATALEN;
						        			break;
			        					case AST_FORMAT_GSM:
        		                            f.datalen = DATALEN;
        		                            f.samples = DATALEN * (DATALEN/GSM_DATALEN);
        		                            break;
			        				    case AST_FORMAT_ALAW:
			        					case AST_FORMAT_ULAW:
			        						f.datalen = DATALEN;
			        						f.samples = DATALEN;
			        						break;
			        					default:
			        						f.datalen = DATALEN;
			        						f.samples = DATALEN;
			        						break;
			        				}//switch (p->ast_ch->rawreadformat)

			        				f.data = nothing + AST_FRIENDLY_OFFSET;
			        				f.src = "chan_dgv";

			        				ast_mutex_lock(&p->lock);
			        				{
			        					QueueFrame(p->port, &f);
			        				}
			        				ast_mutex_unlock(&p->lock);
        	    	        		break;
            	    	    	default:
			    	    			if (p->verbosity==30)
	            						ast_log(LOG_WARNING, "Error, case default on pri_send_ast_control(%d), so nothing to do on '%d' \n", pri_send_ast_control, p->port);                    	
                    				break;
                    		}//switch (pri_send_ast_control)                        
	                    }
    	            }
        	        else {    	
	                	if (p->verbosity==30)
    	                	ast_log(LOG_WARNING, "RECEBEU PRI_EVENT_HANGUP-no ch=%d chanpos=%d - passo 3\n",e->hangup.channel, p->port);
	
    	                if (p->pri && p->call) {
        	                pri_hangup(p->pri, p->call, e->hangup.cause);
            	            p->call = NULL;
                	        p->reserved = 0;
                    	}
                	}
                	if (e->hangup.cause == PRI_CAUSE_REQUESTED_CHAN_UNAVAIL) {

                    	if (option_verbose >= 3)
                        	ast_verbose(VERBOSE_PREFIX_3 "Forcing restart of channel %d (card %d / span %d) since channel reported in use\n", EventContext.port,
                                        dchans[dc].card+1, dchans[dc].span+1);

                    	pri_reset(pri, e->hangup.channel);
                    	//todo: verificar estado resetting
                    	//pri->pvts[chanpos]->resetting = 1;
                    	//ast_log(LOG_WARNING, "RECEBEU PRI_EVENT_HANGUP_REQ-no ch=%d chanpos=%d -- passo 4\n",e->hangup.channel,chanpos);
                	}
            	}
            	else {
                	ast_log(LOG_WARNING, "Hangup on bad channel %d\n",EventContext.port); 
				}
	    		break;
    	    case PRI_EVENT_HANGUP_REQ: //network side
        	    //abs_port must consider the card/span
            	EventContext.port = dg_GetAbsolutePortNumber(dchans[dc].card, 
            						dg_pri_to_port(e->hangup.channel) + ((dchans[dc].span-1)*30) );
            	//EventContext.port = dg_GetAbsolutePortNumber(dchans[dc].card, e->hangup.channel + ((dchans[dc].span-1)*30) );
            	p = find_ch_port(EventContext.port);
            	//check if port is free
            	if (p) {
	                if (p->verbosity==30)
    	                ast_log(LOG_NOTICE, ">>>>>--> PRI_EVENT_HANGUP_REQ (idx %d) PRI 1: %s (%d) (%d)\n", dc,pri_event2str(e->gen.e), e->gen.e, p->port);

        	        if (p->ast_ch) 
            	    {
                	    p->ast_ch->hangupcause = e->hangup.cause;
                    
						if (p->verbosity==30)
    	                	ast_log(LOG_NOTICE, ">>>>>--> PRI_EVENT_HANGUP_REQ with cause %x on port %d \n", e->hangup.cause, p->port);                    
                    
        	            int pri_send_ast_control = 0;
                    
            	        switch(e->hangup.cause)
                	    {
                    		case PRI_CAUSE_USER_BUSY:
								if (p->verbosity==30)
                    				ast_log(LOG_NOTICE, ">>>>>--> PRI_EVENT_HANGUP_REQ with cause PRI_CAUSE_USER_BUSY on %d \n", p->port);                    			
	                    		pri_send_ast_control = 1;                      	
    	                    	break;
        	            	case PRI_CAUSE_CALL_REJECTED:
								if (p->verbosity==30)
                	    			ast_log(LOG_NOTICE, ">>>>>--> PRI_EVENT_HANGUP_REQ with cause PRI_CAUSE_CALL_REJECTED on %d \n", p->port);
                    			pri_send_ast_control = 2;                    			
                    			break;
                    		case PRI_CAUSE_NETWORK_OUT_OF_ORDER:
								if (p->verbosity==30)
                    				ast_log(LOG_NOTICE, ">>>>>--> PRI_EVENT_HANGUP_REQ with cause PRI_CAUSE_NETWORK_OUT_OF_ORDER on %d \n", p->port);
                    			pri_send_ast_control = 2;
                    			break;	                             
                    		case PRI_CAUSE_NORMAL_CIRCUIT_CONGESTION:
								if (p->verbosity==30)
                    				ast_log(LOG_NOTICE, ">>>>>--> PRI_EVENT_HANGUP_REQ with cause PRI_CAUSE_NORMAL_CIRCUIT_CONGESTION on %d \n", p->port);
                    			pri_send_ast_control = 2;
                    			break;	                             
                    		case PRI_CAUSE_SWITCH_CONGESTION:
								if (p->verbosity==30)
                    				ast_log(LOG_NOTICE, ">>>>>--> PRI_EVENT_HANGUP_REQ with cause PRI_CAUSE_SWITCH_CONGESTION on %d \n", p->port);
                    			pri_send_ast_control = 2;
                    			break;	                             
                    		case PRI_CAUSE_DESTINATION_OUT_OF_ORDER:
								if (p->verbosity==30)
    	                			ast_log(LOG_NOTICE, ">>>>>--> PRI_EVENT_HANGUP_REQ with cause PRI_CAUSE_DESTINATION_OUT_OF_ORDER on %d \n", p->port);
        	            		pri_send_ast_control = 2;
            	        		break;	                             
                	    	case PRI_CAUSE_NORMAL_TEMPORARY_FAILURE:
								if (p->verbosity==30)
                    				ast_log(LOG_NOTICE, ">>>>>--> PRI_EVENT_HANGUP_REQ with cause PRI_CAUSE_NORMAL_TEMPORARY_FAILURE on %d \n", p->port);
                    			pri_send_ast_control = 2;
	                    		break;
    	                	default:  
								if (p->verbosity==30)
            	        			ast_log(LOG_NOTICE, ">>>>>--> PRI_EVENT_HANGUP_REQ with cause default on %d \n", p->port);                    	
                	    		pri_send_ast_control = 3;
                    			break;
                    	}
                                			
            	    	struct ast_frame f;
                    	switch (pri_send_ast_control)
                    	{
                    		case 1://case PRI_CAUSE_USER_BUSY
			        			if (p->verbosity==30)
            						ast_log(LOG_NOTICE, "Sending AST_CONTROL_BUSY into PRI_EVENT_HANGUP_REQ on '%d' \n", p->port);
            					                    	
                            	f.frametype = AST_FRAME_CONTROL;
                            	f.subclass = AST_CONTROL_BUSY;
                            	ast_mutex_lock(&p->lock); 
                            	{
                            		QueueFrame(p->port, &f);
                            	} 
                            	ast_mutex_unlock(&p->lock);          			
                    			break;
                    		case 2://case PRI_CAUSE_ALL
			        			if (p->verbosity==30)
            						ast_log(LOG_NOTICE, "Sending AST_CONTROL_CONGESTION into PRI_EVENT_HANGUP_REQ on '%d' \n", p->port);
            					
	                            f.frametype = AST_FRAME_CONTROL;
    	                        f.subclass = AST_CONTROL_CONGESTION;
        	                    ast_mutex_lock(&p->lock); 
            	                {
                	            	QueueFrame(p->port, &f);
                    	        } 
                        	    ast_mutex_unlock(&p->lock);            			                    	
                    			break;
                    		case 3://case default
			        			if (p->verbosity==30)
            						ast_log(LOG_NOTICE, "Sending AST_FRAME_VOICE into PRI_EVENT_HANGUP_REQ on '%d' \n", p->port);
			        			
                        		p->ast_ch->_softhangup |= AST_SOFTHANGUP_DEV;			        			

		        				f.frametype = AST_FRAME_VOICE;
		        				f.subclass = p->ast_ch->rawreadformat;
		        				
		        				switch (p->ast_ch->rawreadformat)
		        				{
		        					case AST_FORMAT_SLINEAR:
				        				f.datalen = DATALEN * 2;
					        			f.samples = DATALEN;
					        			break;
		        					case AST_FORMAT_GSM:
    		                            f.datalen = DATALEN;
    		                            f.samples = DATALEN * (DATALEN/GSM_DATALEN);
    		                            break;
		        				    case AST_FORMAT_ALAW:
		        					case AST_FORMAT_ULAW:
		        						f.datalen = DATALEN;
		        						f.samples = DATALEN;
		        						break;
		        					default:
		        						f.datalen = DATALEN;
		        						f.samples = DATALEN;
		        						break;		        						
		        				}//switch (p->ast_ch->rawreadformat)

		        				f.data = nothing + AST_FRIENDLY_OFFSET;
		        				f.src = "chan_dgv";								

	            				ast_mutex_lock(&p->lock);
    	        				{
        	    					QueueFrame(p->port, &f);
            					}
            					ast_mutex_unlock(&p->lock);
                    			break;
	                    	default:
				        		if (p->verbosity==30)
        	    					ast_log(LOG_WARNING, "Error, case default on pri_send_ast_control_REQ(%d), so nothing to do on '%d' \n", pri_send_ast_control, p->port);                    	
            	        		break;
                	    }//switch (pri_send_ast_control)
	                }
    	            else 
        	        {            	
    	                if (p->verbosity==30)
        	                ast_log(LOG_NOTICE, "RECEBEU PRI_EVENT_HANGUP_REQ -no ch=%d chanpos=%d - hangup cause %x\n",e->hangup.channel, p->port, e->hangup.cause);

                    	p->alreadyhungup = 1;

	                    sprintf(data,"pri_hangup - p3 - (%d)\n",p->port);
    	                primsg(NULL, data);

                    	pri_hangup(p->pri, p->call, e->hangup.cause);
                    	p->call = NULL;
                	}
                
                	if (e->hangup.cause == PRI_CAUSE_REQUESTED_CHAN_UNAVAIL) 
                	{
	                    if (option_verbose >= 3)
    	                    ast_verbose(VERBOSE_PREFIX_3 "Forcing restart of channel %d (card %d / span %d) since channel reported in use\n", EventContext.port,
                                        dchans[dc].card+1, dchans[dc].span+1);

	                    pri_reset(pri, e->hangup.channel);
    	                //todo: verificar estado resetting
        	            //pri->pvts[chanpos]->resetting = 1;
            	        //ast_log(LOG_WARNING, "RECEBEU PRI_EVENT_HANGUP_REQ-no ch=%d chanpos=%d -- passo 4\n",e->hangup.channel,chanpos);
                	}
            	}
            	else {
	                ast_log(LOG_WARNING, "Hangup on bad channel %d\n",EventContext.port); 
    	        }
	            break;
    	    case PRI_EVENT_HANGUP_ACK:			
            	//abs_port must consider the card/span
            	EventContext.port = dg_GetAbsolutePortNumber(dchans[dc].card, 
            						dg_pri_to_port(e->hangup.channel) + ((dchans[dc].span-1)*30) );
            
            	//EventContext.port = dg_GetAbsolutePortNumber(dchans[dc].card, e->hangup.channel + ((dchans[dc].span-1)*30) );
            	p = find_ch_port(EventContext.port);
            	//check if port is free
            	if (p) {
                	if (p->verbosity==30)
                    	ast_log(LOG_NOTICE, ">>>>>--> PRI_EVENT_HANGUP_ACK (idx %d) PRI 1: %s (%d) (%d)\n", dc,pri_event2str(e->gen.e), e->gen.e, p->port);

                	p->call = NULL;
                	p->pri = NULL;
                	p->reserved=0;
            	}
            	break;
        	case PRI_EVENT_DCHAN_DOWN:
	            ast_log(LOG_ERROR, ">>>>>--> LINK DOWN (idx %d) PRI 2 (default): %s (%d) - dc = %d\n", dc,pri_event2str(e->gen.e), e->gen.e, dc);
            	break;
        	case PRI_EVENT_RESTART:
	        	//if (p->verbosity==30)             
    	        ast_log(LOG_NOTICE, ">>>>>--> PRI_EVENT_RESTART (idx %d)\n", dc);
                            
        		//dchans[dc].alarmed = 0;
            	break;
        	case PRI_EVENT_DCHAN_UP:
	            ast_log(LOG_NOTICE, ">>>>>--> LINK UP (idx %d)\n", dc);
    	        //When link is up, all ports from this span will be restarted

        	    if (!dchans[dc].reseting && dchans[dc].alarmed) {
    	            dchans[dc].reseting = 1;
	                for (port=1; port<=dg_GetCardPortsCount(dchans[dc].card); port++) {
        	            if (port<=30)   
            	            span=1;
                	    else
                    	    span=2;
                    
                    	if (span==dchans[dc].span) {
	                        p = find_ch_port(port);
    	                    if (p) {
        	                    if (p->pri) {
            	                    if (p->call) {
                	                    pri_hangup(p->pri, p->call, -1);
                    	                pri_destroycall(p->pri, p->call);
                        	            p->call = NULL;
                            	    }
                                	p->pri = NULL;
	                            }
    	                        if (p->ast_ch)
        	                    {
            	                    ast_log(LOG_NOTICE, ">>>>>--> Sending hangup to reset channels\n");
                	                //has a channel associated
                    	            
                        	        struct ast_frame f = { AST_FRAME_CONTROL, AST_CONTROL_HANGUP, };
                            	    ast_mutex_lock(&p->lock); {
                                	    QueueFrame(p->port, &f);
	                                } ast_mutex_unlock(&p->lock);
    	                            p->call_state=DG_AST_NONE;                                
	                            }
    	                        p->reserved=0;
        	                    dchans[dc].alarmed = 0;
            	                if (p->verbosity==30)
                	                ast_log(LOG_NOTICE, ">>>>>--> Reseting port %d card %d span %d\n", port, dchans[dc].card,dchans[dc].span);
                    	    }
            	        }
        	        }
    	            dchans[dc].reseting = 0;
	            }
	            break;
    	    default:
        	    //if (p->verbosity==30) 
            	ast_log(LOG_NOTICE, ">>>>>--> (idx %d) PRI 2 (default): %s (%d) - dc = %d (case default) \n", dc,pri_event2str(e->gen.e), e->gen.e, dc);
            	break;
		}
	}

	//---------------------------------------------------------------
	// Function that handles dchannels data from pri
	//---------------------------------------------------------------
	static void *dchan(void *data)
	{      
		// Joint D-channel
		struct timeval *next, tv;
		pri_event *e;
		fd_set fds;
		int res;
    	int dc=0;
    	//int max_fds=0;
    	//short card,span;
    	fd_set master_fds;
    	/* stores pri file descriptor */
    	FD_ZERO(&master_fds);
    	for (dc=0; dc<dchans_count; dc++) 
    	{   
	        if (dchans[dc].fds[RX]>-1)
            	FD_SET(dchans[dc].fds[RX] , &master_fds);
    	}

		thd_pri_running = 1;

    	while(thd_pri_running)
    	{
	        ast_mutex_lock(&ccs_lock);
        	/* runs through pris */
        	FD_ZERO(&fds);
        	fds = master_fds;

	        //send data to ccs
        	for (dc=0; dc<dchans_count; dc++) 
        	{
	            if (dchans[dc].pri!=0) 
            	{
	                if ((next = pri_schedule_next(dchans[dc].pri))) 
                	{
	                    gettimeofday(&tv, NULL);
	                    tv.tv_sec = next->tv_sec - tv.tv_sec;
    	                tv.tv_usec = next->tv_usec - tv.tv_usec;
        	            if (tv.tv_usec < 0) {
            	            tv.tv_usec += 1000000;
                	        tv.tv_sec -= 1;
                    	}
	                    if (tv.tv_sec < 0) {
    	                    tv.tv_sec = 0;
        	                tv.tv_usec = 0;
            	        }
                	}
            	}
        	}
        	ast_mutex_unlock(&ccs_lock);

			res = select(fdmax + 2, &fds, NULL, NULL, next ? &tv : NULL);
        
        	ast_mutex_lock(&ccs_lock);

			if (res < 0) {
	            ast_log(LOG_ERROR, ">>>>>--> Error in Select() controling dchan (isdn)");
			} else if (res==0) {
        	    //select returns timeout - no events
	            for (dc=0; dc<dchans_count; dc++) 
    	        {
        	        /* Just a timeout, run the scheduler */
            	    if (dchans[dc].pri!=0)
                	{
	                    e = pri_schedule_run(dchans[dc].pri);
    	            }
        	    }
			} else {
            	for (dc=0; dc<dchans_count; dc++) 
            	{
	                if (FD_ISSET(dchans[dc].fds[RX],  &fds)) {
    	                if (dchans[dc].pri!=0)
        	            {
            	            e = pri_check_event(dchans[dc].pri);
                	        if (e) {
                    	        //trata os eventos de pri
                        	    pri_event_handle(dchans[dc].pri, e, dc);
                        	}
                    	}
                	}
            	}
			}
			ast_mutex_unlock(&ccs_lock);
		}
		return NULL;
	}
#endif//#ifdef USE_LIBPRI

//------------------------------------------------------------------------------------
//thread used to outgoing call(fxs card), like ss_thread used in isdn with overlapdial
//------------------------------------------------------------------------------------
static void *fxss_thread(void *data)
{
	struct dgv_pvt *p = data;    
    char exten[AST_MAX_EXTENSION]="";
    int len = 0;
    int res;    
    char fxss_name[256];
    
	pthread_t threadid;
	pthread_attr_t attr;
	
    if ((p) && (p->ast_ch))
    { 	   
        if (option_verbose >= 3) 
            ast_verbose(VERBOSE_PREFIX_3 "Starting fx simple switch on '%s'\n", p->ast_ch->name);

		p->fxss_thread_running = 1;

		strncpy(fxss_name, p->ast_ch->name, sizeof(fxss_name));

        /* Now loop looking for an extension */
        ast_copy_string(exten, p->exten, sizeof(exten));
        len = strlen(exten);
        res = 0;
      
		while ((len < AST_MAX_EXTENSION-1) && 
		       (ast_matchmore_extension(p->ast_ch, p->ast_ch->context, exten, 1, p->cid_num)) && 
		       (!p->immediate))
        {
			if ((p) && (p->ast_ch))
			{
				if (len == 0)
					p->fv_thread_timeout = p->dialtone_timeout;
				else				
	            	if (ast_exists_extension(p->ast_ch, p->ast_ch->context, exten, 1, p->cid_num))
        	        	p->fv_thread_timeout = matchdigittimeout;
            		else
	                	p->fv_thread_timeout = gendigittimeout;
   			
				/*chan_zap.c send a AST_FRAME_VOICE in zt_read, here we call a thread to send it*/   			
    			pthread_attr_init(&attr);
    			pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);    			
            	if (ast_pthread_create(&threadid, &attr, fv_thread, p)>=0)
            	{
	    			if (option_verbose >= 3)
    	            	ast_log(LOG_NOTICE,"fv_thread Created - port %s \n", p->ast_ch->name);
        	    }
            	else
	            {
    	        	ast_log(LOG_ERROR, "Error creating fv_thread - port %s \n", p->ast_ch->name);
        	    }        	    
        	    pthread_attr_destroy(&attr);        	    

            	res = ast_waitfordigit(p->ast_ch, p->fv_thread_timeout);
			
				if (p->ast_ch)
				{
					if (dg_GetPortStatus(p->port)==spNone)
					{
    					if (option_verbose >= 3) 
        					ast_verbose(VERBOSE_PREFIX_3 "fxss_thread: Calling ast_hangup because '%s' already hungup\n", fxss_name);
						
						ast_hangup(p->ast_ch);
						p->fxss_thread_running = 0;						
						return NULL;
					}
				}
				else
				{
					if (option_verbose >= 3) 
        				ast_verbose(VERBOSE_PREFIX_3 "!p or !p->ast_ch, so stopping fx simple switch on '%s'\n", fxss_name);
					p->fxss_thread_running = 0;        								
    				return NULL;
				}

            	if (res < 0) 
            	{
                	ast_log(LOG_DEBUG, "fxss_thread: waitfordigit returned < 0...\n");
					ast_hangup(p->ast_ch);
					p->fxss_thread_running = 0;                
                	return NULL;
            	} 
            	else 
            		if (res) 
            		{
                		exten[len++] = res;
                		exten[len] = '\0';
            		} 
            		else
            		{   
        		    	if (option_verbose >= 3)
		        			ast_verbose(VERBOSE_PREFIX_3 "fxss_thread: Timeout on '%s'\n", fxss_name);
		                break;          			         	
            		}
			}
			else
			{
            	if (option_verbose >= 3)
		        	ast_verbose(VERBOSE_PREFIX_3 "fxss_thread: already hangup on '%s'\n", fxss_name);
                break;
			}
			usleep(1);
        }//while
       
		if (p)
		{
	        /* if no extension was received ('unspecified') on overlap call, use the 's' extension */
    	    if ((ast_strlen_zero(exten)) && (!p->immediate)) 
        	{            
            	if (option_verbose >= 3)
                	ast_verbose(VERBOSE_PREFIX_3 "Calling dgv_indicate because of empty extension received on overlap call \n");           
            
	            if (p->ast_ch)
    	        {          
#ifdef CC_AST_HAS_INDICATE_DATA        
				        dgv_indicate(p->ast_ch, AST_CONTROL_CONGESTION, NULL, 0);
#else//#ifdef CC_AST_HAS_INDICATE_DATA	
    				    dgv_indicate(p->ast_ch, AST_CONTROL_CONGESTION);
#endif//#ifdef CC_AST_HAS_INDICATE_DATA
	            }
    	    }   
        	else//if ((ast_strlen_zero(exten)) && (!p->immediate))
        	{         	
        		if (p->ast_ch) 
        		{
		        	if (ast_exists_extension(p->ast_ch, p->ast_ch->context, exten, 1, p->cid_num)) 
		        	{
	    	        	if (option_verbose >= 3)
    	    	        	ast_verbose(VERBOSE_PREFIX_3 "Starting the real PBX... \n");
                		        	
    		    	    /* Start the real PBX */           
        		    	ast_copy_string(p->ast_ch->exten, exten, sizeof(p->ast_ch->exten));
        	    
		        	    ast_setstate(p->ast_ch, AST_STATE_RING);
    		        	p->call_state=DG_AST_RINGING;         	
            	
	    	            if (strlen(p->szPortID) > 1)
                 		{
							ast_set_callerid(p->ast_ch, p->szPortID, NULL, NULL);
									
							if (option_verbose >= 3)
                        		ast_verbose(VERBOSE_PREFIX_3 "fxss_thread (%d): set callerid %s \n", p->port, p->szPortID);	    	            				
    	                } 
    	                else             		
    	                {
    	                	if (option_verbose >= 3)
			                	ast_verbose(VERBOSE_PREFIX_3 "fxss_thread (%d): No callerid information. \n", p->port);
    	                }            	
            	
	    		        res = ast_pbx_run(p->ast_ch);
    	    		    if (res) 
    	        		{
        	        		ast_log(LOG_WARNING, "PBX exited non-zero!\n");
		            	}         
    		    	} 
    		    	else//if (ast_exists_extension(p->ast_ch, p->ast_ch->context, exten, 1, p->cid_num)) 
    	    		{
		            	if (option_verbose >= 3)
    		    		{
            				if (p->immediate)
        	    			{
            					if (option_verbose >= 3)        	    				
		            	    		ast_verbose(VERBOSE_PREFIX_3 "Going to extension s|1 because of immediate option is enable \n");
	            			}            		
    	        			else
            				{
            					if (option_verbose >= 3)            					
		    	            		ast_verbose(VERBOSE_PREFIX_3 "Going to extension s|1 because of empty extension received on overlap call \n");
	            			}	                	
		    	    	}
    	    	
    			        exten[0] = 's';
        			    exten[1] = '\0';

		        		if (ast_exists_extension(p->ast_ch, p->ast_ch->context, exten, 1, p->cid_num))
			        	{
    			        	/* Start the real PBX */           
	    	    		    ast_copy_string(p->ast_ch->exten, exten, sizeof(p->ast_ch->exten));
        		    
		        		    ast_setstate(p->ast_ch, AST_STATE_RING);
    		        		p->call_state=DG_AST_RINGING;            	
            	
	    	                if (strlen(p->szPortID) > 1)
    	                	{
								ast_set_callerid(p->ast_ch, p->szPortID, NULL, NULL);
									
								if (option_verbose >= 3)
                        			ast_verbose(VERBOSE_PREFIX_3 "fxss_thread (%d): set callerid %s \n", p->port, p->szPortID);	    	            				
    	                	} 
    	                	else             		
    	                	{
    	                		if (option_verbose >= 3)
			                      	ast_verbose(VERBOSE_PREFIX_3 "fxss_thread (%d): no callerid information. \n", p->port);
    	                	}            	
            	
		    	        	res = ast_pbx_run(p->ast_ch);
	    		    	    if (res) 
    			    	    {
        			    	    ast_log(LOG_WARNING, "PBX exited non-zero!\n");
	        		   		}          	
		    	    	} 
		    	    	else//if (ast_exists_extension(p->ast_ch, p->ast_ch->context, exten, 1, p->cid_num))
	    		    	{    	    	
		    	        	if (p->ast_ch)
        			    	{
	    	    	    		ast_log(LOG_WARNING, "No such possible extension '%s' in context '%s' \n", exten, p->ast_ch->context);        		    	            
#ifdef CC_AST_HAS_INDICATE_DATA        
							        dgv_indicate(p->ast_ch, AST_CONTROL_CONGESTION, NULL, 0);
#else//#ifdef CC_AST_HAS_INDICATE_DATA
					    		    dgv_indicate(p->ast_ch, AST_CONTROL_CONGESTION);
#endif//#ifdef CC_AST_HAS_INDICATE_DATA            
        		    		}
            				else
    	    			    	ast_log(LOG_WARNING, "No such possible extension '%s' in context... \n", exten);            		    	    
		    	    	}//if (ast_exists_extension(p->ast_ch, p->ast_ch->context, exten, 1, p->cid_num))
        			}//if (ast_exists_extension(p->ast_ch, p->ast_ch->context, exten, 1, p->cid_num))
        		}//if (p->ast_ch)
        	}//if ((ast_strlen_zero(exten)) && (!p->immediate))
		}//if (p)	
    }//if ((p) && (p->ast_ch))
   
    if (option_verbose >= 3) 
        ast_verbose( VERBOSE_PREFIX_3 "Stopping fx simple switch on '%s'\n", fxss_name);

	if (p)
		p->fxss_thread_running = 0;
		
    return NULL;
}

/*Send frametype as AST_FRAME_VOICE; ***zt_read*/
static void *fv_thread(void *data)
{
    struct dgv_pvt *p = data;

	char nothing[256];
	
	char fv_name[256];
	short fv_rawreadformat;     		

    int i;

    if (p)
    {
	    if (p->ast_ch)
    	{
			int timeout = (p->fv_thread_timeout + 1000) / 20;    		
    		strncpy(fv_name, p->ast_ch->name, 100);
    		fv_rawreadformat = p->ast_ch->rawreadformat;    		
    		
        	for (i=0; i<timeout; i++)
        	{
				if ((p) && (p->ast_ch))
				{
					struct ast_frame fr = { AST_FRAME_VOICE, };
					fr.subclass = fv_rawreadformat;					
					
					switch (p->ast_ch->rawreadformat)
					{
						case AST_FORMAT_SLINEAR:
							fr.datalen = DATALEN * 2;
		        			fr.samples = DATALEN;
		        			break;
						case AST_FORMAT_GSM:
		                    fr.datalen = DATALEN;
		                    fr.samples = DATALEN * (DATALEN/GSM_DATALEN);
		                    break;
					    case AST_FORMAT_ALAW:
						case AST_FORMAT_ULAW:
							fr.datalen = DATALEN;
							fr.samples = DATALEN;
							break;
						default:
							fr.datalen = DATALEN;
							fr.samples = DATALEN;
							break;
					}//switch (p->ast_ch->rawreadformat)				
					
					fr.data = nothing + AST_FRIENDLY_OFFSET;
					fr.src = "chan_dgv";
				
					if ((option_debug == 30) || (option_debug == 50))
						ast_verbose(VERBOSE_PREFIX_3 "Sending QueueFrame into fv_thread on '%s' \n", fv_name);        								
					
					if ((p) && (p->ast_ch))
					{										
						ast_mutex_lock(&p->lock); 
						{
							QueueFrame(p->port, &fr);
						} 
						ast_mutex_unlock(&p->lock);
						
	            		usleep(20 * 1000L);						
					}
					else						
						break;            		
				}
				else
					break;
        	}        
    	}
    }
    
	return NULL;
}


//----------------------------------------------------------------------
// sets individual card configuration
//----------------------------------------------------------------------
short set_card_config(char *s, short sync, short use_crc4)
{
	char *copy;
	char *piece;
	char *c=NULL;
	int start=0, finish=0, x;
    short num_spans=0,spn=0;
    
#ifdef USE_LIBPRI
    	char szTemp[255];
#endif//#ifdef USE_LIBPRI

	c = copy = ast_strdupa(s);
    if (!copy)
       return 0;

	while ((piece = strsep(&c, ","))) 
	{		
		if (sscanf(piece, "%d-%d", &start, &finish) == 2)
		{			
			//Range
		} 
		else 
		{
			if (sscanf(piece, "%d", &start)) 
			{
				//Just one
				finish = start;				
			} 
			else 
			{
				ast_log(LOG_ERROR, "Syntax error parsing set_card configuration '%s' at '%s'. Ignoring.\n", s, piece);
				continue;
			}
		}

		for (x = start; x <= finish; x++) 
		{
			if (x < 0 || x > dg_GetCardsCount()) 
			{
				ast_log(LOG_WARNING, "Ignoring invalid card %d (maximum port is %d)\n", x,dg_GetCardsCount());
			}
			else
			{
				if (dg_GetCardInterface(x)==DG_DIGITAL_INTERFACE) 
				{
					ast_log(LOG_NOTICE,"card=%d sync_mode = %d-%s\n",x,sync,sync_mode_desc[sync]);

                    /* check which type of framer will be used */
                    if (dg_GetCardPortsCount(x)>30)
                        num_spans = 2;
                    else
                        num_spans = 1;

                    for (spn=0; spn < num_spans; spn++) 
                    {                    	
                        if (span_type[spn] == DG_SPAN_CCS) 
                        {                        	
#ifdef USE_LIBPRI
	                            //set ccs callback - doesn't matter call several times
                            	dg_SetCCSCallback(ccs_handler);
                            	ast_log(LOG_NOTICE,"CCS TYPE card=%d span=%d\n",x,spn+1);
                            	dg_EnableCCSMode(x,spn+1);                         

                            	//starts driver to handle exchange data from libpri to voicerlib                            
                            	sprintf(szTemp, "/dev/dgvfifo%d", dev_idx++);
                            	dchans[dchans_count].fds[TX] = open(szTemp,O_RDWR  | O_NONBLOCK);
                            	ast_log(LOG_NOTICE,"Card %d - Span %d Creating %s\n", x,spn+1,szTemp);
	
                            	sprintf(szTemp,"/dev/dgvfifo%d",dev_idx++);
                            	dchans[dchans_count].fds[RX] = open(szTemp,O_RDWR | O_NONBLOCK);
                            	ast_log(LOG_NOTICE,"Card %d - Span %d Creating %s\n", x,spn+1,szTemp);
	
                            	if (dchans[dchans_count].fds[TX]<0 || dchans[dchans_count].fds[RX]<0) 
                            	{
	                                ast_log(LOG_ERROR, "Error opening /dev/dgvfifo! You need to install this device driver before!");
                                	return 0;
                            	}
                            	else
                            	{
	                                //stores card and span number
                                	dchans[dchans_count].card = x;
                                	dchans[dchans_count].span = spn + 1;
                                	dchans[dchans_count].pritype = span_pritype[spn];
                                	dchans[dchans_count].switchtype = span_switchtype[spn];
                                	dchans[dchans_count].overlapdial = span_overlapdial[spn];
                                	dchans[dchans_count].pridialplan = span_pridialplan[spn];
                                	dchans[dchans_count].prilocaldialplan = span_prilocaldialplan[spn];
	                                
                                	ast_log(LOG_NOTICE,"%d - Card %d - Span %d Creating fifo device TX=%d RX=%d\n", dchans_count, x, spn+1, dchans[dchans_count].fds[TX],
                                                                                                	dchans[dchans_count].fds[RX]);
	
                                	//save the greater file descriptor used in select()
                                	if (dchans[dchans_count].fds[TX] > fdmax)
	                                    fdmax = dchans[dchans_count].fds[RX];    //check for the last that was opened

    	                            pri_set_message(primsg);
        	                        pri_set_error(prierr);

            	                    //initializes pri structure
                	                if (!(dchans[dchans_count].pri = pri_new(dchans[dchans_count].fds[RX], dchans[dchans_count].pritype, dchans[dchans_count].switchtype))) {
                    	                ast_log(LOG_ERROR,"Fatal error starting pri structure!");
                        	            return 0;
                            	    }
                                	else
                                	{
	                                    ast_log(LOG_NOTICE,"************** Pri started for span %d\n", spn+1);
                                    	pri_set_debug(dchans[dchans_count].pri, PRI_DEBUG_LEVEL);

                                    	//set overlapdial state
                                    	pri_set_overlapdial(dchans[dchans_count].pri, dchans[dchans_count].overlapdial);
                                    	
                                    	ast_log(LOG_NOTICE, "setting card %d \n", dchans[dchans_count].card);                                    	
                                    	ast_log(LOG_NOTICE, "setting span %d \n", dchans[dchans_count].span);                                    	
                                    	ast_log(LOG_NOTICE, "setting pritype %d \n", dchans[dchans_count].pritype);
										ast_log(LOG_NOTICE, "setting switchtype %d \n", dchans[dchans_count].switchtype);
                                    	ast_log(LOG_NOTICE, "setting overlapdial %d \n", dchans[dchans_count].overlapdial);										
                                    	ast_log(LOG_NOTICE, "setting pridialplan %d \n", dchans[dchans_count].pridialplan);
                                    	ast_log(LOG_NOTICE, "setting prilocaldialplan %d \n", dchans[dchans_count].prilocaldialplan);                           	                                    	                                    	
                                	}

	                                //increment counter
    	                            dchans_count++;
        	                    }
#endif//#ifdef USE_LIBPRI
                        }
                        else
                            ast_log(LOG_NOTICE,"CAS TYPE card=%d span=%d\n", x, spn+1);                            
                    } //spans
					
					dg_SetCardSyncMode(x, sync);					

                    //crc4 options
                    switch (use_crc4)
					{
                    	case 0:
                        	dg_SetE1CRC4Option(x, CFG_FRAMER_E1_A, 0);
                        	dg_SetE1CRC4Option(x, CFG_FRAMER_E1_B, 0);
                        	ast_log(LOG_NOTICE, "card=%d -> All CRC4 Disable \n",x);                    	
                    		break;                    		
                    	case 1:
                            dg_SetE1CRC4Option(x, CFG_FRAMER_E1_A, 1);
                            dg_SetE1CRC4Option(x, CFG_FRAMER_E1_B, 0);
                            ast_log(LOG_NOTICE, "card=%d -> CRC4 Enable E1-A and Disable E1-B \n",x);                    	
                    		break;                    		
                    	case 2:
                            dg_SetE1CRC4Option(x, CFG_FRAMER_E1_A, 0);
                            dg_SetE1CRC4Option(x, CFG_FRAMER_E1_B, 1);
                            ast_log(LOG_NOTICE, "card=%d -> CRC4 Disable E1-A and Enable E1-B \n",x);                    	
                    		break;                    		
                    	case 3:
                            dg_SetE1CRC4Option(x, CFG_FRAMER_E1_A, 1);
                            dg_SetE1CRC4Option(x, CFG_FRAMER_E1_B, 1);
	                        ast_log(LOG_NOTICE, "card=%d -> All CRC4 Enable \n",x);                    	
                    		break;
					}                    		                    		                    		
			    }
			}
		}
	}
	return 1;
}

//----------------------------------------------------------------------
// sets individual port configuration
//----------------------------------------------------------------------
short set_port_config(short reload, char *s, char *port_id, char *context, short category, short min_rx, short max_rx, short interdigit_timeout, 
					  short txgain, short rxgain, short cid_pos, short echo_can,
                      char *accountcode, short signalling, short b_subscriber_hangup, short group_b_value, 
                      short silence_threshold, char *musicclass, short pickupgroup, short gen_dtmf, short gen_ringback, char *callprogress,
                      char *language, short dtmf_to_asterisk, short dialtone_timeout, short wait_dialtone, short gen_busy, 
                      short drop_tx_samples_factor, short relaxdtmf, short echo_training,
                      short ring_type, short flash_min_time, short flash_max_time, short immediate, char *mailbox,
                      short fax_detection)
{
	char *copy;
	char *piece;
	char *c=NULL;
	int start=0, finish=0, x, ret;
	struct dgv_pvt *p;
	
	c = copy = ast_strdupa(s);
    if (!copy)
       return 0;

	while ((piece = strsep(&c, ","))) 
	{
		if (sscanf(piece, "%d-%d", &start, &finish) == 2)
		{
			//Range
		} 
		else 
			if (sscanf(piece, "%d", &start)) 
			{
				//Just one
				finish = start;
			} 
			else 
			{
				ast_log(LOG_ERROR, "Syntax error parsing set_port configuration '%s' at '%s'. Ignoring.\n", s, piece);
				continue;
			}

		for (x = start; x <= finish; x++) 
		{
			if (x < 1 || x > dg_GetPortsCount()) 
			{
				ast_log(LOG_WARNING, "Ignoring invalid port %d (maximum port is %d)\n", x,dg_GetPortsCount());
			}
			else
			{
				
				p = find_ch_port(x);
				if (p!=NULL)
				{
					strncpy(p->szPortID,port_id,AST_MAX_EXTENSION);
					strncpy(p->context,context,AST_MAX_CONTEXT);
					strncpy(p->accountcode,accountcode,AST_MAX_ACCOUNT_CODE);	//account code
					p->category = category;
					//p->auto_r2dmfc = auto_r2dmfc;
					p->max_digits_rx = max_rx;
					p->interdigit_timeout = interdigit_timeout;
					p->min_digits_rx = min_rx;
					p->tx_gain = txgain;
					p->rx_gain = rxgain;
					p->ask_for_cid_position = cid_pos;
					p->echo_can = echo_can;
                    p->echo_training = echo_training;
                    p->drop_tx_samples_factor = drop_tx_samples_factor;
                    p->relaxdtmf = relaxdtmf;
					p->last_dtmf = -1;
                    p->gen_ringback = gen_ringback;
                    p->gen_busy = gen_busy;
                    p->gen_dtmf = gen_dtmf;
                    p->dtmf_to_asterisk = dtmf_to_asterisk;
                    p->b_subscriber_hangup = b_subscriber_hangup;
                    p->group_b_value = group_b_value;
                    p->silence_threshold = silence_threshold;
					p->audio_debug = 0;
                    p->verbosity = 0;
					p->faxhandled = 0;
					p->fxo_bridged_dial = 0;
										
					//fxs
					p->ring_type      = ring_type;
					p->flash_min_time = flash_min_time;
					p->flash_max_time = flash_max_time;
					p->immediate      = immediate;//fxs and pri channels
					ast_copy_string(p->mailbox, mailbox, sizeof(p->mailbox));					 
#ifdef USE_LIBPRI
	                    //isdn specific configurations
    	                ast_copy_string(p->internationalprefix, internationalprefix, sizeof(p->internationalprefix));
        	            ast_copy_string(p->nationalprefix, nationalprefix, sizeof(p->nationalprefix));
            	        ast_copy_string(p->localprefix, localprefix, sizeof(p->localprefix));
                	    ast_copy_string(p->privateprefix, privateprefix, sizeof(p->privateprefix));
                    	ast_copy_string(p->unknownprefix, unknownprefix, sizeof(p->unknownprefix));
                    	
                    	//isdn controls
                    	p->alerting = 0;
                    	p->outgoing = 0;
#endif//#ifdef USE_LIBPRI
 					p->use_callerid = use_callerid;
                    ast_copy_string(p->customcascfg, customcascfg, sizeof(p->customcascfg)); 					
 					
                    p->dialtone_timeout = dialtone_timeout;         //hardcoded - must be configured

                    p->wait_dialtone = wait_dialtone;
                    p->fax_detection = fax_detection;

                    p->linetone_timeout_sched = -1;
                    p->digit_timeout_sched = -1;                    
                    p->ring_timeout_sched = -1;                    
                    p->all_echotraining_sched = -1;                    
                    
					p->sig = signalling;   /* signalling type */
                    p->pickupgroup=pickupgroup;
                    p->flashing=0;
                    if (ast_strlen_zero(callprogress) && !ast_strlen_zero(dgv_conf.szDefaultCallProgressFile))
                    {
                        strlcpy(callprogress,dgv_conf.szDefaultCallProgressFile, TAMSTR);
                    }
                    //language
                    if (!ast_strlen_zero(language))
                    {
                        strlcpy(p->language, language, MAX_LANGUAGE);
                    }
                    else
                        strlcpy(p->language, global_language, MAX_LANGUAGE);
                    

                    ast_copy_string(p->musicclass, musicclass, sizeof(p->musicclass));  /* music on hold */

					p->tlastdigit  = ast_tvnow();

                    p->reserved = 0;

					if (option_verbose >= 3) 
						ast_log(LOG_NOTICE, "port=%d id=%s context=%s - cat=%d max_rx=%d txgain=%d rxgain=%d echocan=%s\n", p->port, 
															 p->szPortID, p->context,p->category, p->max_digits_rx, p->tx_gain,
															 p->rx_gain, p->echo_can ? "yes":"no" );

					//gain
					dg_SetPortGain(p->port, PLAY_GAIN, p->tx_gain);
                    //set rec gain instead of rx_gain to avoid affect echo and agc algorythms
					dg_SetPortGain(p->port, REC_GAIN, p->rx_gain);
                    //RX Gain is always zero
                    dg_SetPortGain(p->port, RX_GAIN, 0);
                    dg_SetPortGain(p->port, TX_GAIN, 0);

                    //set relaxdtmf
                    switch(p->relaxdtmf)
                    {
                        case 0:                        
                            dg_SetTwist(p->port, 6 , 18);
							break;
                        case 1:
                            dg_SetTwist(p->port, 9 , 12);
                            break;
                        case 2:
                            dg_SetTwist(p->port, 9 , 6 );
                            break;
                        case 3:
                            dg_SetTwist(p->port, 12, 0 );
                            break;
                    }

					//fxs configurations
					dg_SetFlash(p->port, p->flash_min_time, p->flash_max_time);
					
                    //Callprogress is started only in the beginning
                    //call progress to detect busy, etc
                    if (reload<2)
                    {
                        dg_DestroyCallProgress(p->port);
                        dg_DisableInputBuffer(p->port);
                    }

                    switch(p->sig)
                    {
            
                        case DG_SIG_R2_NDIS_FIXED:
                        case DG_SIG_R2_NDIS_VARIABLE:
                            //ast_log(LOG_NOTICE, "port %d criando thread e1\n", p->port);
                            //Creates E1 Thread and configure all necessary data
                            //RETIRA!!!!!!!!!!!!!!!!!!!!!!!
                            if (reload<2)
                                    ret=dg_CreateE1Thread(p->port);
                            else
                                ret=DG_EXIT_SUCCESS;
    
                            if (ret!=DG_EXIT_SUCCESS)
                                ast_log(LOG_ERROR, "Error creating E1 thread for port %d - cod %x\n",p->port,ret);
                            else
                            {
                                dg_SetPortId(p->port, p->szPortID);
                                //set the same category  for all channels
                                dg_ConfigE1Thread(p->port,E1CFG_GROUP_II,p->category);
                                //set to receive x digits
                                //ast_log(LOG_ERROR, "MAX DIGITS E1 thread for port %d - value %x\n",p->port,p->max_digits_rx);
    
                                if (p->sig==DG_SIG_R2_NDIS_FIXED) 
                                {
                                    dg_ConfigE1Thread(p->port,E1CFG_GROUP_B, p->group_b_value);
                                    dg_ConfigE1Thread(p->port,E1CFG_MAXDIGITS_RX, p->max_digits_rx);
                                }
                                else
                                    if (p->sig==DG_SIG_R2_NDIS_VARIABLE)
                                    {
                                        dg_ConfigE1Thread(p->port,E1CFG_MAXDIGITS_RX, 0);
                                        dg_ConfigE1Thread(p->port,E1CFG_GROUP_B, 0);
    
                                    }
                                //set to ask for cid after n digits
                                dg_ConfigE1Thread(p->port,E1CFG_SEND_ID_AFTERDIGIT,p->ask_for_cid_position);
                                //set to not generate ringback tones
                                dg_ConfigE1Thread(p->port,E1CFG_GENERATE_RINGBACK, 0);
                                //threshold during conversation
                                dg_ConfigE1Thread(p->port,SILENCE_THRESHOLD_AFTER,p->silence_threshold);
                            }
                            break;
                        case DG_SIG_PRI:
                            break;
                        case DG_SIG_CUSTOMCAS:
                            //ast_log(LOG_NOTICE, "port %d criando thread cas\n", p->port);
                            //Creates Custom CAS Thread and configure all necessary data                            
                            if (reload<2)
                                ret=dg_CreateCustomCAS(p->port, p->customcascfg);
                            else
                                ret=DG_EXIT_SUCCESS;
    
                            if (ret!=DG_EXIT_SUCCESS)
                                ast_log(LOG_ERROR, "Error creating Custom CAS thread for port %d - cod %x\n",p->port,ret);
                            else                           	
                                dg_SetPortId(p->port, p->szPortID);
                                                                
                            break;                        
                        case DG_SIG_FXO:
                            //threshold during conversation/answer
                            dg_SetSilenceThreshold(p->port, p->silence_threshold);
                            break;
            			default:
            				break;
                    }

                    if (reload<2)
                    {
                        ret = dg_CreateCallProgress(p->port, p->callprogress);
                        if (ret != DG_EXIT_SUCCESS)
                        {
                            //thread creation problem
                            if (ret != 0x455)
                                ast_log(LOG_ERROR, "Failed to create CallProgress thread - error code = %d\n",ret);
                            else
                            {
                                //message to show callprogress problem
                                //ast_log(LOG_ERROR, "Failed to create CallProgress thread - error code = %d\n",ret);
                                if (option_verbose >= 3)
                                {
	                                ast_verbose(VERBOSE_PREFIX_3 "*********************************************\n");
    	                            ast_verbose(VERBOSE_PREFIX_3 "* Unable to find the configuration file for *\n");
        	                        ast_verbose(VERBOSE_PREFIX_3 "* call progress. Check if the file:         *\n");
            	                    ast_verbose(VERBOSE_PREFIX_3 "*                  %s                       *\n", p->callprogress);
                	                ast_verbose(VERBOSE_PREFIX_3 "* is available. Without the correct file    *\n");
                    	            ast_verbose(VERBOSE_PREFIX_3 "* the channel will be unable to detect busy *\n");
                        	        ast_verbose(VERBOSE_PREFIX_3 "* and another call progress signal!!!!      *\n");
                            	    ast_verbose(VERBOSE_PREFIX_3 "*********************************************\n");
                                }
                            }
                        }	
                        //else
                        //     ast_log(LOG_ERROR, "CallProgress OK (%d) -  code = %d\n",p->port,ret);
                        // 
                        //check for specific configurations for FXO
                        switch(p->sig)
                        {
                            case DG_SIG_FXO:
                                //for fxo cards, enables CALLERID Detection
                                //we don't do autopickup
                                dg_IdleSettings(p->port,0,5,500,1,0,wtDTMF,500,20,"");
                                ret = dg_IdleStart(p->port);
                                if (ret!=DG_EXIT_SUCCESS) {
                                    ast_log(LOG_ERROR, "dg_IdleStart for port %d - cod %x\n",p->port,ret);
                                }
                                break;
                        }

                        //Enables InputBuffer at beginning 
                        if ((ret=dg_EnableInputBuffer(p->port,DG_DISABLE)) != DG_EXIT_SUCCESS)
                        {
                            if (ret==DG_ERROR_THREAD_ALREADY_RUNNING) 
                            {
                                /* try again */
                                dg_DisableInputBuffer(p->port);
                                ret=dg_EnableInputBuffer(p->port,DG_DISABLE);
                                if (ret!= DG_EXIT_SUCCESS) 
                                {
                                    ast_log(LOG_NOTICE, "Error creating InputBuffer port %d - cod =%x - state %x\n",
                                        p->port,ret, GetInputBufferStatus(p->port));
                                }

                            }
                        }

                        while(GetInputBufferStatus(p->port)!=INPUT_BUFFER_ON)
                        {
                            usleep(100);
                        }
                        dg_PauseInputBuffer(p->port,DG_PAUSE);
                    } //reload < 2
				}
			}
		}
	}
	return 0;
}

//----------------------------------------------------------------------
// Searchs groups and set port with the selected group
//----------------------------------------------------------------------
short set_port_group(short group, char *s)
{
	char *copy;
	char *piece;
	char *c=NULL;
	int start=0, finish=0, x;
	struct dgv_pvt *p;
	
	c = copy = ast_strdupa(s);
    if (!copy)
       return 0;

	while ((piece = strsep(&c, ","))) 
	{
		if (sscanf(piece, "%d-%d", &start, &finish) == 2)
		{
			//Range
		} 
		else 
			if (sscanf(piece, "%d", &start)) 
			{
				//Just one
				finish = start;
			} 
			else 
			{
				ast_log(LOG_ERROR, "Syntax error parsing group configuration '%s' at '%s'. Ignoring.\n", s, piece);
				continue;
			}

		for (x = start; x <= finish; x++) 
		{
			if (x < 0 || x > dg_GetPortsCount()) 
			{
				ast_log(LOG_WARNING, "Ignoring invalid group %d (maximum group is %d)\n", x,dg_GetPortsCount());
			}
			else
			{
				p = find_ch_port(x);
				if (p!=NULL)
				{
					p->group = group;
					if (p->verbosity==20)
						ast_log(LOG_NOTICE, "group=%d  port=%d\n", p->group, p->port);
				}
			}
		}
	}
	return 0;
}

//--------------------------------------------------------------------------
// This functions return de tone code based on tone str
//--------------------------------------------------------------------------
short tone_lookup(char *tone_str)
{
	if (strcasecmp(tone_str, "tone1")==0)
		return GENERATE_TONE1;
	else	
		return GENERATE_TONE2;
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
//Generic gettickcount
unsigned int dgv_gettick(void)
{
    struct timeval tv;
    gettimeofday(&tv, NULL);
    return tv.tv_sec*1000L+tv.tv_usec/1000L;
}


//------------------------------------------------------------
// Convert audio format constants
//------------------------------------------------------------
static inline short ast2dgvformat(int ast_format)
{
	switch(ast_format) {
		case AST_FORMAT_ALAW:
			return ffWaveALaw;
		case AST_FORMAT_SLINEAR:
			return ffWavePCM;
		case AST_FORMAT_ULAW:
			return ffWaveULaw;
		case AST_FORMAT_GSM:
			return ffGsm610;
		default:
			return -1;
	}
}

//------------------------------------------------------------
// starting dial procedure
//------------------------------------------------------------
void start_dial(struct dgv_pvt *p)
{
    short ret=1;
    
    dg_event_data_structure EventContext;
    
	struct ast_channel *chan_tmp;
    struct dgv_pvt *p_bridged;    
        
#ifdef USE_LIBPRI
    	unsigned int exclusive=0;
    	int dp_strip;
    	int prilocaldialplan;
    	int pridialplan;
    	int dc=0;
    	char *c;
#endif//#ifdef USE_LIBPRI
    if (p->verbosity==10)
        ast_log(LOG_NOTICE, "--------- start_dial --------------\n");

    switch(p->sig) 
    {   	
#ifdef USE_LIBPRI
	        case DG_SIG_PRI:
    	        //isdn dial
        	    if (!p->call) 
				{
                	p->call = pri_new_call(p->pri);
                	if (p->call==NULL) {
	                    //todo: creating error handle'
	                	p->call_state = DG_AST_NONE;
	                	ast_setstate(p->ast_ch, AST_STATE_BUSY);
	     
                	}
                	else
                	{
	                    if (p->verbosity==30)
                        	ast_log(LOG_NOTICE, "PRI Dialing using port %d to %s\n",p->port,p->szNumberToDial);
	
                    	dc = dg_pri_findpri(p->pri);
	
                    	/* traditional behavior */
                    	if (dchans[dc].pritype == PRI_NETWORK)
	                        exclusive = 0;
                    	else
	                        exclusive = 1;

	                    p->digital = IS_DIGITAL(p->ast_ch->transfercapability);
                    
    	                if (option_verbose >= 3)
            	        {
							ast_log(LOG_NOTICE, "PRI Dialing to %s using port %d cid_num: %s cid_name: %s \n", p->szNumberToDial, p->port, p->pri_cid_num, p->cid_name);                    
                	        ast_verbose(VERBOSE_PREFIX_3 "Requested transfer capability: 0x%.2x - %s\n", p->ast_ch->transfercapability, ast_transfercapability2str(p->ast_ch->transfercapability));
                    	}

                    	dp_strip = 0;
                    	pridialplan = dchans[dc].pridialplan - 1;

                    	c = p->szNumberToDial;
                    	if (pridialplan == -2) { /* compute dynamically */
	                        if (strncmp(c + p->stripmsd, p->internationalprefix, strlen(p->internationalprefix)) == 0) {
    	                        dp_strip = strlen(p->internationalprefix);
        	                    pridialplan = PRI_INTERNATIONAL_ISDN;
            	            } else if (strncmp(c + p->stripmsd, p->nationalprefix, strlen(p->nationalprefix)) == 0) {
                	            dp_strip = strlen(p->nationalprefix);
                    	        pridialplan = PRI_NATIONAL_ISDN;
                        	} else {
                            	pridialplan = PRI_LOCAL_ISDN;
                        	}
                    	}
                    	prilocaldialplan = dchans[dc].prilocaldialplan - 1;
                    
                    	ast_mutex_lock(&ccs_lock);
                        	if (pri_call(p->pri, 
                             	         p->call, 
                                 	     p->digital ? PRI_TRANS_CAP_DIGITAL : p->ast_ch->transfercapability,                     //transmode
                                    	 port_to_pri(p->ch_board),                  //card channel
                                     	 exclusive,                                 //exclusive
                                     	 0,                                         //nonisdn
                                     	 p->pri_cid_num,                            //caller number
                                     	 prilocaldialplan,                          //caller plan
                                     	 p->cid_name, 
                                     	 PRES_NUMBER_NOT_AVAILABLE /*PRES_ALLOWED_USER_NUMBER_PASSED_SCREEN*/,    
                                     	 p->szNumberToDial, 
                                     	 pridialplan /* PRI_LOCAL_ISDN */ /*PRI_NATIONAL_ISDN*/, 
                                     	 PRI_LAYER_1_ALAW)) 
                        	{
                            	//error during pri_call
                            	// todo: drop call
                            	ast_mutex_unlock(&ccs_lock);
                            	if (p->verbosity==30)
	                                ast_log(LOG_NOTICE, "ERROR PRI Dialing using port %d to %s\n",p->port,p->szNumberToDial);
    	
        	                }
            	            else
                	        {
                    	    	if (option_verbose >= 3)
                        	    	ast_log(LOG_NOTICE, "PRI Dialing using port %d to %s\n",p->port,p->szNumberToDial);
                        	    	
                            	if (p->verbosity==30)
                            	{
									if (p->digital)
										ast_log(LOG_NOTICE, "PRI Dialing(%d): is digital \n", p->port);									
									else
										ast_log(LOG_NOTICE, "PRI Dialing(%d): isn't digital \n", p->port);																			
									ast_log(LOG_NOTICE, "PRI Dialing(%d): ch_board %d \n", p->port, p->ch_board);                            		
                            		ast_log(LOG_NOTICE, "PRI Dialing(%d): port_to_pri(p->ch_board) %d \n", p->port, port_to_pri(p->ch_board));
                            		ast_log(LOG_NOTICE, "PRI Dialing(%d): exclusive %d \n", p->port, exclusive);                          		

									ast_log(LOG_NOTICE, "PRI Dialing(%d): p->pri_cid_num %s \n", p->port, p->pri_cid_num);
									ast_log(LOG_NOTICE, "PRI Dialing(%d): prilocaldialplan %d \n", p->port, prilocaldialplan);
									ast_log(LOG_NOTICE, "PRI Dialing(%d): p->cid_name %s \n", p->port, p->cid_name);
									ast_log(LOG_NOTICE, "PRI Dialing(%d): p->szNumberToDial %s \n", p->port, p->szNumberToDial);
									ast_log(LOG_NOTICE, "PRI Dialing(%d): pridialplan %d \n", p->port, pridialplan);
                            	}                      	    	
                            	
	                            //must call EV_AFTERDIAL to complete call
    	                        ast_mutex_unlock(&ccs_lock);
        	                    p->call_state = DG_AST_DIALING;
            	                ast_setstate(p->ast_ch, AST_STATE_DIALING);
                	            EventContext.command = EV_AFTERDIAL;
                    	        EventContext.port = p->port;
                        	    event_handler(&EventContext);
                        	}
	                }
				}
				break;
#endif//#ifdef USE_LIBPRI
        case DG_SIG_FXS:
        	break;                   
        case DG_SIG_FXO:                
			if (ast_strlen_zero(p->szNumberToDial))
	    	{
                if (option_verbose >= 3)
                	ast_log(LOG_NOTICE, "No number to dial, answering port %d \n", p->port);
                	
				p->fxo_bridged_dial = 1;
				
                p->ast_ch->nativeformats = AST_FORMAT_ALAW;
                ast_set_read_format(p->ast_ch, AST_FORMAT_ALAW);
                ast_set_write_format(p->ast_ch, AST_FORMAT_ALAW);

                //re-enable inputbuffer using G711 format
                dg_SetRecordFormat(p->port, ast2dgvformat(AST_FORMAT_ALAW));

                //check if there is a bridged channel with
                if ((chan_tmp=ast_bridged_channel(p->ast_ch))!=NULL)
                {
					//tell to bridged channel to stop playbuffer
                    if (chan_tmp->tech_pvt!=NULL && !strncmp(chan_tmp->name,type,3))
                    {                    	
						p_bridged = (struct dgv_pvt *)chan_tmp->tech_pvt;
                        if (p->verbosity==10)
                        	ast_log(LOG_NOTICE, "start_dial: changing bridged format for %s \n", chan_tmp->name);

                        //change bridged channel
                        chan_tmp->nativeformats = AST_FORMAT_ALAW;
                        ast_set_read_format(chan_tmp, AST_FORMAT_ALAW);
                        ast_set_write_format(chan_tmp, AST_FORMAT_ALAW);

                        //re-enable inputbuffer using G711 format
                        dg_SetRecordFormat(p_bridged->port, ast2dgvformat(AST_FORMAT_ALAW));
                    }
                }
                	
               	ast_setstate(p->ast_ch, AST_STATE_DIALING);//only to asterisk log
			
				p->call_state = DG_AST_CALLING;
				ast_setstate(p->ast_ch, AST_STATE_RINGING);
					    
                EventContext.port = p->port;
                EventContext.command = EV_ANSWERED;
                event_handler(&EventContext);
                break;
	    	}
        default:
            //dialing
            ret = dg_Dial(p->port, p->szNumberToDial, dgv_conf.afterdialpause, dgv_conf.default_dialtype);

            if (p->verbosity==10)
                ast_log(LOG_NOTICE, "Dialing using port %d to %s\n", p->port, p->szNumberToDial);

            if (ret == EXIT_SUCCESS)
            {
                p->call_state = DG_AST_DIALING;
                ast_setstate(p->ast_ch, AST_STATE_DIALING);
            }
            else
            {
                //failed to dial
                p->call_state = DG_AST_NONE;
                ast_setstate(p->ast_ch, AST_STATE_BUSY);
            }
            break;
    }
}

//------------------------------------------------------------
// thread allowing ringback tones
//------------------------------------------------------------
static void *ToneThread(void *param)
{
	//unsigned int last_msec;
	int x;
	short port = *(short *)param;
	struct dgv_pvt *p;
	unsigned int time_values[4];
	short tone_to_generate1=0,tone_to_generate2=0;
	
	short nForSleep;

	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	p = find_ch_port(port);
	if (p==NULL)
	{		
		//	ast_log(LOG_ERROR, "ToneThread ERROR: port %d\n",port);			
		return NULL;
	}
	
	if (!p->tone_thread_running)
	    return NULL;

	if (p->verbosity==40)
	    ast_log(LOG_NOTICE, ">>>>> ToneThread Starting: port %d\n",port);

	
	if (!p->tone_thread_running)
	    return NULL;

	//check tone type
	switch(p->tone_thread_type)
	{
        case AST_CONTROL_RINGING:
			if (p->verbosity==40)        
        		ast_log(LOG_NOTICE, "ToneThread: port %d - type = AST_CONTROL_RINGING \n", port);            
       	
			//copy values to ringback tones and times
			for (x=0;x<4;x++)
				time_values[x] = dgv_conf.ringback_times[x];
				
			tone_to_generate1 = dgv_conf.ringbacktone1;
			tone_to_generate2 = dgv_conf.ringbacktone2;
			
            //wait to start
            for (nForSleep = 1; nForSleep <= 5; nForSleep++)//usleep(time_values[0]*500L);
            {
				if (!p->tone_thread_running)
	    			return NULL;            	
            	
				usleep(time_values[0]*100L);
            }           
			break;
		case AST_CONTROL_BUSY:
			if (p->verbosity==40)
        		ast_log(LOG_NOTICE, "ToneThread: port %d - type = AST_CONTROL_BUSY \n", port);
		
			//copy values to busy tones and times
			for (x=0;x<4;x++)
				time_values[x] = dgv_conf.busy_times[x];

			tone_to_generate1 = dgv_conf.busytone1;
			tone_to_generate2 = dgv_conf.busytone2;
			break;		
        default:      
        	ast_log(LOG_NOTICE, "ToneThread: port %d - no tone_thread_type, type = %d \n", port, p->tone_thread_type);
            p->tone_thread_running = 0;
			break;	
	}

	while(p->tone_thread_running)
	{
		if (p->verbosity==40)
			ast_log(LOG_NOTICE, "ToneThread: port %d - generating tone step 1\n",port);

		//turn on tone 1
		dg_GenerateMF(port,tone_to_generate1,0);		

		//wait for duration
        for (nForSleep = 1; nForSleep <= 10; nForSleep++)//usleep(time_values[0]*1000L);
        {
			if ((p->sig == DG_SIG_FXS && dg_GetPortStatus(p->port)==spNone && p->ast_ch==NULL) ||
				(p->sig != DG_SIG_FXS && p->ast_ch==NULL) ||			
			    (p->tone_thread_running==0))        			
			{		    
				dg_GenerateMF(port,GENERATE_OFF,0);
	        	break;
			}
			usleep(time_values[0]*100L);
        }	

		//check if thread is already asked to terminate		
		if ((p->sig == DG_SIG_FXS && dg_GetPortStatus(p->port)==spNone && p->ast_ch==NULL) ||
			(p->sig != DG_SIG_FXS && p->ast_ch==NULL) ||			
		    (p->tone_thread_running==0))		
		{		
			dg_GenerateMF(port,GENERATE_OFF,0);
	        break;
		}		
		
        if (p->verbosity==40)
            ast_log(LOG_NOTICE, "ToneThread: port %d - waiting time1\n",port);
            
		//turn off tone 1
		dg_GenerateMF(port,GENERATE_OFF,0);

		if (p->verbosity==40)
			ast_log(LOG_NOTICE, "ToneThread: port %d - generating silence step 1\n",port);

		//wait for silence pause
        for (nForSleep = 1; nForSleep <= 10; nForSleep++)//usleep(time_values[1]*1000L);
        {
       		if ((p->sig == DG_SIG_FXS && dg_GetPortStatus(p->port)==spNone && p->ast_ch==NULL) ||
		   	    (p->sig != DG_SIG_FXS && p->ast_ch==NULL) ||			
			    (p->tone_thread_running==0))		
			{		    
				dg_GenerateMF(port,GENERATE_OFF,0);
	        	break;
			}
			        	
			usleep(time_values[1]*100L);				
        }
		
		if ((p->sig == DG_SIG_FXS && dg_GetPortStatus(p->port)==spNone && p->ast_ch==NULL) ||
			(p->sig != DG_SIG_FXS && p->ast_ch==NULL) ||			
		    (p->tone_thread_running==0))			
		{
       	    dg_GenerateMF(port,GENERATE_OFF,0);
            break;
		}
		
		//turn on tone 2
		if (p->verbosity==40)
			ast_log(LOG_NOTICE, "ToneThread: port %d - generating tone step 1\n",port);

		dg_GenerateMF(port,tone_to_generate2,0);
		
        for (nForSleep = 1; nForSleep <= 10; nForSleep++)//usleep(time_values[2]*1000L);
        {
			if ((p->sig == DG_SIG_FXS && dg_GetPortStatus(p->port)==spNone && p->ast_ch==NULL) ||
				(p->sig != DG_SIG_FXS && p->ast_ch==NULL) ||			
			    (p->tone_thread_running==0))		
			{		    
				dg_GenerateMF(port,GENERATE_OFF,0);
	        	break;
			}
			        	
			usleep(time_values[2]*100L);
        }	
		
		if ((p->sig == DG_SIG_FXS && dg_GetPortStatus(p->port)==spNone && p->ast_ch==NULL) ||
			(p->sig != DG_SIG_FXS && p->ast_ch==NULL) ||			
		    (p->tone_thread_running==0))
		{
       	    dg_GenerateMF(port,GENERATE_OFF,0);
            break;
		}
		
		//turn off tone 2
		if (p->verbosity==40)
			ast_log(LOG_NOTICE, "ToneThread: port %d - generating silence step 2\n",port);

		dg_GenerateMF(port,GENERATE_OFF,0);
		
        for (nForSleep = 1; nForSleep <= 10; nForSleep++)//usleep(time_values[3]*1000L);
        {
			if ((p->sig == DG_SIG_FXS && dg_GetPortStatus(p->port)==spNone && p->ast_ch==NULL) ||
				(p->sig != DG_SIG_FXS && p->ast_ch==NULL) ||			
			    (p->tone_thread_running==0))		
			{		    
				dg_GenerateMF(port,GENERATE_OFF,0);
	        	break;
			}
			        	
			usleep(time_values[3]*100L);
        }
		
		if ((p->sig == DG_SIG_FXS && dg_GetPortStatus(p->port)==spNone && p->ast_ch==NULL) ||
			(p->sig != DG_SIG_FXS && p->ast_ch==NULL) ||			
		    (p->tone_thread_running==0))
		{
       	    dg_GenerateMF(port,GENERATE_OFF,0);
            break;
		}
		

	}//while
	
	//turn off flags
	p->tone_thread_running = 0;
	p->tone_thread_id = 0;
	if (p->verbosity==40)
		ast_log(LOG_NOTICE, "****** ToneThread: port %d - end! \n",port);

	return NULL;
}

//------------------------------------------------------------
// thread allowing stutter dial tone
//------------------------------------------------------------
static void *StutterToneThread(void *param)
{
	short port = *(short *)param;
	struct dgv_pvt *p;
	unsigned int time_values=100;
	
	short nForStutterTone;
	short nForSleep;

	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	p = find_ch_port(port);
	if (p==NULL)
		return NULL;
	
	if ((!p->stuttertone_thread_running) || (p->tone_thread_running)) 
	    return NULL;

	if (p->verbosity==40)
	    ast_log(LOG_NOTICE, ">>>>> StutterToneThread Starting: port %d\n",port);

	
	if ((!p->stuttertone_thread_running) || (p->tone_thread_running))
	    return NULL;

	for (nForStutterTone=0; nForStutterTone<3; nForStutterTone++)
	{
		if (p->verbosity==40)
			ast_log(LOG_NOTICE, "StutterToneThread: port %d - generating stutter tone step 1\n",port);

		//turn on tone 1
		dg_GenerateMF(port, GENERATE_TONE1, 0);

		//wait for duration
        for (nForSleep = 1; nForSleep <= 10; nForSleep++)
        {
        	if ((p->ast_ch == NULL) || (!p->stuttertone_thread_running) || (p->tone_thread_running))		
			{		    
				dg_GenerateMF(port, GENERATE_OFF, 0);
	        	break;
			}
        	
			usleep(time_values*100L);
        }		
		
		//check if thread is already asked to terminate		
        if ((p->ast_ch == NULL) || (!p->stuttertone_thread_running) || (p->tone_thread_running))		
		{		    
			dg_GenerateMF(port, GENERATE_OFF, 0);
	        break;
		}
		
        if (p->verbosity==40)
            ast_log(LOG_NOTICE, "StutterToneThread: port %d - waiting time1\n",port);
            
		//turn off tone 1
		dg_GenerateMF(port, GENERATE_OFF, 0);

		if (p->verbosity==40)
			ast_log(LOG_NOTICE, "StutterToneThread: port %d - generating silence step 1\n",port);

		//wait for silence pause
        for (nForSleep = 1; nForSleep <= 10; nForSleep++)
        {
        	if ((p->ast_ch == NULL) || (!p->stuttertone_thread_running) || (p->tone_thread_running))		
			{		    
				dg_GenerateMF(port, GENERATE_OFF, 0);
	        	break;
			}
			        	
			usleep(time_values*100L);				
        }
		
		if ((p->ast_ch == NULL) || (!p->stuttertone_thread_running) || (p->tone_thread_running))			
		{
       	    dg_GenerateMF(port, GENERATE_OFF, 0);
            break;
		}
	}
	
	if ((p->ast_ch == NULL) || (!p->stuttertone_thread_running) || (p->tone_thread_running))			
    	dg_GenerateMF(port, GENERATE_OFF, 0);
	else
    	dg_GenerateMF(port, GENERATE_TONE1, 0);	
		
	//turn off flags
	p->stuttertone_thread_running = 0;
	p->stuttertone_thread_id = 0;
	if (p->verbosity==40)
		ast_log(LOG_NOTICE, "****** StutterToneThread: port %d - end! \n",port);

	return NULL;
}

//------------------------------------------------------------
// Callback that receives samples from card and passes to ast.
// size is the number of samples (not datalen)
//------------------------------------------------------------
void read_samples(short port, short size, void *SampleData)
{
	struct ast_frame fr;

	struct dgv_pvt *p;

	short format_to_use;
    int afmt;

    char *readbuf;
    
    //int j;
	//short x;    
    //char sin_tx[1024];

#ifdef DEBUG_Q
	unsigned char ctmp;
  	char szTmp[1024];
	char char_tx[1024];
	char *tmp8;
#endif//#ifdef DEBUG_Q
	p = find_ch_port(port);
	if (p)
	{	
		if (p->ast_ch == NULL)
		{
			if (p->verbosity==60)
				ast_log(LOG_WARNING, "dgv_read: port %d - ast_ch NULL \n",p->port);
			//todo: forcar desligamento
			return;
		}
		
		if (p->ast_ch->_state != AST_STATE_UP)
		{
			//channel isn't up
            if (p->verbosity==60)
                ast_log(LOG_WARNING, "dgv_read: port %d - ast_ch not UP - state is %d, but send voice...\n",p->port,p->ast_ch->_state);

			//fr.frametype = AST_FRAME_NULL;
			//return;
		}
		//else
		{
			if (p->audio_debug && option_debug==60) //120
  			    ast_log(LOG_WARNING, "dgv_read: port %d read_last_format = %x , readformat=%x - supported=%x - rawread=%d\n",port,p->read_last_format , p->ast_ch->readformat , supported_fmts,p->ast_ch->rawreadformat);

            if (p->ast_ch)            
                afmt = p->ast_ch->rawreadformat;      
            else 
                afmt = p->pref_format;
            
            if (p->read_last_format == -1 || p->read_last_format != afmt)
            {
                p->read_last_format = afmt;
                format_to_use = ast2dgvformat(afmt);
                dg_SetRecordFormat(p->port, format_to_use);
	            if (p->verbosity==60)
                {
                    ast_log(LOG_NOTICE, "dgv_read: port %d Assigning read format:%x %s\n",p->port,p->read_last_format,ast2dgvformatname(p->read_last_format));
                }
            }

            //copy data to internal buf[]
            //
            readbuf = ((char *)(p->buf)) + AST_FRIENDLY_OFFSET;
            /*if (size < DGV_MAX_BUF)
                memcpy(readbuf, SampleData , size);
            else
                memcpy(readbuf , SampleData , DGV_MAX_BUF-1);*/

			fr.frametype = AST_FRAME_VOICE;
			fr.src = type;
			fr.mallocd = 0;
			fr.delivery.tv_sec = 0;
			fr.delivery.tv_usec = 0;
			fr.offset = AST_FRIENDLY_OFFSET;
			//samples
			fr.samples = size;
			//subclass is the recordformat used
			fr.subclass = afmt;
#ifdef CC_AST_HAS_VERSION_1_4
			fr.has_timing_info = 0;
#endif//#ifdef CC_AST_HAS_VERSION_1_4

            //if its fax but not SLINEAR, drop samples
            if (p->faxhandled && p->ast_ch->nativeformats != AST_FORMAT_SLINEAR) {

                if (p->verbosity==60)
                    ast_log(LOG_NOTICE, "dgv_read: port %d - Fax handled - forcing slin\n",p->port);

                fr.frametype = AST_FRAME_NULL;
                return;
            }

            //ast_log(LOG_NOTICE, "dgv_read: port %d sending format:%x\n",p->port,afmt);
			switch(p->read_last_format)
			{
                case AST_FORMAT_SLINEAR:
        
                    if (size<=DATALEN*20 /*|| p->faxhandled*/)     //400ms
                    {
                        if (size < (DGV_MAX_BUF/2))
                            memcpy(readbuf, SampleData , size*2);
                        else
                            memcpy(readbuf , SampleData , DGV_MAX_BUF-1);

                        fr.data = (char *)readbuf;                    
                        fr.samples = size;
                        fr.datalen = size*2;

                        //ast_log(LOG_NOTICE, "dgv_read(%d): slin datalen=%d samples=%d size=%d \n",p->port, fr.datalen, fr.samples, size);

						if (dg_GetPortStatus(p->port)==spDialing)
						{
							fr.frametype = AST_FRAME_NULL;
							fr.subclass = 0;
							fr.datalen = 0;
							fr.samples = 0;
							fr.mallocd = 0;
							fr.offset = 0;
							fr.src = "chan_dgv";
							fr.data = NULL;
							if (p->verbosity==60)
								ast_log(LOG_NOTICE, "Sending AST_FRAME_NULL because dg_GetPortStatus == spDialing on %d \n", p->port);							
						}
						
                        ast_mutex_lock(&p->lock); {
                            QueueFrame(port, &fr);
                        } ast_mutex_unlock(&p->lock);

                    }
                    else
                    {
                         if (p->verbosity==60)
                            ast_log(LOG_NOTICE, "dgv_read(%d): dropping samples - too much delay datalen=%d delay=%d\n",p->port,
                                    size,size/8);
                    }
                    break;
                case AST_FORMAT_GSM:

                        if (size<=GSM_DATALEN*20)     //400ms
                        {
                            if (size < DGV_MAX_BUF)
                                memcpy(readbuf, SampleData, size);
                            else
                            {
                                memcpy(readbuf, SampleData, DGV_MAX_BUF-1);
                                size = DGV_MAX_BUF-1;
                            }

                            fr.data = (short *)readbuf;
                            fr.datalen = size; /*GSM_DATALEN * (size/GSM_DATALEN);*/ 
                            fr.samples = DATALEN * (size/GSM_DATALEN); 

                            if (option_debug==60 && p->verbosity==60)
                                //if (size>GSM_DATALEN) 
                                //{
                                ast_log(LOG_NOTICE, "dgv_read(%d): gsm datalen=%d samples=%d size=%d \n",p->port, fr.datalen, fr.samples, size);
                                //}    

							if (dg_GetPortStatus(p->port)==spDialing)
							{
								fr.frametype = AST_FRAME_NULL;
								fr.subclass = 0;
								fr.datalen = 0;
								fr.samples = 0;
								fr.mallocd = 0;
								fr.offset = 0;
								fr.src = "chan_dgv";
								fr.data = NULL;
								if (p->verbosity==60)
									ast_log(LOG_NOTICE, "Sending AST_FRAME_NULL because dg_GetPortStatus == spDialing on %d \n", p->port);								
							}

                            ast_mutex_lock(&p->lock); {
                                QueueFrame(port, &fr);
                            } ast_mutex_unlock(&p->lock);


                        }
                        else
                        {
                            if (p->verbosity==60)
                                ast_log(LOG_NOTICE, "dgv_read(%d): dropping samples - too much delay datalen=%d delay=%d\n",p->port,
                                        size,size*20/33);
                        }
						break;
			    case AST_FORMAT_ALAW:
				case AST_FORMAT_ULAW:
#ifdef DEBUG_Q
							fwrite(SampleData, 1, size, p->rec);
#endif//#ifdef DEBUG_Q
						//a_law and u_law waits for 160 samples with 160 datalen
                        if (size<=DATALEN*20 && p->force_drop_rx==0 /*|| p->faxhandled*/ )     //400ms
                        {
                        	if (size < DGV_MAX_BUF)
                            {
                                memcpy(readbuf, SampleData, size);
                            	//if (p->verbosity==60)
                                //	ast_log(LOG_NOTICE, "dgv_read(%d): size < DGV_MAX_BUF on port %d \n", p->port);
                            }    
                            else
                            {
                                memcpy(readbuf, SampleData, DGV_MAX_BUF-1);
	                            if (p->verbosity==60)
                                	ast_log(LOG_NOTICE, "dgv_read(%d): size >= DGV_MAX_BUF \n", p->port);                                
                            }

                           	fr.data = (short *)readbuf;
                           	fr.samples = size;
                           	fr.datalen = size;
                       
                           	//sin wave -
                           	/*if (p->audio_debug)
                           	{
                           	    x=0;
                            	for (j=0; j<size; j++)
                               	{	
                                	sin_tx[j] = senoide[x++];
                                  	if (x==32) x=0;
                               	}	
								fr.data = (char *)sin_tx;                      
							}*/

                            //ast_log(LOG_NOTICE, "dgv_read(%d): g711 datalen=%d samples=%d size=%d x=%d\n",p->port, fr.datalen, fr.samples, size, x);

							if (dg_GetPortStatus(p->port)==spDialing)
							{
								fr.frametype = AST_FRAME_NULL;
								fr.subclass = 0;
								fr.datalen = 0;
								fr.samples = 0;
								fr.mallocd = 0;
								fr.offset = 0;
								fr.src = "chan_dgv";
								fr.data = NULL;
								if (p->verbosity==60)
									ast_log(LOG_NOTICE, "Sending AST_FRAME_NULL because dg_GetPortStatus == spDialing on %d \n", p->port);								
							}
						
							ast_mutex_lock(&p->lock); {
                            	QueueFrame(port, &fr);                               
                            } ast_mutex_unlock(&p->lock);
                            if (option_debug == 60)
                            	ast_log(LOG_NOTICE, "dgv_read(%d): g711 datalen=%d samples=%d size=%d \n",p->port, fr.datalen, fr.samples, size);
						}
                        else
                        {

                            if (p->force_drop_rx>0) 
                                p->force_drop_rx--;

                            if (p->verbosity==60)
                                ast_log(LOG_NOTICE, "dgv_read(%d): ulaw/alaw dropping samples - too much delay datalen=%d delay=%d\n",p->port,
                                        size,size*8);
                        }
						break;
			}
			p->samples_rx_count += fr.samples;
#ifdef DEBUG_Q
/*			if (p->verbosity==60)
			{
    			if ((p->samples_rx_count%32767)==0)
				{
					ast_log(LOG_NOTICE, "dgv_read: Port %d rx_count=%d\n",p->port,p->samples_rx_count);
				}
			}*/
#endif//#ifdef DEBUG_Q
		}	
	}
}

//------------------------------------------------------------------
// Queue frames to asterisk channel
//------------------------------------------------------------------
void QueueFrame(short port, struct ast_frame *fp)
{
	struct dgv_pvt *p;
	struct ast_frame f;
	int trycnt;

	f = *fp;
	//ast_log(LOG_NOTICE, "QueueFrame 1:\n");
	p = find_ch_port(port);
		
	if (p)	
	{
		//ast_log(LOG_NOTICE, "QueueFrame 2: Port %d\n",p->port);
		/* check if has a frame to queue to Asterisk */ 
		if ((f.frametype >= 0) && (f.frametype != AST_FRAME_NULL) && (p->ast_ch))
		{
			if (p->verbosity==70 && f.frametype!=AST_FRAME_VOICE)
					ast_verbose(VERBOSE_PREFIX_3 "%s: handle_owned: Prepared frame type[%d]subclass[%d], owner=[%s]\n",
									p->dev, f.frametype, f.subclass, p->ast_ch->name);

			trycnt = 0;
			for (;;) 
			{
				if ((p) && (p->ast_ch))				 
				{
					if (ast_mutex_trylock(&p->ast_ch->lock)) 
					{
						ast_mutex_unlock(&p->lock);
						usleep(1);
						ast_mutex_lock(&p->lock);
						trycnt++;
                        if (trycnt>100)
                        {
    						if (p->verbosity==70)
    						{
    							ast_verbose(VERBOSE_PREFIX_3 "chan_dgv handled_owned (%d): Too much delay to delivery frame %d - frame type[%d]subclass[%d]\n", p->port,trycnt,f.frametype, f.subclass);
    						}
                        }
                        break;
					}
					else
					{
                         if (p->verbosity==70)
                             ast_verbose(VERBOSE_PREFIX_3 "handled_owned : frame %d\n", trycnt);
						//ast_log(LOG_NOTICE, "QueueFrame 3: Port %d - type[%d]subclass[%d]\n",p->port,f.frametype, f.subclass);

						if (p && p->ast_ch!=NULL && fp!=NULL)
						{
						    ast_queue_frame(p->ast_ch, &f);
						}
						else
						    ast_log(LOG_NOTICE, "QueueFrame 3,2: CHANNEL NULL Port %d - type[%d]subclass[%d]\n",p->port,f.frametype,f.subclass);

						if (p->ast_ch==NULL)
						    if (option_verbose >= 3)
								ast_verbose(VERBOSE_PREFIX_3 "handled_owned : ast_ch == NULL\n");

						ast_mutex_unlock(&p->ast_ch->lock);
						break;
					}
				}
				else
					break;
			}
		}
	}
}

//-------------------------------------------------------------------------------------------
// This function handles digivoice voicerlib events
//------------------------------------------------------------
void event_handler(void *context_data)
{
	//struct dg_event_data_strucuture *EventContext;
	short ret=0;
    
	short ndis_size=0;
	char cAux;
	dg_event_data_structure *EventContext;
	struct dgv_pvt *p, *p_bridged;
    
	char szTemp[TAMSTR];
	struct ast_channel *chan_tmp;
	char tmpfax[256];
	
	short sTempPort;

    //int fxss_len = 0;
#ifdef USE_LIBPRI
    	short port=0;
    	short span=0;
    	short dc=0;
#endif//#ifdef USE_LIBPRI
    pthread_t threadid;
    pthread_attr_t attr;

	if (!channel_up)
	{
        ast_log(LOG_WARNING, "*** Channel not ready to receive events! ****\n");
		return;
	}

	EventContext = ((dg_event_data_structure*)context_data);

	/* Copy received Data */
	switch (EventContext->command)
	{
		case EV_E1_ALARM:
			if (EventContext->port == 1)
				sTempPort = dg_GetAbsolutePortNumber(EventContext->card + 1, 1);
			else
				sTempPort = dg_GetAbsolutePortNumber(EventContext->card + 1, 31);

			p = find_ch_port(sTempPort);
			break;
		default:
			p = find_ch_port(EventContext->port);
			break;
	}

    if (p==NULL)
    {
		switch (EventContext->command)
		{
			case EV_E1_ALARM:   	
				ast_log(LOG_NOTICE, "  -- Event 0x%x on card %d, port %d, data %x \n", EventContext->command, EventContext->card, EventContext->port, EventContext->data);
			default:
				return;
		}
    }

	/*if (p->verbosity == 50)
	{
		ast_log(LOG_NOTICE, "  -- Event 0x%x on port %d data=%x\n",EventContext->command,EventContext->port,EventContext->data);
	}*/

	switch (EventContext->command)
	{
        case EV_ERRORDETECTED:
            ast_log(LOG_WARNING, "EV_ERRORDETECTED (%d) data %x\n",EventContext->card+1,EventContext->data);
            break;

		case EV_E1_ALARM:
			//port means E1_A ou E1_B
			switch(EventContext->data)
		    {
					case ALARM_RSLIP:		//escorregamento					
						dgv_e1alarms[EventContext->card][EventContext->port - 1].nALARM_RSLIP = EventContext->data_aux;
					
						//sprintf(szTemp, "Card %d E1-%d - RSLIP - count=%d", EventContext->card+1, EventContext->port,EventContext->data_aux);
					
						ast_log(LOG_NOTICE, "Card %d E1-%d - RSLIP - count=%d \n", EventContext->card + 1, EventContext->port, EventContext->data_aux);						
						break;
					case ALARM_RAIS:		//alarme remoto
						dgv_e1alarms[EventContext->card][EventContext->port - 1].nALARM_RAIS = EventContext->data_aux;
						
						ast_log(LOG_NOTICE, "Card %d E1-%d - RAIS (%s) \n", EventContext->card + 1, EventContext->port, showalarmstatus_desc[EventContext->data_aux]);						
						break;
					case ALARM_AISS:		//indicacao de alarme
						dgv_e1alarms[EventContext->card][EventContext->port - 1].nALARM_AISS = EventContext->data_aux;
						
						ast_log(LOG_NOTICE, "Card %d E1-%d - AISS (%s) \n", EventContext->card+1,EventContext->port, showalarmstatus_desc[EventContext->data_aux]);
						break;
					case ALARM_AIS16S:		//indicacao de alarme canal  16
						dgv_e1alarms[EventContext->card][EventContext->port - 1].nALARM_AIS16S = EventContext->data_aux;					
					
						ast_log(LOG_NOTICE, "Card %d E1-%d - AIS16S (%s) \n", EventContext->card+1,EventContext->port, showalarmstatus_desc[EventContext->data_aux]);
						break;
					case ALARM_LOSS:		//perda de sinal
						if (EventContext->data_aux)//LOSS ON
						{
							dgv_e1alarms[EventContext->card][EventContext->port - 1].nALARM_RSLIP = 0;
							dgv_e1alarms[EventContext->card][EventContext->port - 1].nALARM_RAIS = 0;
							dgv_e1alarms[EventContext->card][EventContext->port - 1].nALARM_AISS = 0;
							dgv_e1alarms[EventContext->card][EventContext->port - 1].nALARM_AIS16S = 0;						
							dgv_e1alarms[EventContext->card][EventContext->port - 1].nALARM_CRC4SYNC = 0;
							dgv_e1alarms[EventContext->card][EventContext->port - 1].nALARM_MFSYNC = 0;
							dgv_e1alarms[EventContext->card][EventContext->port - 1].nALARM_SYNC = 0;
						}				
						dgv_e1alarms[EventContext->card][EventContext->port - 1].nALARM_LOSS = EventContext->data_aux;
								
                        switch(p->sig)
                        {
#ifdef USE_LIBPRI                        	
	                            case DG_SIG_PRI:
    	                            //find index
        	                        dc = dg_pri_findindex(EventContext->card+1,EventContext->port);
        
            	                    if (EventContext->data_aux) {
                	                    ast_log(LOG_NOTICE, "Card %d E1-%d - ISDN LOSS (%s) \n", EventContext->card+1,EventContext->port, showalarmstatus_desc[EventContext->data_aux]);
                                    
                    	                //alarmed - must block isdn
                        	            dchans[dc].alarmed = 1;
                            	        //ends all calls
                                	    if (!dchans[dc].reseting) {
                                    	    dchans[dc].reseting = 1;
                                        	for (port=1; port<=dg_GetCardPortsCount(dchans[dc].card); port++) 
                                        	{
                                            	if (port<=30)   
	                                                span=1;
    	                                        else
        	                                        span=2;
                                                
            	                                if (span==dchans[dc].span) 
                	                            {                                            	
                    	                            p = find_ch_port(port);
                        	                        
                            	                    if (p) {                                                	
                                	                    if (p->pri) {                                                    	
                                    	                    if (p->call) {                                                        	
                                        	                    pri_hangup(p->pri, p->call, -1);
                                            	                pri_destroycall(p->pri, p->call);
                                                	            p->call = NULL;                                                            
                                                    	    }                                                        
                                                        	p->pri = NULL;                                                        
	                                                    }
    		                                            if (p->ast_ch)                                                    
            	                                        {                                                    	
                	                                        ast_log(LOG_NOTICE, ">>>>>--> Sending hangup to reset channels \n");
                    	                                    //has a channel associated
                        	                                struct ast_frame f = { AST_FRAME_CONTROL, AST_CONTROL_HANGUP, };
                            	                            ast_mutex_lock(&p->lock); {
                                	                            QueueFrame(p->port, &f);
                                    	                    } ast_mutex_unlock(&p->lock);
                                        	                p->call_state=DG_AST_NONE;
                                            	        }                                                    
                                                	    p->reserved=0;
                                                    	dchans[dc].alarmed = 0;
	                                                    if (p->verbosity==50)
    	                                                    ast_log(LOG_NOTICE, ">>>>>--> Reseting port %d card %d span %d \n", port, dchans[dc].card,dchans[dc].span);
        	                                        }
            	                                }
                	                        }
                    	                    dchans[dc].reseting = 0;
                        	            }
                            	    }
                                	else
	                                {
    	                                //alarm cleared
        	                            ast_log(LOG_NOTICE, "Card %d E1-%d - ISDN LOSS Cleared (%s) \n", EventContext->card+1,EventContext->port, showalarmstatus_desc[EventContext->data_aux]);
            	                        dchans[ret].alarmed = 0;       
                	                }
                    	            break;
#endif//#ifdef USE_LIBPRI
                            default:
                                ast_log(LOG_NOTICE, "Card %d E1-%d - R2 LOSS (%s) \n", EventContext->card+1,EventContext->port, showalarmstatus_desc[EventContext->data_aux]);
                                break;
                        }
						break;
					case ALARM_RESERVED:	//reservado
						dgv_e1alarms[EventContext->card][EventContext->port - 1].nALARM_CRC4SYNC = EventContext->data_aux;
						
						//sprintf(szTemp,"Card %d E1-%d - RESERVED (%s)", EventContext->card+1,EventContext->port,(EventContext->data_aux==1 ? "on":"off"));
						ast_log(LOG_NOTICE, "Card %d E1-%d - CRC4SYNC (%s) \n", EventContext->card+1,EventContext->port, showalarmstatus_desc[EventContext->data_aux]);
						break;
					case ALARM_MFSYNC:		//sincronismo de multiquadro
						dgv_e1alarms[EventContext->card][EventContext->port - 1].nALARM_MFSYNC = EventContext->data_aux;
					
						ast_log(LOG_NOTICE, "Card %d E1-%d - MFSYNC (%s) \n", EventContext->card+1,EventContext->port, showalarmstatus_desc[EventContext->data_aux]);
						break;
					case ALARM_SYNC:		//sincronismo de quadro
						dgv_e1alarms[EventContext->card][EventContext->port - 1].nALARM_SYNC = EventContext->data_aux;
					
						ast_log(LOG_NOTICE, "Card %d E1-%d - SYNC (%s) \n", EventContext->card+1,EventContext->port, showalarmstatus_desc[EventContext->data_aux]);
						break;
		   	}
		   	break; 

        case EV_CALLERID:
            if (p && p->sig==DG_SIG_FXO)
            {
                dg_GetCallerId(p->port, szTemp);
                
                if (strlen(szTemp)>1) 
                    strncpy(p->cid_num, szTemp+1, AST_MAX_EXTENSION);
                else
                    p->cid_num[0] = '\0';
                
                if (p->verbosity==50)
                {
                    if (strlen(szTemp)>1)
                        ast_verbose(VERBOSE_PREFIX_3 "Dgv/%d: EV_CALLERID CallerID Received - %s (%d)\n",p->port, p->cid_num, strlen(p->cid_num));
                    else
                        ast_verbose(VERBOSE_PREFIX_3 "Dgv/%d: No CallerID information.\n",p->port);
                }
            }
			break;
		case EV_E1CHANGESTATUS:
		
            if (p->verbosity==50)
                ast_log(LOG_NOTICE, "EV_E1CHANGESTATUS (%d) data %x\n",p->port,EventContext->data);

            switch(EventContext->data)
            {
                case C_SEIZURE:
                    //ask for first r2d number
                    if (p->sig==DG_SIG_R2_NDIS_VARIABLE && p->max_digits_rx>0)
                    {
                        /* first allocate channel to avoid double use -AST_STATE_PRERING*/
                        if (!p->reserved) 
                        {
                            if (p->call_state==DG_AST_NONE || p->call_state==DG_AST_WAITFORDIALTONE)
                            {
                                ast_mutex_lock(&p->lock); {
                                    p->call_state=DG_AST_R2_SIGNALLING;
                                    if (p->verbosity==50)
                                        ast_log(LOG_NOTICE, ">> Seizure on port %d\n",p->port);
    
                                    p->reserved=1;	/* reserve channel */
    
                                } ast_mutex_unlock(&p->lock);
                            }
                            dg_SetStartE1RxCount(p->port, 1);
                        }
                        else {
                            if (p->verbosity==50)
                                ast_log(LOG_NOTICE, ">> Seizure on port %d - but it's already reserved - don't send seizure ack!\n",p->port);
                        }
                        
                    }
                    break;
                case C_GROUP_II:
                    
                    if (p->sig==DG_SIG_R2_NDIS_VARIABLE && p->call_state==DG_AST_R2_SIGNALLING)
                    {
                        //send group B
                        if (p->digit_timeout_sched > -1)
                        {
                            ast_sched_del(sched, p->digit_timeout_sched);
                            p->digit_timeout_sched = -1;
                        }
                        if (p->verbosity==50)
                                ast_log(LOG_NOTICE,"E1 Events (%d) : C_GROUP_II - Sending group B\n", p->port);

                        //automatically sending group B
                        dg_R2SendGroupB(p->port, p->group_b_value);

                        ast_mutex_lock(&p->lock); {

                            p->call_state=DG_AST_RINGING;
                            if (p)
                            {
                                if (p->verbosity==50)
                                    ast_log(LOG_NOTICE, "C_GROUP_II : Calling dgv_new port %d using context=%s\n",p->port,p->context);
                                if (dgv_new(p,AST_STATE_RING, p->context,START_PBX,0)!=NULL)
                                {
                                    ast_set_callerid(p->ast_ch, p->cid_num, NULL,NULL);
                                }
                            }
                            else
                            {
                            	if (option_verbose >= 3)
                                	ast_verbose(VERBOSE_PREFIX_3 "C_GROUP_II: No port available!\n");
                                p->call_state=DG_AST_NONE;
                            }
                        } ast_mutex_unlock(&p->lock);							
                    }
                    break;
                case C_ID_RECEIVED:
                    if (p->sig==DG_SIG_R2_NDIS_VARIABLE && p->call_state==DG_AST_R2_SIGNALLING)
                    {
                        ast_mutex_lock(&p->lock); 
                        {                    	
	                        //get callerid and set into asterisk channel
    	                    dg_GetCallerId(p->port,szTemp);
        	                if (strlen(szTemp)>1) 
            	                strncpy(p->cid_num, szTemp+1, AST_MAX_EXTENSION);
                    	    else
	                            p->cid_num[0] = '\0';
	                            
    	                    if (p->verbosity==50)
        	                {            	                
            	                if (strlen(szTemp)>1)
                	                ast_verbose(VERBOSE_PREFIX_3 "Dgv/%d: E1 CallerID Event - %s (%d)\n",p->port, p->cid_num, strlen(p->cid_num));
                    	    }
                        	//ast_set_callerid(p->ast_ch, szTemp, NULL,NULL);
                        	dg_SetNextE1RxCount(p->port, 1);
                        } 
                        ast_mutex_unlock(&p->lock);                        	
                    }
                    break;

                case C_NUMBER_RECEIVED:		/* Get called number */
                    dg_GetE1Number(p->port,szTemp);
                    ndis_size=strlen(szTemp);
                    if (p->verbosity==50)
                        ast_log(LOG_NOTICE,"E1 Events (%d): Partial Number received = %s\n", p->port, szTemp);
                
                    if (!ast_strlen_zero(szTemp))
                    {
                        strncpy(p->rdnis, szTemp, sizeof(p->rdnis));
                        if (p->verbosity==50)
                            ast_log(LOG_NOTICE, "C_NUMBER_RECEIVED: %s\n",p->rdnis);

                        strncpy(p->exten, p->rdnis,sizeof(p->exten) - 1);
                    }
                    else
                       strcpy(p->exten, "s");

                    if (p->sig==DG_SIG_R2_NDIS_VARIABLE)   //not using rd2/mfc automatic
                    {
                        //manual R2DMFC
                        if (ndis_size == p->ask_for_cid_position)
                        {
                            if (p->verbosity==50)
                                ast_log(LOG_NOTICE,"E1 Events (%d): Asking for CID...\n", p->port);
                            dg_R2AskForId(p->port);
                        }
                        else
                        {
                            if (p->verbosity==50)
                                    ast_log(LOG_NOTICE,"E1 Events (%d) : Receiving digit count %d min=%d max=%d...\n", p->port, ndis_size,p->min_digits_rx, p->max_digits_rx);

                            if (ndis_size >= p->min_digits_rx && ndis_size<p->max_digits_rx /*&& !p->check_exten_after_timeout*/)
                            {
                                if (p->verbosity==50)
                                    ast_log(LOG_NOTICE,"E1 Events (%d): Checking extensions %s (%s@%s)...\n", p->port, szTemp, p->exten, p->context);

                                if (ast_matchmore_extension(NULL, p->context, p->exten, 1, p->cid_num))
                                {
                                    if (p->verbosity==50)
                                        ast_log(LOG_NOTICE,"E1 Events (%d): Still has possible extensions. Waiting more digits...\n", p->port);

                                    dg_SetNextE1RxCount(p->port, 1);
                                }
                                else
                                    if (ast_exists_extension(NULL, p->context, p->exten, 1, p->cid_num))
                                    {
                                        //extension found! accept call
                                        if (p->verbosity==50)
                                            ast_log(LOG_NOTICE,"E1 Events (%d): Extension (%s@%s) found!!!.\n", p->port, p->exten, p->context);

                                        //when send groupB you're saying that you receive enough numbers
                                        dg_R2AskForGroupII(p->port, DG_A3_COMPELED);
                                        //...
                                        //turnoff dtmf timer if it exists
                                        // 
                                        if (p->digit_timeout_sched > -1)
                                        {
                                            ast_sched_del(sched, p->digit_timeout_sched);
                                            p->digit_timeout_sched = -1;
                                            
                                        }
                                    }
                                    else
                                    {
                                        if (p->verbosity==50)
                                            ast_log(LOG_NOTICE,"E1 Events (%d): No more possible extensions. Dropping call.\n", p->port);
                                        //turnoff timer
                                        if (p->digit_timeout_sched > -1)
                                        {
                                            ast_sched_del(sched, p->digit_timeout_sched);
                                            p->digit_timeout_sched = -1;
                                        }
                                            

                                        dg_HangUp(p->port);
                                        //....p->reserved=0;
                                        if ((p->sig==DG_SIG_FXO) || (p->sig==DG_SIG_CUSTOMCAS) || (p->sig==DG_SIG_FXS))
                                            p->reserved=0;
                                        else
                                            p->reserved=2;      //wait receiving idle

                                    }
                            }
                            else
                                if (strlen(szTemp)==p->max_digits_rx)
                                {
                                    if (p->verbosity==50)
                                        ast_log(LOG_NOTICE,"E1 Events (%d): Completed Number received = %s\n",  p->port, szTemp);
                                    dg_R2AskForGroupII(p->port, DG_A3_PULSED);
                                }
                                else
                                    dg_SetNextE1RxCount(p->port, 1);
                          }
                        }
                    break;

                case C_E1_IDLE:      //recebimento de idle para liberar o canal
                    if (p->verbosity==50)
                        ast_log(LOG_NOTICE, "C_E1_IDLE (%d): state=%x\n",p->port,p->call_state);

                    break;

                case C_B_ENDCALL:
                    if (p->verbosity==50)
                        ast_log(LOG_NOTICE, "C_B_ENDCALL (%d): state=%x\n",p->port,p->call_state);

                    //Disables Audio
                    //dg_DisableInputBuffer(p->port);
                    dg_PauseInputBuffer(p->port, DG_PAUSE);

                    if (p->call_state==DG_AST_CONVERSATION)
                    {
                        if (p->b_subscriber_hangup) 
                        {

                            if (p->verbosity==50)
                                ast_log(LOG_NOTICE, "C_B_ENDCALL (%d): End call\n",p->port);
                            //end call - send hangup to asterisk
                            //send hangup
                            //.......p->reserved=0;
                            if ((p->sig==DG_SIG_FXO) || (p->sig==DG_SIG_CUSTOMCAS) || (p->sig==DG_SIG_FXS))
                                p->reserved=0;
                            else
                                p->reserved=2;      //wait receiving idle

                            struct ast_frame f = { AST_FRAME_CONTROL, AST_CONTROL_HANGUP, }; 
                            ast_mutex_lock(&p->lock); {
                                QueueFrame(EventContext->port, &f);
                            } ast_mutex_unlock(&p->lock);
                            p->call_state=DG_AST_NONE;
                        }
                        else
                        {
                            
                            //check if there is a bridged channel with
                            if (p->ast_ch)
                            {
                                if ((chan_tmp=ast_bridged_channel(p->ast_ch))!=NULL)
                                {
                                    //tell to bridged channel to stop playbuffer
                                    if (chan_tmp->tech_pvt!=NULL && !strncmp(chan_tmp->name,type,3))
                                    {
                                        p_bridged = (struct dgv_pvt *)chan_tmp->tech_pvt;
                                            if (p->verbosity==50)
                                                ast_log(LOG_NOTICE, "C_B_ENDCALL: put %s quite\n",p_bridged->dev);
                                           dg_StopPlayBuffer(p_bridged->port);
                                    }
                                }
                            }
                        }
                    }
                    break;
                case C_ANSWERED:
                    if (p->call_state==DG_AST_CONVERSATION)
                    {
                        //b subscriber returns after C_B_ENDCALL
                        if (p->verbosity==50)
                            ast_log(LOG_NOTICE, "B Subscriber returns to call\n");
                        //dg_EnableInputBuffer(p->port, DG_DISABLE);
                        dg_PauseInputBuffer(p->port, DG_RELEASE);
                    }
                    break;
			}
			break;
		case EV_R2:
				if (p->verbosity==50)
					ast_verbose(VERBOSE_PREFIX_3 "Port %d \tR2D: %x\n",p->port, EventContext->data);

                // New verbosity to save R2 signalling to file
                if (p->verbosity==50) {
                    sprintf(szTemp,"R2:%x",EventContext->data);
                    save_log(p->port, szTemp);
                }

                /* check if is blocking collect calls */
                if (p->flashing && EventContext->data==R2_IDLE)
                {
                    /* Send hangup */
                    struct ast_frame f = { AST_FRAME_CONTROL, AST_CONTROL_HANGUP, };
                    ast_mutex_lock(&p->lock); {
                        QueueFrame(EventContext->port, &f);
                    } ast_mutex_unlock(&p->lock);
                }

                /* change the reserved flag*/
                if (EventContext->data==R2_IDLE) {
                    if (p->reserved==2)
                    {
                        if (p->verbosity==50)
                            ast_log(LOG_NOTICE, "EV_R2 (%d) - Changing reserved to 0 because receive %x\n",p->port,EventContext->data);
                        p->reserved = 0;
                    }
                }

			break;
        case EV_LINEREADY:        
        	if (p)
        	{
        		if (p->sig == DG_SIG_FXS) 
        		{
					if (p->verbosity == 50)
						ast_log(LOG_NOTICE, "EV_LINEREADY port %d data: %x \n", p->port, EventContext->data);
						        			
        			switch (p->call_state)        		
        			{
        				case DG_AST_NONE:
							//dynamically enable echo cancelation						
    	    			    if (p->echo_can)
			        	    {
            					if (dg_EnableEchoCancelation(p->port, p->echo_can, p->echo_training) != DG_EXIT_SUCCESS)
            					{
		        			    	if (p->immediate)
									{    
										if (p->verbosity==50)
        			  						ast_log(LOG_NOTICE, "EV_LINEREADY (%d): Starting Immediate. \n", p->port);
        			  						
						            	if (option_verbose >= 3)
							                ast_verbose(VERBOSE_PREFIX_3 "Going to immediate on extension s|1 \n");						                
						    	            
						        	    //fxs immediate					        	    
    		        	            	if (dgv_new(p,AST_STATE_RESERVED, p->context, DONT_START_PBX,0)!=NULL)
    	    	    	            	{
    										pthread_attr_init(&attr);
    										pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);    	    	    	            		
	            		    	        	if (ast_pthread_create(&threadid, &attr, fxss_thread, p)>=0) 
    	            	    		        {
        	            	    		    	p->call_state = DG_AST_DIALING;
    											if (option_verbose >= 3)
                	            	    		ast_log(LOG_NOTICE,"fxss_thread Created - port %d \n", p->port);                                	                                    	
		                	            	}
	    		                	        else                                
    	    		                	    	ast_log(LOG_ERROR, "Error creating fxss_thread - port %d \n", p->port);
    	    		                	   	pthread_attr_destroy(&attr);
    		        	            	}
										else
    	    		        	        	ast_log(LOG_ERROR,"Error creating dgv_new to fxs immediate - port %d \n", p->port);
									}
									else// not p->immediate
									{         						
	            			        	ast_log(LOG_ERROR, "Port %d ERROR ENABLING DSP ECHO CANCELATION \n", p->port);
	            			        	
					        		  	if (ast_app_has_voicemail(p->mailbox, NULL))
					        		  	{
											p->stuttertone_thread_running = 1;

        							    	pthread_attr_init(&attr);	
        							   		pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);                
        									if (pthread_create(&p->stuttertone_thread_id, &attr, StutterToneThread, &(p->port)) < 0)
                							{
												p->stuttertone_thread_running = 0;
               								}
               								else
                   								if (p->verbosity==50)
                       								ast_log(LOG_NOTICE, "EV_LINEREADY: stutter dial tone thread created port %d\n",p->port);                       			
               								pthread_attr_destroy(&attr);        		  		
        		  						}        		  		
        		  						else
        		  						{        		  											
			        						dg_GenerateMF(p->port, GENERATE_TONE1, 0);
			        						if (p->verbosity==50)
												ast_log(LOG_NOTICE, "EV_LINEREADY GENERATE_TONE1 on %d \n", p->port);	
        		  						}

    	    							dg_SetDetectionType(p->port, DETECT_DTMF, DG_ENABLE);
       								
										//fxs overlapdial										
    		                    		if (dgv_new(p,AST_STATE_RESERVED, p->context, DONT_START_PBX,0)!=NULL) 
    	    	                		{
											pthread_attr_init(&attr);
											pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);    	    	                			                                
	            	            			if (ast_pthread_create(&threadid, &attr, fxss_thread, p)>=0) 
                	            			{
                    	        				p->call_state = DG_AST_DIALING;
    											if (option_verbose >= 3)
                            	    				ast_log(LOG_NOTICE,"fxss_thread Created - port %d \n", p->port);                                	                                    	
	                            			}
    	                        			else                                
        	                    				ast_log(LOG_ERROR, "Error creating fxss_thread - port %d \n", p->port);
											pthread_attr_destroy(&attr);
            	            			}
										else
	    	                    		{
    		                    			ast_log(LOG_ERROR,"Error creating dgv_new to fxs overlapdial - port %d \n", p->port);
	        		               		}
            						}
            					}
								else
    	    			        {    	    			        	
			        	        	if (p->verbosity == 50)
            				        	ast_log(LOG_NOTICE, "Port %d Enabling(1) DSP Echo Cancelation - taps = %s \n", p->port, taps_desc[p->echo_can]);          				        	

									dg_PickUp(p->port, p->echo_training + 64);//only used as a timer
			                	}
	        			    }
	        			    else
	        			    {     	
	        			    	if (p->immediate)
								{    
									if (p->verbosity==50)
        		  						ast_log(LOG_NOTICE, "EV_LINEREADY (%d): Starting Immediate. \n", p->port);
        		  						
					            	if (option_verbose >= 3)
						                ast_verbose(VERBOSE_PREFIX_3 "Going to immediate on extension s|1 \n");						                
						                
					        	    //fxs immediate
    	        	            	if (dgv_new(p,AST_STATE_RESERVED, p->context, DONT_START_PBX,0)!=NULL)
    	        	            	{
										pthread_attr_init(&attr);
										pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);    	        	            		
	            	    	        	if (ast_pthread_create(&threadid, &attr, fxss_thread, p)>=0) 
    	            	    	        {
        	            	    	    	p->call_state = DG_AST_DIALING;
    										if (option_verbose >= 3)
                	            	    		ast_log(LOG_NOTICE,"fxss_thread Created - port %d \n", p->port);                                	                                    	
		                	            }
    		                	        else                                
        		                	    	ast_log(LOG_ERROR, "Error creating fxss_thread - port %d \n", p->port);
										pthread_attr_destroy(&attr);
    	        	            	}
									else
    	    	        	        	ast_log(LOG_ERROR,"Error creating dgv_new to fxs immediate - port %d \n", p->port);
								}
								else// not p->immediate
								{
									if (p->verbosity==50)
        		  						ast_log(LOG_NOTICE, "EV_LINEREADY (%d): Starting DIAL_TONE. \n", p->port);									
        		  						        			    	
					        		if (ast_app_has_voicemail(p->mailbox, NULL))
					        		{
										p->stuttertone_thread_running = 1;

        							    pthread_attr_init(&attr);	
        							   	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);                
        								if (pthread_create(&p->stuttertone_thread_id, &attr, StutterToneThread, &(p->port)) < 0)
                						{
											p->stuttertone_thread_running = 0;
               							}
               							else
                   							if (p->verbosity==50)
                       							ast_log(LOG_NOTICE, "EV_LINEREADY: stutter dial tone thread created port %d\n",p->port);                       			
               							pthread_attr_destroy(&attr);        		  		
        		  					}        		  		
        		  					else
        		  					{        		  											
			        					dg_GenerateMF(p->port, GENERATE_TONE1, 0);
			        					if (p->verbosity==50)
											ast_log(LOG_NOTICE, "EV_LINEREADY GENERATE_TONE1 on %d \n", p->port);			        					
        		  					}        							        		  						        			    	
        		  						        			    	
	    	    					dg_SetDetectionType(p->port, DETECT_DTMF, DG_ENABLE);
        						
									//fxs overlapdial
    	        	            	if (dgv_new(p,AST_STATE_RESERVED, p->context, DONT_START_PBX,0)!=NULL) 
        	        	        	{
        								pthread_attr_init(&attr);
										pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
	            	    	        	if (ast_pthread_create(&threadid, &attr, fxss_thread, p)>=0) 
    	            	    	        {
        	            	    	    	p->call_state = DG_AST_DIALING;
    										if (option_verbose >= 3)
                	            	    		ast_log(LOG_NOTICE,"fxss_thread Created - port %d \n", p->port);                                	                                    	
		                	            }
    		                	        else                                
        		                	    	ast_log(LOG_ERROR, "Error creating fxss_thread - port %d \n", p->port);
										pthread_attr_destroy(&attr);
	        	    	            }	
									else
    	    	        	        	ast_log(LOG_ERROR,"Error creating dgv_new to fxs overlapdial - port %d \n", p->port);
								}
	        			    }
        	                break;
						case DG_AST_CALLING:					
							if (p->ast_ch)
							{								
								dg_DisableAnswerDetection(p->port);
								dg_DisableCallProgress(p->port);

								//dynamically enable echo cancelation
    	    			    	if (p->echo_can)
			        	    	{
	            					if (dg_EnableEchoCancelation(p->port, p->echo_can, p->echo_training) != DG_EXIT_SUCCESS)
    	        					{
										ast_log(LOG_ERROR, "Port %d ERROR ENABLING DSP ECHO CANCELATION \n", p->port);										    	        						
										p->call_state = DG_AST_CONVERSATION;
										
										struct ast_frame f = { AST_FRAME_CONTROL, AST_CONTROL_ANSWER, };
										ast_mutex_lock(&p->lock); 
										{
											QueueFrame(EventContext->port, &f);
										} 
										ast_mutex_unlock(&p->lock);

										if (p->ast_ch)
											ast_setstate(p->ast_ch, AST_STATE_UP);
        	    					}
									else
    	    				        {    	    			        	
			        		        	if (p->verbosity == 50)
            					        	ast_log(LOG_NOTICE, "Port %d Enabling(2) DSP Echo Cancelation - taps = %s \n", p->port, taps_desc[p->echo_can]);          				        	
	
										dg_PickUp(p->port, p->echo_training + 64);//only used as a timer
			                		}
	        			    	}
	        			    	else
	        			    	{
									if (p->verbosity==50)
        		  						ast_log(LOG_NOTICE, "EV_LINEREADY (%d): ast_setstate(AST_STATE_UP). \n", p->port);
        		  							        			    		
									p->call_state = DG_AST_CONVERSATION;
									
									struct ast_frame f = { AST_FRAME_CONTROL, AST_CONTROL_ANSWER, };
									ast_mutex_lock(&p->lock); 
									{
										QueueFrame(EventContext->port, &f);
									} 
									ast_mutex_unlock(&p->lock);

									if (p->ast_ch)
										ast_setstate(p->ast_ch, AST_STATE_UP);        			    		     			          				    
		        			    } 		

								//enable callprogress during conversation
								ret = dg_EnableCallProgress(p->port, CP_ENABLE_BUSY_OR_FAX);
                				if (ret != DG_EXIT_SUCCESS)
                    				ast_log(LOG_NOTICE, "EnableCallProgress port %d - error %d\n",p->port,ret);
                				else
                    				if (p->verbosity==50)
                        				ast_log(LOG_NOTICE, "EnableCallProgress CP_ENABLE_BUSY_OR_FAX port %d - OK\n",p->port);								
								
								dg_SetDetectionType(p->port, DETECT_DTMF, DG_ENABLE);

			                	ret = dg_PauseInputBuffer(p->port, DG_RELEASE);
								if (ret != DG_EXIT_SUCCESS)
									ast_log(LOG_NOTICE, "Error releasing InputBuffer %d cod = %x - state %x\n",p->port,ret, GetInputBufferStatus(p->port));			
								else
									if (p->verbosity==50)
										ast_log(LOG_NOTICE, "InputBuffer released for port %d cod=%x\n",EventContext->port, ret);		                
							} 	   						
							break;       	              
        			}
        		}
        	}
			break;
         
		case EV_FLASH:
        	if (p)
        	{
        		if (p->sig == DG_SIG_FXS) 
        		{
					if (p->verbosity == 50)
						ast_log(LOG_NOTICE, "EV_FLASH port %d data: %x \n", p->port, EventContext->data);
						
					if (p->call_state == DG_AST_CONVERSATION)
					{
						if (p->verbosity == 50)
							ast_log(LOG_NOTICE, "EV_FLASH port %d - Sending AST_FRAME_CONTROL \n", p->port);
														
						struct ast_frame f = { AST_FRAME_CONTROL, AST_CONTROL_FLASH, };
						ast_mutex_lock(&p->lock); 
						{
							QueueFrame(EventContext->port, &f);
						}
						ast_mutex_unlock(&p->lock);
						
						if (p->verbosity==50)
            				ast_log(LOG_NOTICE, "EV_FLASH: AST_CONTROL_FLASH on port %d \n", p->port);

						if (features_conf.fxs_atxfer[0] != '\0')
						{            
							if (p->verbosity==50)  
								ast_log(LOG_NOTICE, "EV_FLASH: Sending %s to asterisk on port %d \n", features_conf.fxs_atxfer, p->port);						
							
							p->flashing = 1;
										
		                    int nAtxferLen = strlen(features_conf.fxs_atxfer);
							char cAtxfer;
                		    int nForLen;

		                    for (nForLen = 0; nForLen < nAtxferLen; nForLen++)
        		            {
								cAtxfer = features_conf.fxs_atxfer[nForLen];

								if (p->verbosity==50)
									ast_log(LOG_NOTICE, "EV_FLASH: Sending DTMF %d on port %d \n", cAtxfer, p->port);

								struct ast_frame f = { AST_FRAME_DTMF, cAtxfer, };		                        
        		                ast_mutex_lock(&p->lock);
                		        {
									QueueFrame(p->port, &f);
								}
                        		ast_mutex_unlock(&p->lock);                        
                    		}
						}
						else
						{
							if (p->verbosity==50)
								ast_log(LOG_WARNING, "EV_FLASH: FXS transfer command not found!!! \n");
						}
					}
        		}
        	}
			break;            
        case EV_CALLING:
        	if (p)
        	{        
            	if (p->verbosity==50)
                	ast_log(LOG_NOTICE, "EV_CALLING port %d data: %x - state=%x\n",p->port, EventContext->data, p->call_state);        		
        		
				switch (p->sig)
               	{
					case DG_SIG_R2_NDIS_FIXED:
    	            case DG_SIG_R2_NDIS_VARIABLE:
						switch (p->group_b_received)
						{
							case B_FREE_WITHOUTBILLING:							
							case B_NUMBER_UNKNOWN:
							case B_NUMBER_CHANGED:
							case B_OUT_OF_SERVICE:
								break;
							default:
								if (p->call_state != DG_AST_NONE)
								{	
									struct ast_frame f = { AST_FRAME_CONTROL, AST_CONTROL_RINGING, };					
									ast_mutex_lock(&p->lock); {
										QueueFrame(EventContext->port, &f);
									} ast_mutex_unlock(&p->lock);
								}
								break;
						}
            	        break;
					default:
						if (p->call_state != DG_AST_NONE)
						{	
							struct ast_frame f = { AST_FRAME_CONTROL, AST_CONTROL_RINGING, };					
							ast_mutex_lock(&p->lock); {
								QueueFrame(EventContext->port, &f);
							} ast_mutex_unlock(&p->lock);
						}
						break;					
               	}     		
        	}
			break;
		case EV_FAX://fax handling
			if (p->fax_detection)
			{
				if (p->call_state != DG_AST_NONE)
				{
	                switch(p->faxhandled)
    	            {
        	            case 0:
	                        /* first change codecs etc */
    	                    p->faxhandled++;
    
        	                /* disable echo cancelation when receiving fax */
            	            if (p->echo_can)
                	            dg_DisableEchoCancelation(p->port);
	
    	                    p->ast_ch->nativeformats = AST_FORMAT_SLINEAR;
        	                ast_set_read_format(p->ast_ch, AST_FORMAT_SLINEAR);
            	            ast_set_write_format(p->ast_ch, AST_FORMAT_SLINEAR);
	
    	                    //check if there is a bridged channel with
        	                if ((chan_tmp=ast_bridged_channel(p->ast_ch))!=NULL)
            	            {
                	            //tell to bridged channel to stop playbuffer
                    	        if (chan_tmp->tech_pvt!=NULL && !strncmp(chan_tmp->name,type,3))
                        	    {
                            	    p_bridged = (struct dgv_pvt *)chan_tmp->tech_pvt;
                                	if (p->verbosity==50)
	                                  ast_log(LOG_NOTICE, "EV_FAX: changing bridged format for %s \n",chan_tmp->name);

    	                            dg_DisableEchoCancelation(p_bridged->port);
	
    	                            p_bridged->faxhandled++;
	
    	                            //change bridged channel 
        	                        chan_tmp->nativeformats = AST_FORMAT_SLINEAR;
            	                    ast_set_read_format(chan_tmp, AST_FORMAT_SLINEAR);
                	                ast_set_write_format(chan_tmp, AST_FORMAT_SLINEAR);
	
    	                            //re-enable inputbuffer using PCM format
        	                        dg_SetRecordFormat(p_bridged->port, ast2dgvformat(AST_FORMAT_SLINEAR));
	
    	                        }
        	                }
            	            
                	        if (option_verbose >= 3)
                    	    	ast_verbose(VERBOSE_PREFIX_3 "FAX detected port %d\n",p->port);
                        	break;
	                    case 1:
    	                    if (p->verbosity==50)
        	                    ast_log(LOG_NOTICE, "---> EV_FAX Step 2 port %d\n",p->port);
	
    	                    /* first change codecs etc */
        	                p->faxhandled++;
	
    	                    //dg_DisableInputBuffer(p->port);
        	                //........dg_PauseInputBuffer(p->port, DG_PAUSE);
            	            dg_StopPlayBuffer(p->port);
                	        
	
    	                    if (ast_bridged_channel(p->ast_ch)==NULL &&  strncasecmp(p->ast_ch->exten, "fax", 3))
        	                {
	                            snprintf(tmpfax, sizeof(tmpfax), "fax%s", p->ast_ch->exten);
    	                        if (p->verbosity==50)
        	                        ast_verbose(VERBOSE_PREFIX_3 "Checking %s to specific fax extension %s@%s\n", p->ast_ch->name, tmpfax,p->ast_ch->context);
            	                if (ast_exists_extension(p->ast_ch, p->ast_ch->context, tmpfax, 1, p->ast_ch->cid.cid_num))
                	            {
                    	            if (p->verbosity==50)
                        	            ast_verbose(VERBOSE_PREFIX_3 "Redirecting %s to specific fax extension %s\n", p->ast_ch->name, tmpfax);
                            	    /*endif*/
                                	pbx_builtin_setvar_helper(p->ast_ch, "FAXEXTEN", p->ast_ch->exten);
	                                if (ast_async_goto(p->ast_ch, p->ast_ch->context, tmpfax, 1))
    	                            {
        	                            ast_log(LOG_WARNING, "Failed to async goto '%s' into '%s' of '%s'\n", p->ast_ch->name, tmpfax, p->ast_ch->context);
            	                        //p->faxhandled=0;
                	                    //p->ast_ch->nativeformats = p->nativeformats;
                    	            }
                        	        /*endif*/
	                            }
    	                        else
        	                    {
            	                    if (p->verbosity==50)
                	                    ast_verbose(VERBOSE_PREFIX_3 "2nd Checking %s to specific fax extension %s@%s\n", p->ast_ch->name, tmpfax,p->ast_ch->context);
                    	            if (ast_exists_extension(p->ast_ch, p->ast_ch->context, "fax", 1, p->ast_ch->cid.cid_num))
                        	        {
                            	        if (p->verbosity==50)
                                	        ast_verbose(VERBOSE_PREFIX_3 "Redirecting %s to fax extension\n", p->ast_ch->name);
                                    	/*endif*/
	                                    pbx_builtin_setvar_helper(p->ast_ch, "FAXEXTEN", p->ast_ch->exten);
    	                                if (ast_async_goto(p->ast_ch, p->ast_ch->context, "fax", 1))
        	                            {
            	                            if (p->verbosity==50)
                	                            ast_log(LOG_WARNING, "Failed to async goto '%s' into fax of '%s'\n", p->ast_ch->name, p->ast_ch->context);
	
    	                                    //p->faxhandled=0;
        	                                //p->ast_ch->nativeformats = p->nativeformats;
            	                        }
                	                    /*endif*/
                    	            }
                        	        else
                            	    {
                                	    if (p->verbosity==50)
                                    	    ast_log(LOG_NOTICE, "Fax detected, but no fax extension\n");
	
    	                                //p->faxhandled=0;
        	                            //p->ast_ch->nativeformats = p->nativeformats;
            	                    }
                	            }
                    	        /*endif*/
                        	}
	
    	                    //re-enable inputbuffer using PCM format
        	                dg_SetRecordFormat(p->port, ast2dgvformat(AST_FORMAT_SLINEAR));
	
    	                    break;
        	            default:
            	            if (p->verbosity==50)
                	            ast_log(LOG_DEBUG, "Already in a fax extension, not redirecting\n");
	
    	                    break;
        	        }//switch
				}//if (p->call_state != DG_AST_NONE)
			}
			else//if (p->fax_detection)
			{
				if (option_verbose >= 3)
					ast_verbose(VERBOSE_PREFIX_3 "FAX detected(%d), but the FAX option is disabled \n", p->port);				
			}//if (p->fax_detection)	
			
			break;
	  	case EV_AUDIO_SIGNAL:
			if (p->verbosity==50 && p->audio_debug)
				ast_log(LOG_NOTICE, "EV_AUDIO_SIGNAL porta %d dado: %x\n",p->port, EventContext->data);
				
			if (p->last_dtmf!=-1 && EventContext->data==0x20) /* silence*/
			{	
			    p->last_dtmf=-1;

                dg_PauseInputBuffer(p->port, DG_RELEASE);
			}
				
			break;		
		case EV_LINEOFF:
        case EV_BUSY:
        	if (p)
        	{
        		if (EventContext->command == EV_LINEOFF)
        		{
					if (p->verbosity==50)
						ast_log(LOG_NOTICE, "EV_LINEOFF port %d data: %x\n",p->port, EventContext->data);				
        		}		
        		else
        		{	
					if (p->verbosity==50)
                		ast_log(LOG_NOTICE, "EV_BUSY port %d data: %x\n",p->port, EventContext->data);	
        		}
        		
				if ((EventContext->command == EV_BUSY) ||
				   ((EventContext->command == EV_LINEOFF) && (p->sig == DG_SIG_FXS)))
				{			        	
					if (p->call_state != DG_AST_NONE)
					{
						struct ast_frame f = { AST_FRAME_CONTROL, };
						
						switch(p->call_state)
						{
        		            case DG_AST_R2_SIGNALLING:
                	        
                    		    p->call_state = DG_AST_HANGINGUP;

								if (p->digit_timeout_sched > -1)
        		                {
		
        		                    ast_sched_del(sched, p->digit_timeout_sched);
                		            p->digit_timeout_sched = -1;
                        		}
							
								//Disables Audio
								//dg_DisableInputBuffer(p->port);
		                        dg_PauseInputBuffer(p->port, DG_PAUSE);

	                        	p->reserved=0;
	                        	/*if (p->sig==DG_SIG_FXO)
		                            p->reserved=0;
		                        else
	    	                        p->reserved=2;      //wait receiving idle*/

								//send hangup								
								f.subclass = AST_CONTROL_HANGUP;
								ast_mutex_lock(&p->lock); {
									QueueFrame(EventContext->port, &f);
								} ast_mutex_unlock(&p->lock);
								p->call_state=DG_AST_NONE;
								//******************					
								//VERY IMPORTANT!!!
								//Here it is possible that p->ast_ch is
								//already NULL, then be careful!!!!!!
								//******************
			
								break;
							case DG_AST_CONVERSATION:
		                    case DG_AST_RINGING:
								p->call_state = DG_AST_HANGINGUP;
										                 		                        
								if (p->verbosity==50)
									ast_log(LOG_NOTICE, "EV_BUSY port %d - Sending AST_CONTROL_HANGUP \n",p->port);										                 		                                                
		
								//Disables Audio
								//dg_DisableInputBuffer(p->port);
		                        dg_PauseInputBuffer(p->port, DG_PAUSE);
		
		                        p->reserved=0;
		                        
		                        if (p->tone_thread_running == 1)
									p->tone_thread_running = 0;
									
		                        if (p->stuttertone_thread_running == 1)
									p->stuttertone_thread_running = 0;											                        
		                        
								//send hangup								
								f.subclass = AST_CONTROL_HANGUP;
								ast_mutex_lock(&p->lock); {
									QueueFrame(EventContext->port, &f);
								} ast_mutex_unlock(&p->lock);
		
								//******************					
								//VERY IMPORTANT!!!
								//Here it is possible that p->ast_ch is
								//already NULL, then be careful!!!!!!
								//******************
								break;
		                    case DG_AST_HANGINGUP:
        		                //do nothing
                		        break;                		        
		                    case DG_AST_DIALING://only to fxs cards
		                    	if (p->sig == DG_SIG_FXS)
		                    	{		                    	
			                        if (p->verbosity==50)
		               		        	ast_log(LOG_NOTICE, "EV_LINEOFF port %d - Sending AST_CONTROL_HANGUP \n", p->port);
	    	           		        
				                	dg_GenerateMF(p->port, GENERATE_OFF, 0);
       		        
	            	   		        if (p->tone_thread_running == 1)
	               			        	p->tone_thread_running = 0;
	               		        	
		                        	if (p->stuttertone_thread_running == 1)
										p->stuttertone_thread_running = 0;	               		        	
	               		        	
		               				if (p->ast_ch) 
			            				p->ast_ch->hangupcause = AST_CAUSE_UNALLOCATED;

									//send hangup
									f.subclass = AST_CONTROL_HANGUP;
									ast_mutex_lock(&p->lock); {
										QueueFrame(EventContext->port, &f);
									} ast_mutex_unlock(&p->lock);									

									if ((!p->fxss_thread_running) && (p->ast_ch))
									{			
			                        	if (p->verbosity==50)
		               		        		ast_log(LOG_NOTICE, "EV_LINEOFF port %d - Sending ast_hangup \n", p->port);
		               		        		
		               		        	ast_hangup(p->ast_ch);	
									}									

	        		    			p->exten[0] = '\0';       		    			
	        		    			break;
		                    	}           			    	        			
							default:
								if (p->verbosity==50)								
									ast_log(LOG_NOTICE, "EV_BUSY port %d - Sending AST_STATE_BUSY\n",p->port);
									
								//send busy								
								f.subclass = AST_CONTROL_BUSY;
								ast_mutex_lock(&p->lock); {
									QueueFrame(EventContext->port, &f);
								} ast_mutex_unlock(&p->lock);
						
								//******************					
								//VERY IMPORTANT!!!
								//Here it is possible that p->ast_ch is
								//already NULL, then be careful!!!!!!
								break;
						}
					}
					else
					{
						if (p)
						{												
							if (p->sig == DG_SIG_FXS)
							{
								if ((p->tone_thread_running == 1) || (p->stuttertone_thread_running == 1))
								{
									if (p->tone_thread_running == 1)
										p->tone_thread_running = 0;									

			                        if (p->stuttertone_thread_running == 1)
										p->stuttertone_thread_running = 0;
										
			                		dg_GenerateMF(p->port, GENERATE_OFF, 0);
									//p->ast_ch = NULL;									
		                        }//if ((p->tone_thread_running == 1) || (p->stuttertone_thread_running == 1))
							}//if (p->sig == DG_SIG_FXS)
						}//if (p)
					}
					
				}
        	}
			break;
		case EV_DIALTONE:
			//dial tone was received, then dials
			if (p->verbosity==50)
				ast_log(LOG_NOTICE, "EV_DIALTONE port %d\n",p->port);

			if (p)
			{
			    dg_DisableCallProgress(p->port);
				if (p->call_state == DG_AST_WAITFORDIALTONE)
				{
                    ast_mutex_lock(&p->lock); {

	                    if (p->linetone_timeout_sched > -1) //stop line tone timeout
    	                {
        	                ast_sched_del(sched, p->linetone_timeout_sched);
            	            p->linetone_timeout_sched = -1;
                	    }
	
    		        } ast_mutex_unlock(&p->lock); 

                    start_dial(p);                    
				}
                else
                {
                    if (p->verbosity==50)
                        ast_log(LOG_NOTICE, "EV_DIALTONE port %d - another state=%x\n",p->port, p->call_state);
                }
			}
			break;
        case EV_AFTERPICKUP:
        case EV_AFTERFLASH:
        	if (EventContext->command == EV_AFTERPICKUP)
        	{
				if (p->verbosity==50)
					ast_log(LOG_NOTICE, "EV_AFTERPICKUP port %d \n", p->port);
        	}		
        	else
        	{	
				if (p->verbosity==50)
               		ast_log(LOG_NOTICE, "EV_AFTERFLASH port %d \n", p->port);	
        	}        
        
            if ((p->sig == DG_SIG_FXS) && (EventContext->command == EV_AFTERPICKUP))
            {
				switch (p->call_state)
				{
					
					case DG_AST_NONE:			
				        if (p->immediate)
						{    
							if (p->verbosity==50)
        						ast_log(LOG_NOTICE, "EV_AFTERPICKUP (%d): Starting Immediate. \n", p->port);
        			  						
							if (option_verbose >= 3)
								ast_verbose(VERBOSE_PREFIX_3 "Going to immediate on extension s|1 \n");						                
						    	            
							//fxs immediate
    		        		if (dgv_new(p,AST_STATE_RESERVED, p->context, DONT_START_PBX,0)!=NULL)
    	    	    		{
								pthread_attr_init(&attr);
								pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
	            				if (ast_pthread_create(&threadid, &attr, fxss_thread, p)>=0) 
    	            			{
        	            			p->call_state = DG_AST_DIALING;
    								if (option_verbose >= 3)
                	        			ast_log(LOG_NOTICE,"fxss_thread Created - port %d \n", p->port);                                	                                    	
		                		}
	    		        		else                                
    	    		    			ast_log(LOG_ERROR, "Error creating fxss_thread - port %d \n", p->port);
								pthread_attr_destroy(&attr);
    	    	    		}
							else
    	    					ast_log(LOG_ERROR,"Error creating dgv_new to fxs immediate - port %d \n", p->port);
						}
						else// not p->immediate
						{				
							if (p->verbosity==50)
        		  				ast_log(LOG_NOTICE, "EV_AFTERPICKUP (%d): Starting DIAL_TONE. \n", p->port);
		        		  		
		        		  	if (ast_app_has_voicemail(p->mailbox, NULL))
        				  	{
								p->stuttertone_thread_running = 1;

        		    			pthread_attr_init(&attr);	
                				pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);                
                				if (pthread_create(&p->stuttertone_thread_id, &attr, StutterToneThread, &(p->port)) < 0)
                				{
									p->stuttertone_thread_running = 0;
               					}
               					else
                   					if (p->verbosity==50)
                       					ast_log(LOG_NOTICE, "EV_AFTERPICKUP: stutter dial tone thread created port %d\n",p->port);                       			
		               			pthread_attr_destroy(&attr);        		  		
        				  	}        		  		
        				  	else
        				  	{        		  											
			        			dg_GenerateMF(p->port, GENERATE_TONE1, 0);
			        			if (p->verbosity==50)
									ast_log(LOG_NOTICE, "EV_AFTERPICKUP GENERATE_TONE1 on %d \n", p->port);			        			
        				  	}
			        	
							dg_SetDetectionType(p->port, DETECT_DTMF, DG_ENABLE);
        
							//fxs overlapdial
    						if (dgv_new(p,AST_STATE_RESERVED, p->context, DONT_START_PBX,0)!=NULL) 
        					{
								pthread_attr_init(&attr);
								pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
								if (ast_pthread_create(&threadid, &attr, fxss_thread, p)>=0) 
            					{
            						p->call_state = DG_AST_DIALING;
    								if (option_verbose >= 3)
                						ast_log(LOG_NOTICE,"fxss_thread Created - port %d \n", p->port);                                	                                    	
								}
								else                                
        							ast_log(LOG_ERROR, "Error creating fxss_thread - port %d \n", p->port);
								pthread_attr_destroy(&attr);
							}
							else
	    					{
	    						ast_log(LOG_ERROR,"Error creating dgv_new to fxs overlapdial - port %d \n", p->port);
							}
						}
						break;
					case DG_AST_CALLING:
						if (p->verbosity==50)
        		  			ast_log(LOG_NOTICE, "EV_AFTERPICKUP (%d): ast_setstate(AST_STATE_UP). \n", p->port);
        		  										
						p->call_state = DG_AST_CONVERSATION;
						
						struct ast_frame f = { AST_FRAME_CONTROL, AST_CONTROL_ANSWER, };						
						ast_mutex_lock(&p->lock); 
						{
							QueueFrame(p->port, &f);
						}
						ast_mutex_unlock(&p->lock);
		
						if (p->ast_ch)
							ast_setstate(p->ast_ch, AST_STATE_UP);		
		
						break;	
				}
			}
			else//if ((p->sig == DG_SIG_FXS) && (EventContext->command == EV_AFTERPICKUP))
			{
				//AFTERPICKUP or AFTERFLASH is used only to incoming calls	
				if (p->call_state == DG_AST_INCOMING_CALL)
				{
					if (p->flashing)
					{
						p->flashing = 0;

	           			if (option_verbose >= 3)
		    	           	ast_log(LOG_NOTICE, "EV_AFTERPICKUP: Calling CallBlock application on %d \n", p->port);

						dg_Flash(p->port, 1, 1000, 100);
						break;
					}
					
        	    	switch(p->sig)
            	    {
                		case DG_SIG_FXO:
			    		    if (p->ring_timeout_sched > -1)//stop ring timeout
	    	    		    {
    		    		    	if (p->verbosity==50)
        	                		ast_log(LOG_NOTICE, "EV_AFTERPICKUP port %d - deleting ring schedule \n", p->port);

        		        	    ast_sched_del(sched, p->ring_timeout_sched);
            		            p->ring_timeout_sched = -1;
                		    }

	                    	break;
		                default:
    						break;
        		   	}
				
					ast_setstate(p->ast_ch, AST_STATE_UP);
					//call progress to detect busy, etc
					/* todo: configuration to this */
					dg_SetDetectionType(p->port, dgv_conf.default_detectiontype, DG_ENABLE);
			
					//enable inputbuffer streaming with agc disabled
                	dg_PauseInputBuffer(p->port, DG_RELEASE);

                	ret = dg_EnableCallProgress(p->port, CP_ENABLE_BUSY_OR_FAX);
                	if (ret != DG_EXIT_SUCCESS)
	                    ast_log(LOG_NOTICE, "EnableCallProgress port %d - error %d\n",p->port,ret);
    	            else
        	            if (p->verbosity==50)
            	            ast_log(LOG_NOTICE, "EnableCallProgress CP_ENABLE_BUSY_OR_FAX port %d - OK\n",p->port);
	
    	            //dynamically enable echo cancelation
        	        //p->echo_can must contains the number os taps
            	    if (p->echo_can)
                	{
	                    //dg_DisableEchoCancelation(p->port);
    	                if (dg_EnableEchoCancelation(p->port,p->echo_can,p->echo_training) != DG_EXIT_SUCCESS)
        	                ast_log(LOG_ERROR, "Port %d ERROR ENABLING DSP ECHO CANCELATION\n", p->port);
            	        else
                	    {
                    	    if (p->verbosity==50)
                        	    ast_log(LOG_NOTICE, "Port %d Enabling(3) DSP Echo Cancelation - taps = %s\n", p->port,taps_desc[p->echo_can] );
                          	
	                        //only echotraining     
							if (p->all_echotraining_sched > -1) 
							{
           						ast_sched_del(sched, p->all_echotraining_sched);
           						p->all_echotraining_sched = -1;
   	   						}
							p->all_echotraining_sched = ast_sched_add(sched, p->echo_training + 5000, all_echotraining_callback, p);
							if (p->all_echotraining_sched < 0)
								ast_log(LOG_NOTICE, "Port %d ERROR Setting timeout to all_echotraining(2) \n", EventContext->port);
        	            }
            	    }
					p->call_state = DG_AST_CONVERSATION;
				}//if (p->call_state == DG_AST_INCOMING_CALL)
			}//if ((p->sig == DG_SIG_FXS) && (EventContext->command == EV_AFTERPICKUP))
			break;
			
		case EV_GROUP_B:
			if (p)
			{
				p->group_b_received = EventContext->data;
				
				if (p->verbosity==50)
					ast_log(LOG_NOTICE, "EV_GROUP_B port %d, group_b_received %d \n", EventContext->port, p->group_b_received);
			}			
			break;			
		case EV_AFTERDIAL:
			if (p)
			{
				if (p->verbosity==50)
					ast_log(LOG_NOTICE, "EV_AFTERDIAL %d\n",EventContext->port);

				if (p->sig == DG_SIG_FXS) 
        		{        										
        			if (p->call_state == DG_AST_CALLING)
        			{
		       			dg_SetRing(p->port, DG_ENABLE, p->ring_type);

						struct ast_frame f = { AST_FRAME_CONTROL, AST_CONTROL_RINGING, };						
						ast_mutex_lock(&p->lock); 
						{
							QueueFrame(p->port, &f);
						}
						ast_mutex_unlock(&p->lock);						       				
        					
	           			if (option_verbose >= 3)        				
    	           			ast_log(LOG_NOTICE, "EV_AFTERDIAL: Ring started on %d \n", p->port);                   		
                    }
        		}
										
				if (p->call_state==DG_AST_DIALING)
				{
					//enable callprogress to wait, busy, calling, answer, etc...
               		switch(p->sig)
               		{
	                   	case DG_SIG_R2_NDIS_FIXED:
    	               	case DG_SIG_R2_NDIS_VARIABLE:						    
							switch (p->group_b_received)
							{
								case B_FREE_CALLING:
								case B_BUSY:
								case B_CONGESTION:
								case B_COLLECTCALL:
						    		p->call_state = DG_AST_CALLING;
						    		ast_setstate(p->ast_ch, AST_STATE_RINGING);								
									break;
								case B_FREE_WITHOUTBILLING:									
								case B_NUMBER_UNKNOWN:
								case B_NUMBER_CHANGED:
								case B_OUT_OF_SERVICE:								
									if (p->verbosity==50)
										ast_log(LOG_NOTICE, "Sending AST_CONTROL_ANSWER, group_b_received on %d\n", EventContext->port);
									
									//check if there is a bridged channel with
									if (p->ast_ch)
									{							
										p->call_state = DG_AST_CONVERSATION;
					
										struct ast_frame f = { AST_FRAME_CONTROL, AST_CONTROL_ANSWER, };					
										ast_mutex_lock(&p->lock); 
										{
											QueueFrame(EventContext->port, &f);
										} 
										ast_mutex_unlock(&p->lock);

										if (p->ast_ch)
										{
											ast_setstate(p->ast_ch, AST_STATE_UP);
										}
		
					                	ret = dg_PauseInputBuffer(p->port, DG_RELEASE);
										if (ret != DG_EXIT_SUCCESS)
											ast_log(LOG_NOTICE, "Error releasing InputBuffer %d cod = %x - state %x\n",p->port,ret,
                                					GetInputBufferStatus(p->port));			
										else
											if (p->verbosity==50)
												ast_log(LOG_NOTICE, "InputBuffer released for port %d cod=%x\n",EventContext->port, ret);
									}
									break;
								default:
						    		p->call_state = DG_AST_CALLING;
						    		ast_setstate(p->ast_ch, AST_STATE_RINGING);								
									break;
							}						    
            	            break;
           	    	    case DG_SIG_PRI:
						    p->call_state = DG_AST_CALLING;
					    	ast_setstate(p->ast_ch, AST_STATE_RINGING);
	                       	break;
		                case DG_SIG_CUSTOMCAS:
    	                case DG_SIG_FXO:
	    				    ret = dg_EnableCallProgress(p->port,CP_ENABLE_ALL);
						    if (ret==EXIT_SUCCESS)	
						    {	
								p->call_state = DG_AST_CALLING;
								ast_setstate(p->ast_ch, AST_STATE_RINGING);
					    	}
						    else
						    {
						    	ast_log(LOG_ERROR, "ERROR FXO Enabling CallProgress on %d \n", p->port);
								ast_setstate(p->ast_ch, AST_STATE_BUSY);							
						    }	
					    
							//enables answerdetection							
		   					ret = dg_EnableAnswerDetection(p->port);
		   					if (ret!=DG_EXIT_SUCCESS)
		   					{
		   						ast_log(LOG_ERROR, "ERROR FXO Enabling AnswerDetection on %d \n", p->port);
	    					    ast_setstate(p->ast_ch, AST_STATE_BUSY);
	    					}
	                       	break;
	       				default:
	        				break;
    	   	        }
				}	
			}
			break;
		case EV_ANSWERED:
			if (p->call_state==DG_AST_CALLING)
			{
				if (p->verbosity==50)
					ast_log(LOG_NOTICE, "EV_ANSWERED %d\n",EventContext->port);
									
				//check if there is a bridged channel with
				if (p->ast_ch)
				{							
					dg_DisableAnswerDetection(p->port);
					dg_DisableCallProgress(p->port);

					p->call_state = DG_AST_CONVERSATION;
					
					struct ast_frame f = { AST_FRAME_CONTROL, AST_CONTROL_ANSWER, };					
					ast_mutex_lock(&p->lock); {
						QueueFrame(EventContext->port, &f);
					} ast_mutex_unlock(&p->lock);

					if (p->ast_ch)
					{
						ast_setstate(p->ast_ch, AST_STATE_UP);
					}
		
					//enable callprogress during conversation
					ret = dg_EnableCallProgress(p->port, CP_ENABLE_BUSY_OR_FAX);
                	if (ret != DG_EXIT_SUCCESS)
                    	ast_log(LOG_NOTICE, "EnableCallProgress port %d - error %d\n",p->port,ret);
					else
                    	if (p->verbosity==50)
                        	ast_log(LOG_NOTICE, "EnableCallProgress CP_ENABLE_BUSY_OR_FAX port %d - OK\n",p->port);			

					dg_SetDetectionType(p->port,DETECT_DTMF,DG_ENABLE);

					//enable inputbuffer streaming with agc disabled
					//ret = dg_EnableInputBuffer(p->port,DG_DISABLE);
                	ret = dg_PauseInputBuffer(p->port, DG_RELEASE);
					if (ret != DG_EXIT_SUCCESS)
						ast_log(LOG_NOTICE, "Error releasing InputBuffer %d cod = %x - state %x\n",p->port,ret,
                                GetInputBufferStatus(p->port));			
					else
						if (p->verbosity==50)
							ast_log(LOG_NOTICE, "InputBuffer released for port %d cod=%x\n",EventContext->port, ret);
	
	                //dynamically enable echo cancelation
    	            if (p->echo_can)
        	        {
            	        //dg_DisableEchoCancelation(p->port);
                	    if (dg_EnableEchoCancelation(p->port,p->echo_can,p->echo_training)!=DG_EXIT_SUCCESS)
                    	    ast_log(LOG_ERROR, "Port %d ERROR ENABLING DSP ECHO CANCELATION\n", p->port);
	                    else
    	                {
        	                if (p->verbosity==50)
            	                ast_log(LOG_NOTICE, "Port %d Enabling(4) DSP Echo Cancelation - taps = %s\n", p->port,taps_desc[p->echo_can] );
            	                
                        	//only echotraining
							if (p->all_echotraining_sched > -1) 
							{
             					ast_sched_del(sched, p->all_echotraining_sched);
   	        					p->all_echotraining_sched = -1;
    						}
							p->all_echotraining_sched = ast_sched_add(sched, p->echo_training + 5000, all_echotraining_callback, p);
							if (p->all_echotraining_sched < 0)
								ast_log(LOG_NOTICE, "Port %d ERROR Setting timeout to all_echotraining(1) \n", EventContext->port);
                	    }
	                }
				}
			}
			break;
        case EV_SILENCE:
            if(EventContext->data==1)
            {
                if (p->verbosity==50)
                    ast_verbose(VERBOSE_PREFIX_3 "SILENCE on port %d - data=%x\n",EventContext->port,EventContext->data);

                //check if there is a bridged channel with, removed on 0.9.3
                /*if (p->ast_ch)
                {                	
                    if ((chan_tmp=ast_bridged_channel(p->ast_ch))!=NULL)
                    {
                        //tell to bridged channel to stop playbuffer                                     
                        if (chan_tmp->tech_pvt!=NULL && !strncmp(chan_tmp->name, type, 3))                      
                        {      
                            p_bridged = (struct dgv_pvt *)chan_tmp->tech_pvt;
                            
                            if (p->verbosity==50)
                                 ast_log(LOG_NOTICE, "(%d) SILENCE Detected: tells to %s drop tx samples\n",p->port, p_bridged->dev);
   
   							//drop tx samples
							p_bridged->force_drop_tx = 1;						
                        }
                    }
                }*/
            }
            else
            {
                if (p->verbosity==50)
                    ast_verbose(VERBOSE_PREFIX_3 "AUDIO on port %d - data=%x\n",EventContext->port,EventContext->data);
            }
            break;
        case EV_RINGS:
            if (option_verbose >= 3)
                ast_verbose(VERBOSE_PREFIX_3 "Ringing on port %d - callstate=%x\n",EventContext->port, p->call_state);

            switch(p->sig)
            {            			
            	case DG_SIG_FXO:                	
			    	if (p->ring_timeout_sched > -1)//stop ring timeout
    	    		{
    	    			if (p->verbosity==50)
                			ast_log(LOG_NOTICE, "EV_RINGS port %d - deleting ring schedule \n", p->port);    	    		            	
    	    		            	
        	        	ast_sched_del(sched, p->ring_timeout_sched);
            	    	p->ring_timeout_sched = -1;
                	}
               			
        	    	//set ring timeout
    	    		if (p->verbosity==50)
                		ast_log(LOG_NOTICE, "EV_RINGS port %d - Creating ring schedule \n", p->port);        	        			
        	        			
		        	p->ring_timeout_sched = ast_sched_add(sched, ringtimeout, ring_timeout_callback, p);
                	break;
                case DG_SIG_FXS:
                	//TODO: ring timeout                	
                	break;
	            default:
    				break;
			}

			switch (p->call_state) {
                case DG_AST_R2_SIGNALLING:
                    ast_mutex_lock(&p->lock); {
                        p->reserved = 1;
                    } ast_mutex_unlock(&p->lock);    

					ast_setstate(p->ast_ch, AST_STATE_RING);
					p->call_state=DG_AST_RINGING;
					break;
				case DG_AST_NONE:
					//turnoff dtmf timer if it exists
					if (p->digit_timeout_sched > -1)
                    {
                        ast_sched_del(sched, p->digit_timeout_sched);
                        p->digit_timeout_sched = -1;
                    }			

					ast_mutex_lock(&p->lock); {                   
        	    								
            			if (option_verbose >= 3)        	    								
							ast_verbose(VERBOSE_PREFIX_3 "Answering port %d\n",EventContext->port);
							
						p->call_state=DG_AST_RINGING;
						if (p)
						{
							if (p->verbosity==50)
								ast_log(LOG_NOTICE, "Calling dgv_new port %d using context=%s\n",p->port,p->context);
								
							if (dgv_new(p,AST_STATE_RING, p->context,START_PBX,0)!=NULL)
							{
								//get callerid and set into asterisk channel
                                if (p->sig!=DG_SIG_FXO && p->sig!=DG_SIG_PRI) 
                                {                                    
                                    dg_GetCallerId(p->port,szTemp);
                                    if (strlen(szTemp)>1) 
                                        strncpy(p->cid_num, szTemp+1, AST_MAX_EXTENSION);
                                    else
                                        p->cid_num[0] = '\0';

	                                ast_set_callerid(p->ast_ch, p->cid_num, NULL,NULL);

    	                            if (p->verbosity==50)
        	                        {
            	                        if (strlen(szTemp)>1)
                	                        ast_verbose(VERBOSE_PREFIX_3 "Dgv/%d: CallerID Received - %s (%d)\n",p->port, p->cid_num, strlen(p->cid_num));
                    	                else
                        	                ast_verbose(VERBOSE_PREFIX_3 "Dgv/%d: No CallerID information.\n",p->port);
                            		}
								}
								else
								{
									if (p->sig == DG_SIG_FXO)
									{
                        				if (p->ast_ch)
                        				{
			                				if (p->verbosity == 50)
            			        				ast_log(LOG_NOTICE, "EV_RINGS: DG_SIG_FXO so ast_set_callerid on port %d \n", p->port);
                    		
											ast_set_callerid(p->ast_ch, p->cid_num, NULL, NULL);
                        				}									
									}									
									
									if (p->sig == DG_SIG_PRI)
									{
                        				if (p->ast_ch)
                        				{
			                				if (p->verbosity == 30 || p->verbosity == 50)
            			        				ast_log(LOG_NOTICE, "EV_RINGS: !ast_matchmore_extension on DG_SIG_PRI so ast_set_callerid on port %d \n", p->port);
                    		
											ast_set_callerid(p->ast_ch, p->cid_num, p->cid_name, NULL);
                        				}									
									}							
								}								
							}							
						}
						else
						{
							if (option_verbose >= 3)
								ast_verbose(VERBOSE_PREFIX_3 "EV_RINGS: No port available!\n");
								
							p->call_state=DG_AST_NONE;
						}
					} ast_mutex_unlock(&p->lock);
					break;
			 }
           break;
        case EV_DTMF:
            cAux = EventContext->data;
            
			if (p->verbosity==50)
			{	
				ast_log(LOG_NOTICE, "Port %d Digit %c\n", EventContext->port, cAux);
			}

            if (p->verbosity==50) {                
                sprintf(szTemp,"MF: %c",cAux);
                save_log(p->port, szTemp);
            }

			if (p->sig==DG_SIG_R2_NDIS_VARIABLE && p->call_state==DG_AST_R2_SIGNALLING)
			{
				//saves last digit received time
				p->tlastdigit  = ast_tvnow();
				if (p->digit_timeout_sched > -1) {
                    ast_sched_del(sched, p->digit_timeout_sched);
                    p->digit_timeout_sched = -1;
                }
					
				p->digit_timeout_sched = ast_sched_add(sched, p->interdigit_timeout, digit_timeout_callback, p);

				if (p->digit_timeout_sched<0)
					ast_log(LOG_NOTICE, "Port %d ERROR Setting timeout\n",EventContext->port);

			}
            ast_sched_runq(sched);

			if (p)
			{
				if (p->sig == DG_SIG_FXS && p->call_state == DG_AST_DIALING)
				{
					dg_GenerateMF(p->port, GENERATE_OFF, 0);
					
					struct ast_frame f = { AST_FRAME_DTMF, EventContext->data, };					
			    	ast_mutex_lock(&p->lock); 
					{
						QueueFrame(EventContext->port, &f);
					}
					ast_mutex_unlock(&p->lock);
               }
			}
			
			/* Find the port to get pvt structure */
			if (p) 
			{
				if (p->ast_ch!=NULL)
				{
					//transfer DTMF only if it isnt in conversation and not bridged
					if (p->ast_ch->_state == AST_STATE_UP) // && ast_bridged_channel(p->ast_ch)==NULL)
					{
                        //comentado na v0.7.4 - isso estava causando paradas no audio durante a conversacao
						//dg_DisableInputBuffer(p->port);
                        //dg_PauseInputBuffer(p->port, DG_PAUSE);
						if (p->verbosity==50)
							ast_log(LOG_NOTICE, "Detecting DTMF to port %d - data %x\n",EventContext->port,cAux);

						if ( (p->dtmf_to_asterisk==1) || 
						     (p->dtmf_to_asterisk==2 && ast_bridged_channel(p->ast_ch)==NULL) ||
						     (p->flashing==1 && p->sig==DG_SIG_FXS) )
						{
							struct ast_frame f = { AST_FRAME_DTMF, EventContext->data, };					
			    			ast_mutex_lock(&p->lock); 
			    			{
								QueueFrame(EventContext->port, &f);
						    }
						    ast_mutex_unlock(&p->lock);
						}	
						/* stores dtmf to generate on EV_AUDIODETECTED event */ 
						p->last_dtmf = EventContext->data;
					}
					else
					{
					    p->last_dtmf = -1;
					}
					
					if (p->call_state == DG_AST_CONVERSATION && ast_bridged_channel(p->ast_ch)==NULL)
					{
                        if (p->verbosity==50)
                            ast_log(LOG_NOTICE, "Port %d Digit %c - stopplaybuffer\n",EventContext->port, cAux);

						dg_StopPlayBuffer(p->port);	//avoiding noise
					}
				}
			}
			else {
				if (option_verbose >= 3)
					ast_log(LOG_WARNING, "Unable to locate port %d structure!\n", EventContext->port);

			}
			break;
	}
}

//-------------------------------------------------------------------------------------------
// Locate channel passing digivoice port
//-------------------------------------------------------------------------------------------
static struct dgv_pvt *find_ch_port(short port)
{
	struct dgv_pvt *tmp;

	if (!iflast)
		return NULL;

	ast_mutex_lock(&iflock); {
		tmp = iffirst; /* beginning of list */
		while(tmp)
		{
				//specific port
				if (tmp->port == port)
					break;
			tmp = tmp->next;			
		}
	} ast_mutex_unlock(&iflock);
	return tmp;
}

//-------------------------------------------------------------------------------------------
// Find avaiable port on digivoice board
//-------------------------------------------------------------------------------------------
static struct dgv_pvt *find_free_port(short group, short backwards, short dc)
{
	struct dgv_pvt *tmp;

	if (!iflast)
		return NULL;

    if (backwards)
        tmp = iflast; /* get top of list */
    else
        tmp = iffirst; /* beginning of list */
        
    while(tmp)
    {
        /* check free port */
        if (tmp->ast_ch==NULL && !tmp->reserved)	//not reserved
        {
            if (dc<0) {
                //default search, without libpri
                if (group>0)
                {
                    //group
                    if (tmp->group == group)
                    {
                        if (tmp->verbosity==20)
                            ast_log(LOG_NOTICE, "Channel allocated port=%d group=%d\n",tmp->port, tmp->group);	
                        break;
                    }
                }
                else
                {
                    //any port
                    if (tmp->verbosity==20)
                        ast_log(LOG_NOTICE, "Channel allocated port=%d\n",tmp->port);	
                    break; 
                }
            }
            else
            {
#ifdef USE_LIBPRI
	                //libpri specific search
    	            if (dchans[dc].card == tmp->board)
        	        {
            	        if (dchans[dc].span == 1 && tmp->ch_board<=30)
                	        break;
                    	else
                        	if (dchans[dc].span == 2 && tmp->ch_board>30)
                            	break;

	                }
#endif//#ifdef USE_LIBPRI                
            }
        }
        else
        {
            if (tmp->verbosity==20)
                ast_log(LOG_NOTICE, "Port=%d already reserved\n",tmp->port);	
        }

        if (backwards)
            tmp = tmp->previous;
        else
            tmp = tmp->next;			
    }
	return tmp;
}

//-------------------------------------------------------------------------------------------
// Create private structure for each channel
//-------------------------------------------------------------------------------------------
static struct dgv_pvt *initialize_ports(short port)
{
	struct dgv_pvt *tmp;

	tmp = (struct dgv_pvt *)calloc(1, sizeof *tmp);

	if (!tmp)
		return NULL;

	ast_mutex_lock(&iflock); {
		tmp->board = dg_GetCardNumber(port);
		tmp->ch_board = dg_GetRelativeChannelNumber(port);
		tmp->group = -1;	//default value - no group
		tmp->port = port;
		tmp->write_last_format=-1;		//initializes with invalid format
		tmp->read_last_format=-1;
		tmp->ast_ch = NULL;
		tmp->tone_thread_id = 0;			//control thread to ringback and busyback tones
		tmp->stuttertone_thread_id = 0;   	//control thread to stutter dial tone 
		tmp->samples_rx_count = 0;		    //counter for received samples

        tmp->force_drop_rx = 0;
        tmp->force_drop_tx = 0;

		//Assume default values to avoid null strings and unsetted values 
		tmp->category = 1;
		tmp->max_digits_rx = 4;
		tmp->tx_gain = -6;
		tmp->rx_gain = 0;
		tmp->ask_for_cid_position = 2;
		tmp->echo_can = 2;  //64taps
        tmp->echo_training = 800;
        tmp->drop_tx_samples_factor = TX_CUT_FACTOR;
        tmp->relaxdtmf=1;
        tmp->gen_ringback = 1;
        tmp->gen_busy = 1;
        tmp->gen_dtmf = 2;
        tmp->dtmf_to_asterisk = 2;
        tmp->b_subscriber_hangup=0;
        tmp->group_b_value = B_FREE_CALLING;
        tmp->silence_threshold = -18;
        tmp->dialtone_timeout = 3000;
        tmp->wait_dialtone = 1;
        tmp->fax_detection = 1;
        
        //fxs
        tmp->ring_type           = 1;
        tmp->flash_min_time      = 80;
        tmp->flash_max_time      = 650;
        tmp->immediate           = 0;//fxs and pri channels
        tmp->fxss_thread_running = 0;
        tmp->fv_thread_timeout   = tmp->dialtone_timeout + 1000;
        strcpy(tmp->mailbox, "1234\0"); 

        strlcpy(tmp->callprogress, dgv_conf.szDefaultCallProgressFile, TAMSTR);

		strcpy(tmp->szPortID,"1122334455\0");	
		strcpy(tmp->context,"default\0");
		strcpy(tmp->accountcode,"none\0");
		
		//gain
		dg_SetPortGain(tmp->port, TX_GAIN, tmp->tx_gain);
		//ast_log(LOG_NOTICE, "9 port %d\n", p->port);
		dg_SetPortGain(tmp->port, RX_GAIN, tmp->rx_gain);
		//ast_log(LOG_NOTICE, "10 port %d\n", p->port);
		
		//E1 specific configuration
		dg_SetPortId(tmp->port, tmp->szPortID);

		//mutexes
		ast_mutex_init(&tmp->lock);
	  	ast_mutex_init(&tmp->play_lock);
		ast_mutex_init(&tmp->record_lock);

	} ast_mutex_unlock(&iflock);

	return tmp;
}

//-------------------------------------------------------------------------------------------
// Try to alloc Asterisk channel to desired port
//-------------------------------------------------------------------------------------------
static struct ast_channel *dgv_new(struct dgv_pvt *i, int state, char *context, short startpbx, int transfercapability)
{
	struct ast_channel *tmp;
    //int flags;

#ifdef CC_AST_HAS_EXT_CHAN_ALLOC
    tmp = ast_channel_alloc(1, state, i->cid_num, NULL,
#ifdef CC_AST_HAS_EXT2_CHAN_ALLOC
		i->accountcode, i->dnid, i->context, i->amaflags,
#endif//#ifdef CC_AST_HAS_EXT2_CHAN_ALLOC
	"%s/%d", type, i->port);
#else//#ifdef CC_AST_HAS_EXT_CHAN_ALLOC
	tmp = ast_channel_alloc(1);
#endif//#ifdef CC_AST_HAS_EXT_CHAN_ALLOC

    if (tmp) 
    {        
        tmp->tech = &dgv_tech;
#ifndef CC_AST_HAS_VERSION_1_4
        	snprintf(tmp->name, sizeof(tmp->name), "%s/%d", type, i->port);      
#endif//#ifndef CC_AST_HAS_VERSION_1_4
        
        //..asterisk new versions ast trunk ast_string_field_build(tmp, tmp->name, "%s/%d", type, i->port);
        if (i->verbosity==10)
            ast_log(LOG_NOTICE, "Trying to allocate %s\n",tmp->name); 
   
        /* lock physical port storing channel pointer into pvt structure */
        i->ast_ch = tmp;
        tmp->tech_pvt = i;        
        ast_copy_string(i->dev, tmp->name, sizeof(i->dev));//strcpy(i->dev, tmp->name);

        if (i->verbosity==10)
            ast_log(LOG_NOTICE, ">>>>>>>>dgv_new %s - address=%x\n",tmp->name, (int)tmp);

        //copy account code if it is configured
        if (!ast_strlen_zero(i->accountcode))
        {
#ifdef CC_AST_HAS_STRINGFIELD_IN_CHANNEL
        	if (i->verbosity==10)
           		ast_log(LOG_NOTICE, ">>>>>>>>dgv_new %s - has stringfield \n",tmp->name);
            ast_string_field_set(tmp, accountcode, i->accountcode);
#else//#ifdef CC_AST_HAS_STRINGFIELD_IN_CHANNEL
            ast_copy_string(tmp->accountcode, i->accountcode, sizeof(tmp->accountcode));
#endif//#ifdef CC_AST_HAS_STRINGFIELD_IN_CHANNEL
        }

        //copy language structure to channel
        if (!ast_strlen_zero(i->language))
        {
#ifdef CC_AST_HAS_STRINGFIELD_IN_CHANNEL
            ast_string_field_set(tmp, language, i->language);
#else//#ifdef CC_AST_HAS_STRINGFIELD_IN_CHANNEL
            ast_copy_string(tmp->language, i->language, sizeof(tmp->language)-1);            
#endif//#ifdef CC_AST_HAS_STRINGFIELD_IN_CHANNEL
        }
        /* todo: configurable formats */
        tmp->nativeformats = supported_fmts;	/* all native formats supported */
	
    	/* save native formats into pvt structure */
    	i->nativeformats = supported_fmts;

        i->faxhandled = 0;		                 //reset fax flag
        i->fxo_bridged_dial = 0;                 //reset fxo bridged dial flag
        if (i->verbosity==10)
            ast_log(LOG_NOTICE, ">>>>>>>>dgv_new %s - Reseting faxhandled and fxo_bridged_dial \n",tmp->name);

        //preferencial format 
        //todo: make custom order in digivoice.conf
       
        if ((supported_fmts & AST_FORMAT_GSM)!=0)        
            i->pref_format = AST_FORMAT_GSM;                      
        else            
            if ((supported_fmts & AST_FORMAT_ULAW)!=0)
                i->pref_format = AST_FORMAT_ULAW;                
            else            
                if ((supported_fmts & AST_FORMAT_ALAW)!=0)
                    i->pref_format = AST_FORMAT_ALAW;                   
                else
                    i->pref_format = AST_FORMAT_SLINEAR;                 

        if (i->verbosity==10)
            ast_log(LOG_NOTICE, "dgv_new %s - preferencial format is %x\n",tmp->name, i->pref_format);
#ifndef CC_AST_HAS_VERSION_1_4
        	tmp->type = type;				    
#endif//#ifndef CC_AST_HAS_VERSION_1_4
        tmp->writeformat = i->pref_format; /* requested write format (playback) */
        tmp->readformat = i->pref_format;  /* requested read format (record) */
        tmp->rawwriteformat = i->pref_format;
        tmp->rawreadformat = i->pref_format;

        tmp->transfercapability = transfercapability;        

        //reset internal buffer to hold samples
        memset(i->buf, 0, sizeof(i->buf));


        /* copy music class for this channel */
        if (!ast_strlen_zero(i->musicclass))
        {
#ifdef CC_AST_HAS_STRINGFIELD_IN_CHANNEL
            ast_string_field_set(tmp, musicclass, i->musicclass);                                                                     
#else//#ifdef CC_AST_HAS_STRINGFIELD_IN_CHANNEL
            ast_copy_string(tmp->musicclass, i->musicclass, sizeof(tmp->musicclass));
#endif//#ifdef CC_AST_HAS_STRINGFIELD_IN_CHANNEL                     
        }    

        //under development
        if (i->pickupgroup!=0)
        {
            /* Only FXO signalled stuff can be picked up */
            tmp->pickupgroup = i->pickupgroup;
        }

        i->read_last_format = -1;

        //set context to use
        strncpy(tmp->context, context, sizeof(tmp->context)-1);


        /* Copy DNIS to channel structure */
        if (!ast_strlen_zero(i->exten))
            strncpy(tmp->exten, i->exten, sizeof(tmp->exten) - 1);
    
        ast_setstate(tmp, state);

#ifdef USE_LIBPRI
	        i->alreadyhungup = 0;
    	    i->stripmsd = 1;            //todo: customizable
#endif//#ifdef USE_LIBPRI


        //ast_mutex_lock(&usecnt_lock);
        //usecnt++;
        //ast_mutex_unlock(&usecnt_lock);
#ifdef CC_AST_HAS_VERSION_1_4
			ast_atomic_fetchadd_int(&usecnt, 1);
        	//...jitter buffer......ast_jb_configure(tmp, &i->jbconf);
#else//#ifdef CC_AST_HAS_VERSION_1_4
        	ast_mutex_lock(&usecnt_lock);
        	usecnt++;
        	ast_mutex_unlock(&usecnt_lock);
#endif//#ifdef CC_AST_HAS_VERSION_1_4
    
        ast_update_use_count();

        //TODO: em prering tambem nao aloca??
        if (startpbx) {
            if (ast_pbx_start(tmp)) {
                ast_log(LOG_WARNING, "Unable to start PBX on %s\n", tmp->name);
                ast_hangup(tmp);
                i->reserved=0;
                tmp = NULL;
            }
        }
        if (i->verbosity==10)
            ast_log(LOG_NOTICE, "Channel allocated %s\n",tmp->name);
    }
    else
    {
        i->ast_ch = NULL;
        ast_log(LOG_WARNING, "Unable to allocate channel structure\n");
    }

  	return tmp;
}

//-------------------------------------------------------------------------------------------
//
//-------------------------------------------------------------------------------------------
static struct ast_channel *dgv_request(const char *type, int format, void *data, int *cause)
{
	int oldformat;
	struct dgv_pvt *p;
	struct ast_channel *tmp = NULL;
	char *name = strdup(data ? (char *)data : "");
	char *s;
	short port=1;
	short group=-1;
	short backwards=0;
    

#ifdef USE_LIBPRI
	    short dc=0;
    	short spn=0;
#endif//#ifdef USE_LIBPRI

#ifdef DEBUG_Q	
		char szTemp[255];
#endif//#ifdef DEBUG_Q

    if (option_verbose >= 3)
        ast_log(LOG_NOTICE, "Asked to get a channel %s/%s using format %s\n", type, name, ast2dgvformatname(format));

	oldformat = format;
	format &= supported_fmts;
	if (!format) {
		ast_log(LOG_NOTICE, "Asked to get a channel of unsupported format '%s'\n", ast2dgvformatname(oldformat));
		return NULL;
	}
   	//check if it's group or port
    s = strsep(&name, "/");
    if (s) 
    {
        //check for group syntax
        //this version use only sequencial/ascending port hunt
        if (toupper(s[0]) == 'G' || toupper(s[0])=='R') {
            if (option_verbose >= 3)
                ast_log(LOG_NOTICE, "Group %s\n",s);

            group=atoi(s+1);
            if (s[0] == 'G' || s[0] == 'R')			//upper letter use backward search
                backwards = 1;
            else
                backwards = 0;
        }	
        else {
            port = atoi(s); 
            if (option_verbose >= 3)
                ast_log(LOG_NOTICE, "Port %d\n",port);
        }
    }
    else
    {
        if (option_verbose >= 3)
            ast_log(LOG_NOTICE, "Invalid port -> %d\n",port);
        port = 0;				//invalid port
    }

    //keep protected
    ast_mutex_lock(&iflock); {

        if (group != -1)
        {
            p = find_free_port(group,backwards,-1);	//force port to -1 to find using group only
        }    
        else
        {
			//specific port must check if it's free
            //find_ch_port locks iflock, then we need to unlock it first
            ast_mutex_unlock(&iflock);
            p = find_ch_port(port);
            ast_mutex_lock(&iflock);    //and lock it again
            if (p!=NULL)
                if (p->ast_ch != NULL)
                    p = NULL;		//force null to raise busy message
        }

        if (p && !p->reserved)
        {
            p->reserved = 1;
            //setting audio format for recording
            if (p->read_last_format != format)
            {
                if (p->verbosity==10)
                    ast_log(LOG_NOTICE, "dgv_request: (%d) Req. new format: %x (%s)\n",p->port,format,ast2dgvformatname(format));

                //format was changed since last time
                //dg_DisableInputBuffer(p->port);
                dg_PauseInputBuffer(p->port, DG_PAUSE);
    
                //assigns file format
                p->read_last_format = format;
                dg_SetRecordFormat(p->port, ast2dgvformat(p->read_last_format));
            
            }
            if (p->verbosity==10)
                ast_log(LOG_NOTICE, "dgv_request: Calling dgv_new port %d using context=%s\n",p->port,p->context);

#ifdef USE_LIBPRI
            	//set pri structure to this channel
            	if (p->ch_board > 30)
	                spn = 2;
            	else
	                spn = 1;
            	dc = dg_pri_findindex(p->board, spn);
            	if (dc>=0) {
	                p->pri = dchans[dc].pri;
            	}
            	p->outgoing = 1;
#endif//#ifdef USE_LIBPRI

            tmp = dgv_new(p,AST_STATE_RESERVED, p->context,DONT_START_PBX,0);
            
#ifdef DEBUG_Q
	            sprintf(szTemp,"/tmp/reclog%d.raw", p->port);
    	        p->rec = fopen(szTemp,"w");
        	    if (p->rec==NUL L)
            	    ast_log(LOG_ERROR, "Error opening test file!!!");	
                
            	sprintf(szTemp,"/tmp/playlog%d.raw", p->port);
            	p->play = fopen(szTemp,"w");
            	if (p->play==NULL)
	                ast_log(LOG_ERROR, "Error opening test file!!!");	
#endif//#ifdef DEBUG_Q	
        }
        else
        {
            if (option_verbose >= 3)
                ast_log(LOG_WARNING, "dgv_request: Port %d in use or reserved! \n", port);

			if (group != -1)
            	*cause = AST_CAUSE_CONGESTION;			
			else
            	*cause = AST_CAUSE_BUSY;
        }
    } ast_mutex_unlock(&iflock);

  	//free(name);
	return tmp;
}

//-------------------------------------------------------------------------------------------
// Send desired digit to card
//-------------------------------------------------------------------------------------------
#ifdef CC_AST_HAS_VERSION_1_4
	static int dgv_digit_begin(struct ast_channel *ast, char digit)
	{
		struct dgv_pvt *p = (struct dgv_pvt *)ast->tech_pvt;       
	
		if (p)
		{	
 	    	char szDigit[2];        
        
       		szDigit[0] = digit;
       		szDigit[1] = '\0';        
        
        	if (p->verbosity==50)       
            	ast_verbose(VERBOSE_PREFIX_3 "%s:dgv_digit_begin: Asked to play digit[%s]\n", p->dev, szDigit);
    	}
      
    	return 0;
	}
#endif//#ifdef CC_AST_HAS_VERSION_1_4

#if defined(CC_AST_HAS_VERSION_1_4) && defined(CC_AST_HAS_SEND_DIGIT_END_DURATION)
	static int dgv_digit(struct ast_channel *ast, char digit, unsigned int duration)
#else//#if defined(CC_AST_HAS_VERSION_1_4) && defined(CC_AST_HAS_SEND_DIGIT_END_DURATION)
	static int dgv_digit(struct ast_channel *ast, char digit)
#endif//#if defined(CC_AST_HAS_VERSION_1_4) && defined(CC_AST_HAS_SEND_DIGIT_END_DURATION)
	{     
		struct dgv_pvt *p = (struct dgv_pvt *)ast->tech_pvt;
	
    	if (p->verbosity==50)	
#if defined(CC_AST_HAS_VERSION_1_4) && defined(CC_AST_HAS_SEND_DIGIT_END_DURATION)	
			ast_verbose(VERBOSE_PREFIX_3 "%s: dgv_digit: Asked to play digit with duration [%d]\n", p->dev, duration);
#else//#if defined(CC_AST_HAS_VERSION_1_4) && defined(CC_AST_HAS_SEND_DIGIT_END_DURATION)
			ast_verbose(VERBOSE_PREFIX_3 "%s: dgv_digit: Asked to play digit \n", p->dev);
#endif//#if defined(CC_AST_HAS_VERSION_1_4) && defined(CC_AST_HAS_SEND_DIGIT_END_DURATION)
	
		char szDigit[2];

    	ast_mutex_lock(&p->lock); 
    	{    	
    		if (!p->faxhandled)
			{    	
	        	if ( (p->gen_dtmf==1) || 
    	        	 (p->gen_dtmf==2 && ast_bridged_channel(p->ast_ch)==NULL) )
        			szDigit[0] = digit;
        		else
	        		szDigit[0] = ',';
     	
				szDigit[1] = '\0';
        
				if (p->verbosity==50)
	    	    	ast_verbose(VERBOSE_PREFIX_3 "%s: dgv_digit: Asked to play digit[%s]\n", p->dev, szDigit);
        
				//to allow 
    		    if (dg_GetPortInterface(p->port)==DG_DIGITAL_INTERFACE) 
        		{
	        		if (p->verbosity==50)
    	        		ast_verbose(VERBOSE_PREFIX_3 "%s: dgv_digit: DIRECT Digit[%s]\n", p->dev, szDigit);
        	    
            		dg_Dial(p->port, szDigit, 20, dtDirectMF);                
				}
    	    	else
        		{
	        		if ((p->sig==DG_SIG_FXO) && (p->fxo_bridged_dial))
        			{
		        		if (p->verbosity==50)
		            		ast_verbose(VERBOSE_PREFIX_3 "%s: dgv_digit: Don't Dial DTMF [%s] because fxo on bridged dial \n", p->dev, szDigit);        			
        			}
        			else
        			{
		        		if (p->verbosity==50)
		            		ast_verbose(VERBOSE_PREFIX_3 "%s: dgv_digit: DTMF [%s]\n", p->dev, szDigit);
            
    		        	dg_Dial(p->port, szDigit, 20, dgv_conf.default_dialtype);
        			}
				} 	
	    	}
	    	else
    		{
	        	if (p->verbosity==50)
	           		ast_verbose(VERBOSE_PREFIX_3 "dgv_digit: FAX Handling, Dropping DTMF %s on port %d \n", p->dev, p->port);
    		}        	
    	}	 
    	ast_mutex_unlock(&p->lock);
    	return 0;    
	}
	
//-------------------------------------------------------------------------------------------
//
//-------------------------------------------------------------------------------------------
static int dgv_call(struct ast_channel *ast, char *dest, int timeout)
{
	struct dgv_pvt *p = (struct dgv_pvt *)ast->tech_pvt;
	struct dgv_pvt *p_bridged;
	char *c;
  	struct ast_channel *chan_tmp;
	short ret;
	char sCIDToFXS[AST_MAX_EXTENSION] = "";
	
    ast_mutex_lock(&p->lock);   {
    	c = strchr(dest, '/');
    	if ((c) || (!c && p->sig==DG_SIG_FXO))
    	{
            if ((ast->_state != AST_STATE_DOWN) && (ast->_state != AST_STATE_RESERVED)) {
                ast_log(LOG_WARNING, "dgv_call called on %s, neither down or reserved\n", ast->name);//...down nor reserverd
                ast_mutex_unlock(&p->lock);
                return -1;
            }

			if (!c && p->sig==DG_SIG_FXO)
				p->szNumberToDial[0] = '\0';
			else
			{
    			c++;
    			sprintf(p->szNumberToDial, "%s", c);
			}
			
            if (p->verbosity==10)
                ast_log(LOG_NOTICE, "dvg_call: Using card/port %d to dialing to %s - timeout: %d \n", p->port,c, timeout);

    		switch(p->sig)
    		{
    			case DG_SIG_CUSTOMCAS:    			
    			case DG_SIG_FXO:
	                ret = dg_EnableCallProgress(p->port,CP_ENABLE_LINETONE_OR_BUSY);
	                if (ret != DG_EXIT_SUCCESS)
	                {
	                    //could not enabling call progress
	                    dg_DestroyCallProgress(p->port);
	                    ast_log(LOG_WARNING, "Failed to *enable* CallProgress thread - error code = %d\n",ret);
	                    ast_mutex_unlock(&p->lock);
	                    return -1;
	                }
    				break;
    		}
   
    		//copies cid to port_id to allow port to send cid foward
    		if ((p->sig == DG_SIG_R2_NDIS_FIXED) || (p->sig == DG_SIG_R2_NDIS_VARIABLE)) 
    		{
				if (p->verbosity==10)
                {
		        	ast_log(LOG_NOTICE, "dvg_call: p->szPortID = %s \n", p->szPortID);
		            ast_log(LOG_NOTICE, "dvg_call: p->ast_ch->cid.cid_num = %s \n", p->ast_ch->cid.cid_num);
	    	        ast_log(LOG_NOTICE, "dvg_call: p->cid_num = %s \n", p->cid_num);
                }
                
                if ((!ast_strlen_zero(p->szPortID)) && (ast_isphonenumber(p->szPortID)))
					ast_copy_string(p->cid_num, p->szPortID, sizeof(p->cid_num));
                else
                {
	    			if ((!ast_strlen_zero(p->ast_ch->cid.cid_num)) && (ast_isphonenumber(p->ast_ch->cid.cid_num)))
        	          	ast_copy_string(p->cid_num, p->ast_ch->cid.cid_num, sizeof(p->cid_num));
                	else
                		ast_copy_string(p->cid_num, p->szPortID, sizeof(p->cid_num));
                }

                ast_shrink_phone_number(p->cid_num);
            	dg_SetPortId(p->port, p->cid_num);

				if (p->verbosity==10)
                {
		        	ast_log(LOG_NOTICE, "dvg_call: now p->szPortID = %s \n", p->szPortID);
		            ast_log(LOG_NOTICE, "dvg_call: now p->ast_ch->cid.cid_num = %s \n", p->ast_ch->cid.cid_num);
	    	        ast_log(LOG_NOTICE, "dvg_call: now p->cid_num = %s \n", p->cid_num);
                }
    		}   
   
#ifdef USE_LIBPRI    		
    		p->outgoing = 1;    		
    		
    		//copies cid to port_id to allow port to send cid foward
    		if (p->sig == DG_SIG_PRI)
    		{
				if (p->verbosity==10)
                {
		        	ast_log(LOG_NOTICE, "dvg_call: p->szPortID = %s \n", p->szPortID);
		            ast_log(LOG_NOTICE, "dvg_call: p->ast_ch->cid.cid_num = %s \n", p->ast_ch->cid.cid_num);
	    	        ast_log(LOG_NOTICE, "dvg_call: p->cid_num = %s \n", p->cid_num);
	    	        ast_log(LOG_NOTICE, "dvg_call: p->pri_cid_num = %s \n", p->pri_cid_num);
                }
                
                if ((!ast_strlen_zero(p->szPortID)) && (ast_isphonenumber(p->szPortID)))
                  	ast_copy_string(p->pri_cid_num, p->szPortID, sizeof(p->pri_cid_num));
                else
                {
	    			if ((!ast_strlen_zero(p->ast_ch->cid.cid_num)) && (ast_isphonenumber(p->ast_ch->cid.cid_num)))
        	          	ast_copy_string(p->pri_cid_num, p->ast_ch->cid.cid_num, sizeof(p->pri_cid_num));
                	else
                		ast_copy_string(p->pri_cid_num, p->szPortID, sizeof(p->pri_cid_num));
                }

                ast_shrink_phone_number(p->pri_cid_num);
            	dg_SetPortId(p->port, p->pri_cid_num);

				if (p->verbosity==10)
                {
		        	ast_log(LOG_NOTICE, "dvg_call: now p->szPortID = %s \n", p->szPortID);
		            ast_log(LOG_NOTICE, "dvg_call: now p->ast_ch->cid.cid_num = %s \n", p->ast_ch->cid.cid_num);
	    	        ast_log(LOG_NOTICE, "dvg_call: now p->cid_num = %s \n", p->cid_num);
	    	        ast_log(LOG_NOTICE, "dvg_call: now p->pri_cid_num = %s \n", p->pri_cid_num);
                }
    		}
#endif//#ifdef USE_LIBPRI    		
	    			
    		//turn off tonethread of the bridged channel
    		if ((chan_tmp=ast_bridged_channel(ast))!=NULL)
    		{
                if (p->verbosity==10)
                    ast_log(LOG_NOTICE, " bridged -- Port %d dgv_call - tone thread canceled\n",p->port);
    
    			//tell to bridged channel to stop playbuffer
    			if (chan_tmp->tech_pvt!=NULL && !strncmp(chan_tmp->name,type,3))
    			{
    				p_bridged = (struct dgv_pvt *)chan_tmp->tech_pvt;
    				if (p_bridged->tone_thread_id > 0)
    				{
    					p_bridged->tone_thread_running = 0;
    					dg_GenerateMF(p_bridged->port,GENERATE_OFF,0);                       
                        
    					if (p->verbosity==10)
    						ast_log(LOG_NOTICE, "  -- Port %d dgv_call - tone thread canceled\n",p_bridged->port);
    				}
    			}
    		}
            p->call_state = DG_AST_WAITFORDIALTONE;
            
            switch(p->sig)
            {
                case DG_SIG_R2_NDIS_FIXED:
                case DG_SIG_R2_NDIS_VARIABLE:
                case DG_SIG_CUSTOMCAS:
                case DG_SIG_FXO:
                	//set dialtone timeout
                	p->linetone_timeout_sched = ast_sched_add(sched, p->dialtone_timeout, dialtone_timeout_callback, p);
                    break;
#ifdef USE_LIBPRI                    
                case DG_SIG_PRI:
                	//set dialtone timeout
                	p->linetone_timeout_sched = ast_sched_add(sched, 100, dialtone_timeout_callback, p);
                    break;
#endif//#ifdef USE_LIBPRI                    
    			default:
    				break;
            }
            //pickup port
            dg_PickUp(p->port, 20);

            if (p->verbosity==10)
                ast_log(LOG_NOTICE, "dgv_call: dg_PickedUp Port %d \n",p->port);            
    	}
    	else
    	{
			if (p)
			{	
				if (p->sig == DG_SIG_FXS) 
        		{        										
            		if (p->verbosity==10)
                		ast_log(LOG_NOTICE, "dvg_call: Incoming call on port %d \n", p->port);

        			if (p->call_state == DG_AST_NONE)
        			{
        				ast_setstate(p->ast_ch, AST_STATE_RINGING);
        				p->call_state = DG_AST_CALLING;
        				
        				if ((p->ast_ch) && (!ast_strlen_zero(p->ast_ch->cid.cid_num)))//if (strlen(p->cid_num) > 1)
                    	{                    		
                    		sCIDToFXS[0] = 'A';
                    		sCIDToFXS[1] = '1';                 		
                  			strncpy(&sCIDToFXS[2], p->ast_ch->cid.cid_num, strlen(p->ast_ch->cid.cid_num));                  			                    		
                    		sCIDToFXS[strlen(sCIDToFXS)] = 'C';                    		
                    		sCIDToFXS[strlen(p->ast_ch->cid.cid_num) + 3] = '\0';
                    			   
							dg_Dial(p->port, sCIDToFXS, 20, dtTone);

            				if (option_verbose >= 3)							
                				ast_log(LOG_NOTICE, "dgv_call: Sending CallerID(%s) on FXS port %d \n", sCIDToFXS, p->port);        				
                    	}
                    	else
                    	{
	        				dg_SetRing(p->port, DG_ENABLE, p->ring_type);

							struct ast_frame f = { AST_FRAME_CONTROL, AST_CONTROL_RINGING, };
							ast_mutex_lock(&p->lock); 
							{
								QueueFrame(p->port, &f);
							}
							ast_mutex_unlock(&p->lock);						       				
        					
            				if (option_verbose >= 3)        				
                				ast_log(LOG_NOTICE, "dgv_call: Ring started on %d \n", p->port);                   		
                    	}       				
        			}
        			else   
        			{             		
            			ast_log(LOG_WARNING, "dgv_call: Port %d in use or reserved! \n", p->port);
        			}  			
        		}        		
			}  				
    	}
	} ast_mutex_unlock(&p->lock);

	return 0;
}

//-------------------------------------------------------------------------------------------
// Writes a frame of voice data to a digivoice channel 
//-------------------------------------------------------------------------------------------
static int dgv_write(struct ast_channel *ast, struct ast_frame *frame)
{
#ifdef DEBUG_Q
		unsigned char ctmp;
		char szTmp[1024];
		char char_tx[1024];
		char *tmp8;
#endif//#ifdef DEBUG_Q
	short x = 0;
	int remaining_space, total_space;

	struct dgv_pvt *p = (struct dgv_pvt *)ast->tech_pvt;

    //ast_log(LOG_NOTICE, "dgv_write: port %d write_last_format=%x - subclass=%x\n",p->port,p->write_last_format, frame->subclass);

	/*	ast_mutex_lock(&p->lock); */
	if (frame->frametype != AST_FRAME_VOICE /*|| p->faxhandled > 0 */) {  /* don't play anything when in fax stage 1 s*/
		if (p->verbosity==80)
			ast_verbose("%s: dgv_write: Don't know how to handle from type %d\n", ast->name, frame->frametype);

		return 0;
	} 
	else 
		if (!p->ast_ch) 
		{
			dg_StopPlayBuffer(p->port);
            if (p->verbosity==80)
                ast_log(LOG_WARNING, "Dropping frame since there is no active owner on %s...\n",ast->name);
			return 0;
		}
		else 
			if ((ast->_state != AST_STATE_UP) && (ast->_state != AST_STATE_RING))
			{
				dg_StopPlayBuffer(p->port);
				if (p->verbosity==80)
					ast_verbose("%s: dgv_write: Attempt to Write frame type[%d]subclass[%d] on not up chan(state[%d])\n",ast->name, frame->frametype, frame->subclass,ast->_state);
				return 0;
			}
	
	/* Return if it's not valid data */
	if (!frame->data || !frame->datalen)
		return 0;

	if (p)
	{
		if (p->gen_ringback)
		{
			if (p->tone_thread_running)
			{
	            if (p->verbosity==80)
	                ast_log(LOG_WARNING, "Turning off tone_thread, dgv_write received on %d...\n", p->port);
	            
				ast_mutex_lock(&p->lock); 
				{
					p->tone_thread_running = 0;
					dg_GenerateMF(p->port,GENERATE_OFF,0);
				} 
				ast_mutex_unlock(&p->lock);
			}
		}
	}
	
	//check for format to play	
	if (p->write_last_format != frame->subclass)
	{
		//format was changed since last time
		ast_mutex_lock(&p->play_lock); {
			dg_StopPlayBuffer(p->port);
		} ast_mutex_unlock(&p->play_lock);
		//assigns file format
		p->write_last_format = frame->subclass;
		dg_SetPlayFormat(p->port, ast2dgvformat(p->write_last_format));

		if (p->verbosity==80)
			ast_log(LOG_NOTICE, "dgv_write: (%d) Assigning new format:%x %s (%d)\n", 
                        p->port,p->write_last_format,ast2dgvformatname(p->write_last_format), 
                        ast2dgvformat(p->write_last_format));
	}
    //if (p->audio_debug && option_debug==80)
      //  ast_verbose("%s: dgv_write: Attempt to Write frame type[%d]subclass[%d] (state[%d]) - format %x\n",ast->name, frame->frametype, frame->subclass,ast->_state,p->write_last_format);
	switch(ast2dgvformat(p->write_last_format))
	{
		case ffGsm610:case ffWaveALaw:case ffWaveULaw:
				
#ifdef DEBUG_Q
				fwrite(frame->data, frame->datalen,1, p->play);
				
				/*szTmp[0]=0;
				tmp8 = (char *)frame->data;
				strcat(szTmp, "\n");
				for (x=0;x<frame->datalen;x++)
				{
					ctmp = *(char* )(tmp8 + x);
					sprintf(char_tx,"%02x ", ctmp);
					strcat(szTmp, char_tx);
					if (x==32) strcat(szTmp, "\n");
				}
				ast_log(LOG_NOTICE, "dgv_write: %s\n",szTmp);*/
#endif//#ifdef DEBUG_Q

                //if (option_debug==80)
                //    ast_log(LOG_NOTICE, "dgv_write: port %d samples=%d datalen=%d\n",p->port,frame->samples,frame->datalen);

                remaining_space=0;

                /* check buffer size first - passing 0 as size only returns the current remaining_space */
                ast_mutex_lock(&p->play_lock); 
                {                	                	
                	if (dg_GetPortStatus(p->port)!=spDialing)                				
                		x=dg_PlayBuffer(p->port, (char*)frame->data ,0, &remaining_space);
                	else
                        if (p->verbosity==80)
                            ast_log(LOG_NOTICE, "Canceled dg_PlayBuffer because dg_GetPortStatus == spDialing on %d \n", p->port);
                } 
                ast_mutex_unlock(&p->play_lock);
               
                if (p->audio_debug && option_debug == 80 /*&& remaining_space!=6144*/)
                    ast_log(LOG_NOTICE, "dgv_write: port %d samples=%d datalen=%d fmt=%x\n - fr.subclass=%x - remaining space=%d\n",p->port,frame->samples,
                                            frame->datalen,p->write_last_format,frame->subclass, remaining_space);


                /* Tests to keep playbuffer less than 100ms*/
                if (ast2dgvformat(p->write_last_format)==ffGsm610)      /*   gsm    */
                {
                    total_space = WRITE_BUFFER_LINES*SAMPLES_SIZE_GSM;
                    //gsm*8 means > 160ms
                    if (((total_space-remaining_space) > SAMPLES_SIZE_GSM*p->drop_tx_samples_factor) || p->force_drop_tx>0  )   {
                        if (p->verbosity==80)
                            ast_log(LOG_NOTICE, "playbuffer (%d): too much delay gsm - %d ms. Dropping samples. %d %d \n",p->port,
                                                                                    ((total_space-remaining_space)*20)/33, 
                                                                                    p->drop_tx_samples_factor,
                                                                                    p->force_drop_tx);
                        dg_StopPlayBuffer(p->port);
                        p->force_drop_tx=0;
                    }
                    else
                    {
                        //send current buffer
                        ast_mutex_lock(&p->play_lock); 
                        {
                        	if (dg_GetPortStatus(p->port)!=spDialing)			
                        		x=dg_PlayBuffer(p->port, (char*)frame->data ,frame->datalen, NULL);
                        	else	
                        		if (p->verbosity==80)
                            		ast_log(LOG_NOTICE, "Canceled dg_PlayBuffer because dg_GetPortStatus == spDialing on %d \n", p->port);                        		
                        } 
                        ast_mutex_unlock(&p->play_lock);
                    }
                }
                else
                {
                     total_space = WRITE_BUFFER_LINES*SAMPLES_SIZE;
                     //DATALEN*8 means > 160ms
                     if ( ( ((total_space-remaining_space) > DATALEN*p->drop_tx_samples_factor) || p->force_drop_tx>0 ) && (!p->faxhandled) )
                     {
                         if (p->verbosity==80) 
                         {
                             ast_log(LOG_NOTICE, "playbuffer (%d): too much delay g711 - %d ms. - total=%d remaining=%d - Dropping samples.\n",p->port,
                                                                                         (total_space-remaining_space)/8,
                                                                                            total_space, remaining_space);
                         }
                         //stops buffer to drop all buffer content
                         dg_StopPlayBuffer(p->port);
                         p->force_drop_tx=0;
                     }
                     else
                     {
                         //send current buffer
                         ast_mutex_lock(&p->play_lock); 
                         {
                         	if (dg_GetPortStatus(p->port)!=spDialing)	
                         		x=dg_PlayBuffer(p->port, (char*)frame->data ,frame->datalen, NULL);
                         	else
                         		if (p->verbosity==80)
                            		ast_log(LOG_NOTICE, "Canceled dg_PlayBuffer because dg_GetPortStatus == spDialing on %d \n", p->port);	
                         } 
                         ast_mutex_unlock(&p->play_lock);
                     }

                }

				if (x!=DG_EXIT_SUCCESS)
				{
					ast_log(LOG_NOTICE, "playbuffer: full buffer ret=%d - samples=%d datalen=%d remaining=%d\n",x,frame->samples,frame->datalen,remaining_space);
				}	

			break;
		case ffWavePCM:
		        if (p->audio_debug)
			        ast_log(LOG_NOTICE, "dgv_write: wavepcm samples=%d datalen=%d\n",frame->samples,frame->datalen);

                remaining_space=0;

    			ast_mutex_lock(&p->play_lock); 
    			{
    				if (dg_GetPortStatus(p->port)!=spDialing)
    			  		dg_PlayBuffer(p->port, (short*)frame->data ,0, &remaining_space);
    			  	else
                        if (p->verbosity==80)
                            ast_log(LOG_NOTICE, "Canceled dg_PlayBuffer because dg_GetPortStatus == spDialing on %d \n", p->port);    			  		
    			} 
    			ast_mutex_unlock(&p->play_lock);

                total_space = WRITE_BUFFER_LINES*SAMPLES_SIZE;
                // DATALEN*8 means > 160ms
                if ( ( ((total_space-remaining_space) > DATALEN*p->drop_tx_samples_factor) || p->force_drop_tx>0 ) && (!p->faxhandled) )
                {
                    if (p->verbosity==80)
                        ast_log(LOG_NOTICE, "playbuffer (%d): too much delay slin - %d ms. Dropping samples.\n",p->port,
                                                                            (total_space-remaining_space)/8);

                    dg_StopPlayBuffer(p->port);
                    p->force_drop_tx=0;
                }
                else
                {
                    ast_mutex_lock(&p->play_lock); 
                    {
                    	if (dg_GetPortStatus(p->port)!=spDialing)
                    		dg_PlayBuffer(p->port, (short*)frame->data ,frame->samples, NULL);
                    	else
                    	   	if (p->verbosity==80)
                            	ast_log(LOG_NOTICE, "Canceled dg_PlayBuffer because dg_GetPortStatus == spDialing on %d \n", p->port);
                    } 
                    ast_mutex_unlock(&p->play_lock);
                }

			break;
	}
	return 0;
}


//-------------------------------------------------------------------------------------------
// Hangup procedure
//-------------------------------------------------------------------------------------------
static int dgv_hangup(struct ast_channel *ast)
{
    struct dgv_pvt *p = (struct dgv_pvt *)ast->tech_pvt;

	//dg_event_data_structure EventContext;

	if (!ast->tech) 
	{
		ast_log(LOG_ERROR, "dgv_hangup(%s): channel not connected?\n", ast->name);
		return 0;		
		//ast_mutex_unlock(&p->lock);
	}

	if (!p)
	{
		if (option_verbose >= 3)		
			ast_log(LOG_NOTICE, "dgv_hangup(%s): !p \n", ast->name);
		return 0;		
	}		

    if (p->verbosity==10)
        ast_verbose(VERBOSE_PREFIX_3 "Hanging up Port %d \n",p->port);
	
#ifdef DEBUG_Q
	if (p->rec) fclose(p->rec);
	if (p->play) fclose(p->play);
#endif//#ifdef DEBUG_Q	

    ast_mutex_lock(&p->lock); {

    	ast_setstate(ast, AST_STATE_DOWN);
    		
    	if (p->tone_thread_id > 0)
    	{
    		p->tone_thread_running = 0;
    		dg_GenerateMF(p->port,GENERATE_OFF,0);
    		if (p->verbosity==10)
        		ast_log(LOG_NOTICE, "  -- Port %d hangup - tone thread canceled\n",p->port);
    	}
    	
    	if (p->stuttertone_thread_id > 0)
    	{
    		p->stuttertone_thread_running = 0;
    		dg_GenerateMF(p->port, GENERATE_OFF, 0);
    		if (p->verbosity==10)
        		ast_log(LOG_NOTICE, "  -- Port %d hangup - stutter tone thread canceled\n",p->port);
    	}    	
    	
    	ast_mutex_lock(&p->play_lock); {
    		dg_StopPlayBuffer(p->port);
    	} ast_mutex_unlock(&p->play_lock);
    	
    	ast_mutex_lock(&p->record_lock); {
    		//dg_DisableInputBuffer(p->port);
            dg_PauseInputBuffer(p->port, DG_PAUSE);
    	} ast_mutex_unlock(&p->record_lock);
    	
        if (p->echo_can)
        {
            dg_DisableEchoCancelation(p->port);
    		if (p->verbosity==10 || p->verbosity==50)
				ast_log(LOG_NOTICE, "  -- Port %d Disabling Echo Cancelation \n",p->port);            
        }

    	dg_DisableAnswerDetection(p->port);
        dg_DisableCallProgress(p->port);

#ifdef USE_LIBPRI
    		//isdn controls
    		p->alerting = 0;
    		p->outgoing = 0;
    	
			//last parameter is the cause - need to be setted
    	    if (p->pri) 
        	{
            	if (p->alreadyhungup) 
            	{
                	if (p->verbosity==30)
                    	ast_log(LOG_NOTICE, "Already hungup... Calling hangup once, and clearing call\n");
                    
                	if (p->call)
                	{
                    	ast_mutex_lock(&ccs_lock);
                    	{
	                    	pri_hangup(p->pri, p->call, -1);
    	                }
        	            ast_mutex_unlock(&ccs_lock);
            	    }
                	p->call = NULL;
                	p->pri = NULL;
            	}
            	else 
            	{
	                const char *cause = pbx_builtin_getvar_helper(ast,"PRI_CAUSE");
    	            int icause = ast->hangupcause ? ast->hangupcause : -1;
                
        	        if (p->verbosity==30)
            	        ast_log(LOG_NOTICE, "Not yet hungup... Calling hangup once with icause (%d), and clearing call\n", icause);
                    
	                p->alreadyhungup = 1;
                
    	            if (cause) {
        	            if (atoi(cause))
            	            icause = atoi(cause);
                	}

                	if (p->call)
                	{
	                	ast_mutex_lock(&ccs_lock);
    	            	{
	    	            	pri_hangup(p->pri, p->call, icause);
    	    	        }
        	    	    ast_mutex_unlock(&ccs_lock);
                	}
					p->call = NULL;                
            	}
        	}
#endif//#ifdef USE_LIBPRI

		if (p->sig == DG_SIG_FXS)
		{
			if (p->ast_ch)
			{								 
				switch (p->call_state)
				{
					case DG_AST_CALLING:
		                if (p->verbosity==10)
							ast_log(LOG_NOTICE, "dgv_hangup: Disabling SetRing on port %d \n", p->port);
											
						dg_SetRing(p->port, DG_DISABLE, p->ring_type);
						//p->ast_ch = NULL;											
						break;						
					case DG_AST_CONVERSATION:
					case DG_AST_RINGING:
						if (option_verbose >= 3)
        					ast_log(LOG_NOTICE, "Calling dgv_indicate with AST_CONTROL_BUSY on port %d \n", p->port);
        
#ifdef CC_AST_HAS_INDICATE_DATA
							dgv_indicate(p->ast_ch, AST_CONTROL_BUSY, NULL, 0);
#else//#ifdef CC_AST_HAS_INDICATE_DATA
   							dgv_indicate(p->ast_ch, AST_CONTROL_BUSY);    			
#endif//#ifdef CC_AST_HAS_INDICATE_DATA				
						break;
					case DG_AST_DIALING:
					case DG_AST_HANGINGUP:
		                if (p->verbosity==10)
		                	ast_log(LOG_NOTICE, "dgv_hangup: Nothing to do when DIALING or HANGINGUP on port %d \n", p->port);
		                
		                	//ast_log(LOG_NOTICE, "dgv_hangup: Setting p->ast_ch = NULL on port %d \n", p->port);												
						//p->ast_ch = NULL;
						break;			
					default:
		                if (p->verbosity==10)
							ast_log(LOG_NOTICE, "dgv_hangup: p->call_state = %d on port %d \n", p->call_state, p->port);
												
						break;						
				}
			}
		}
		else
		{
			if (option_verbose >= 3)
               	ast_log(LOG_NOTICE, "Calling dg_HangUp on port %d \n", p->port);				
				
			dg_HangUp(p->port);				
		}	

    	//ids strings
    	p->rdnis[0]   = '\0';
    	p->exten[0]   = '\0';
    	p->cid_num[0] = '\0';
    	p->rdnis[0]   = '\0';
    	p->dnid[0]    = '\0';
    
        if (p->reserved>0) 
        {
            if ((p->sig==DG_SIG_FXO) || (p->sig==DG_SIG_CUSTOMCAS) || (p->sig==DG_SIG_FXS) || (p->sig==DG_SIG_PRI))
                p->reserved=0;
            else
                p->reserved=2;      //wait receiving idle
        }

    	p->call_state = DG_AST_NONE;
    	
        p->write_last_format=-1;		//initializes with invalid format
        p->read_last_format=-1;
    	/* unlink digivoice port from channel */
    	
    	//if (p->sig != DG_SIG_FXS)
		p->ast_ch = NULL;   	
    	
    	ast->tech_pvt = NULL;
        p->flashing = 0;
                
        if (p->linetone_timeout_sched > -1) //stop line tone timeout
        {
            ast_sched_del(sched, p->linetone_timeout_sched);
            p->linetone_timeout_sched = -1;
        }

    }  ast_mutex_unlock(&p->lock);


#ifdef CC_AST_HAS_VERSION_1_4
		ast_atomic_fetchadd_int(&usecnt, -1);
#else//#ifdef CC_AST_HAS_VERSION_1_4
		ast_mutex_lock(&usecnt_lock);
		usecnt--;
		ast_mutex_unlock(&usecnt_lock);
#endif//#ifdef CC_AST_HAS_VERSION_1_4

/*    ast_mutex_lock(&usecnt_lock); {
		usecnt--;
	} ast_mutex_unlock(&usecnt_lock);*/


	ast_update_use_count();

	if (p->verbosity==10)
        ast_verbose(VERBOSE_PREFIX_3 "Port %d hangup\n",p->port);

	return 0;
}

//-------------------------------------------------------------------------------------------
//
//-------------------------------------------------------------------------------------------
static int dgv_answer(struct ast_channel *ast)
{
#ifdef DEBUG_Q
    	char szTemp[255];
#endif//#ifdef DEBUG_Q

	struct dgv_pvt *p = (struct dgv_pvt *)ast->tech_pvt;

	int ret=0;

    ast_mutex_lock(&p->lock); 
    {
    	if (p->call_state==DG_AST_RINGING)
    	{
    	    /* stops tone thread */
    	    p->tone_thread_running = 0;
    	    p->stuttertone_thread_running = 0;
    	    
            if (p->sig==DG_SIG_FXS)
            {
				ast_setstate(p->ast_ch, AST_STATE_UP);

				//enable inputbuffer streaming with agc disabled
                dg_PauseInputBuffer(p->port, DG_RELEASE);

                ret = dg_EnableCallProgress(p->port, CP_ENABLE_BUSY_OR_FAX);
                if (ret != DG_EXIT_SUCCESS)
                    ast_log(LOG_NOTICE, "EnableCallProgress port %d - error %d\n",p->port,ret);
                else
                    if (p->verbosity==50)
                        ast_log(LOG_NOTICE, "EnableCallProgress CP_ENABLE_BUSY_OR_FAX port %d - OK\n",p->port);

				p->call_state = DG_AST_CONVERSATION;
            }
            else
            { 
				p->call_state=DG_AST_INCOMING_CALL;
            	if (p->verbosity==50)
                	ast_log(LOG_NOTICE, "Port %d picking up\n",p->port);
                
#ifdef DEBUG_Q
		   			sprintf(szTemp,"/tmp/reclog%d.raw", p->port);
   					p->rec = fopen(szTemp,"w");
   					if (p->rec==NULL)
   						ast_log(LOG_ERROR, "Error opening test file!!!");	
    					
	   				sprintf(szTemp,"/tmp/playlog%d.raw", p->port);
   					p->play = fopen(szTemp,"w");
   					if (p->play==NULL)
	   					ast_log(LOG_ERROR, "Error opening test file!!!");	
#endif//#ifdef DEBUG_Q		
		
#ifdef USE_LIBPRI
	            	ast_mutex_lock(&ccs_lock);
    	        	{
		    	        if (p->verbosity==30)
    		    	        ast_log(LOG_NOTICE,"Answering PRI - %d\n",p->pri_channel);
        	        
						pri_answer(p->pri, p->call, p->pri_channel, 0);
	            	}	
    	        	ast_mutex_unlock(&ccs_lock);
#endif//#ifdef USE_LIBPRI
				if (p->flashing)
   		            dg_PickUp(p->port, 1500);
				else
				{
		            if (p->max_digits_rx==0)
    		            dg_PickUp(p->port, 20);
        		    else
            		    dg_PickUp(p->port, 1000);
				}				
				
	            if (p->verbosity==50)
    	            ast_log(LOG_NOTICE, "dgv_answer: dg_PickedUp Port %d \n",p->port);
            }                    
    	}
    } 
    ast_mutex_unlock(&p->lock);
  
	return 0;
}

static int dgv_fixup(struct ast_channel *oldchan, struct ast_channel *newchan)
{
	struct dgv_pvt *p = (struct dgv_pvt *)newchan->tech_pvt;
	int res = 0;

	ast_mutex_lock(&p->lock);

    if (p->verbosity==10)
        ast_log(LOG_NOTICE, "New owner for channel %s is %s\n", p->dev, newchan->name);

	if (p->ast_ch == oldchan) {
		p->ast_ch = newchan;
	}

	if (newchan->_state == AST_STATE_RINGING){
        if (option_verbose >= 3)
            ast_verbose(VERBOSE_PREFIX_3 "%s: dgv_fixup Calling dgv_indicate\n", p->dev);

        res=ast_mutex_unlock(&p->lock);        
#ifdef CC_AST_HAS_INDICATE_DATA        
        	dgv_indicate(newchan, AST_CONTROL_RINGING, NULL, 0);
#else//#ifdef CC_AST_HAS_INDICATE_DATA
        	dgv_indicate(newchan, AST_CONTROL_RINGING);
#endif//#ifdef CC_AST_HAS_INDICATE_DATA                
	}
    else
        res= ast_mutex_unlock(&p->lock);

	return 0;
}

//-------------------------------------------------------------------------------------------
// Rotina para tratar o que o asterisk passa para a placa
// os sinais devem ser utilizados para gerar RINGBACK, ocupado, etc...
//-------------------------------------------------------------------------------------------
#ifdef CC_AST_HAS_INDICATE_DATA
	static int dgv_indicate(struct ast_channel *ast, int condition, const void *data, size_t datalen)
#else//#ifdef CC_AST_HAS_INDICATE_DATA
	static int dgv_indicate(struct ast_channel *ast, int condition)
#endif//#ifdef CC_AST_HAS_INDICATE_DATA
	{
		struct dgv_pvt *p = (struct dgv_pvt *)ast->tech_pvt;
	
    	pthread_attr_t attr;

		short ret;
	
		if (p == NULL)
			return 0;

		if (p->verbosity==100)
			ast_log(LOG_NOTICE, "dgv_indicate: condition = %d - port=%d - state=%x\n", condition, p->port, p->call_state);		
   
    	switch(condition)
    	{
        	case -1:
            	if (p->verbosity==100)
                	ast_log(LOG_NOTICE, "dgv_indicate -1 : - port %d - flag %d\n",p->port,p->tone_thread_running);
                
            	if (p->tone_thread_running)
            	{
	                ast_mutex_lock(&p->lock); {
                    	p->tone_thread_running = 0;
                    	dg_GenerateMF(p->port,GENERATE_OFF,0);
                	} ast_mutex_unlock(&p->lock);
            	}

	            //check if it is UP to enable samples
            	if (p->call_state ==  DG_AST_CONVERSATION)
            	{
	                ast_mutex_lock(&p->lock); {				  
                    	//ret = dg_EnableInputBuffer(p->port,DG_DISABLE);
                    	ret = dg_PauseInputBuffer(p->port, DG_RELEASE);
                	} ast_mutex_unlock(&p->lock);
                	if (ret != DG_EXIT_SUCCESS)
	                    ast_log(LOG_WARNING, "Error creating InputBuffer - port %d - cod %d\n",p->port, ret);			
                	else
	                    if (p->verbosity==100)
                            ast_log(LOG_NOTICE, ">InputBuffer Created - port %d - cod %d\n",p->port, ret);
            	}
            	break;
        	case AST_CONTROL_CONGESTION:                
            	if (p->verbosity==100)
                	ast_log(LOG_NOTICE, "dgv_indicate (%d): Congestion - generate busy\n",p->port);

            	if (dg_R2SendGroupB(p->port, B_BUSY)==DG_EXIT_SUCCESS)
            		if (option_verbose >= 3)
                		ast_log(LOG_NOTICE, "dgv_indicate: SENDING CONGESTION\n");	

//            if (p->ast_ch)
//                ast_setstate(p->ast_ch, AST_STATE_BUSY);

	            /*f.frametype = AST_FRAME_CONTROL;
    	        f.subclass = AST_CONTROL_HANGUP;
        	    ast_mutex_lock(&p->lock); {
            	    QueueFrame(p->port, &f);
            	} ast_mutex_unlock(&p->lock);*/

            	if (p->gen_busy)
            	{
					p->tone_thread_running = 1;
    	        	//p->tone_thread_type = condition;
        	    	p->tone_thread_type = AST_CONTROL_BUSY;
            		pthread_attr_init(&attr);	
                	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);                
                	if (pthread_create(&p->tone_thread_id, &attr, ToneThread, &(p->port) )<0)
                	{
						p->tone_thread_running = 0;
               		}
	               	else
    	               	if (p->verbosity==100)
        	               	ast_log(LOG_NOTICE, "dgv_indicate: tone_thread congestion created port %d\n",p->port);
            	   	pthread_attr_destroy(&attr);
				}
            	break;
        	case AST_CONTROL_RINGING:        		
            	ast_mutex_lock(&p->lock); {
                	//dg_DisableInputBuffer(p->port);
                	dg_PauseInputBuffer(p->port, DG_PAUSE);
            	} ast_mutex_unlock(&p->lock); 

#ifdef USE_LIBPRI
			if ((!p->alerting) && p->sig==DG_SIG_PRI && p->pri && !p->outgoing && (p->ast_ch->_state != AST_STATE_UP)) 
			{
				if (p->pri)
				{
	                if (p->verbosity==30)
    	                ast_log(LOG_NOTICE, "dgv_indicate: sending pri_acknowledge port %d, pri_channel %d \n", p->port, p->pri_channel);
	                
            		ast_mutex_lock(&p->lock); {
            			pri_acknowledge(p->pri, p->call, p->pri_channel, !p->digital);						
            		} ast_mutex_unlock(&p->lock);
				}
				p->alerting = 1;
			}
#endif//#ifdef USE_LIBPRI           	
            	
	            if (p->call_state==DG_AST_R2_SIGNALLING)
    	        {
        	        dg_R2SendGroupB(p->port, p->group_b_value);
            	}
            	else
            	{
	                if (p->verbosity==100)
    	                ast_log(LOG_NOTICE, "dgv_indicate: tonethread %x %x\n",p->ast_ch->_state, p->call_state);

        	        if ( (p->ast_ch->_state == AST_STATE_UP || p->ast_ch->_state == AST_STATE_RING) && p->gen_ringback) 
            	    {
                	    if (p->tone_thread_running==0)
                    	{                    
	                        p->tone_thread_running = 1;
    	                    p->tone_thread_type = condition;
                        
        	                pthread_attr_init(&attr);	
            	            pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
                	        if (pthread_create(&p->tone_thread_id, &attr, ToneThread, &(p->port) )<0)
                    	    {
                        	    p->tone_thread_running = 0;
                        	}
                        	else
	                            if (p->verbosity==100)
                                	ast_log(LOG_NOTICE, "dgv_indicate: tone_thread created port %d\n",p->port);
	                        pthread_attr_destroy(&attr);
    	                }
        	        }
            	}
            	break;
        	case AST_CONTROL_BUSY:        	        
	            if (p->call_state==DG_AST_R2_SIGNALLING)
    	        {    
                	ast_mutex_lock(&p->lock); {
                    	//dg_DisableInputBuffer(p->port);
                    	dg_PauseInputBuffer(p->port, DG_PAUSE);
                	} ast_mutex_unlock(&p->lock); 
    
                	dg_R2SendGroupB(p->port, B_BUSY);
                	//dg_R2SendGroupB(p->port, B_FREE_CALLING);
                	//dgv_answer(p->ast_ch);

                	if (p->ast_ch)
                	{
                    	ast_setstate(p->ast_ch, AST_STATE_BUSY);
                    	if (option_verbose >= 3)
	                    	ast_log(LOG_NOTICE, "dgv_indicate: BUSY %x %x\n",p->ast_ch->_state, p->call_state);
    	            }
        	        //retirado de dentro do if - avaliar efeitos colaterais
                
            	    //...p->reserved=0;
                	if ((p->sig==DG_SIG_FXO) || (p->sig==DG_SIG_CUSTOMCAS) || (p->sig==DG_SIG_FXS) || (p->sig==DG_SIG_PRI))
                    	p->reserved=0;
                	else
                    	p->reserved=2;      //wait receiving idle

                	struct ast_frame f = { AST_FRAME_CONTROL, AST_CONTROL_HANGUP, };
                	ast_mutex_lock(&p->lock); {
	                    QueueFrame(p->port, &f);
    	            } ast_mutex_unlock(&p->lock);
        	    }
            	else
            	{
                	if (p->gen_busy)
                	{
                    	p->tone_thread_running = 1;
   	                	p->tone_thread_type = condition;
   	                	
       	            	pthread_attr_init(&attr);	
           	        	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
               	    	if (pthread_create(&p->tone_thread_id, &attr, ToneThread, &(p->port) )<0)
                    	{
	   	                    p->tone_thread_running = 0;
                    	}
   	                	else
	       	                if (p->verbosity==100)
           	                	ast_log(LOG_NOTICE, "dgv_indicate: tone_thread busy created port %d\n",p->port);
               	    	pthread_attr_destroy(&attr);
                	}                
                	//schedule to hangup within 10s anyway
                	//ast_channel_setwhentohangup(p->ast_ch, 10);  removed on version 0.9.5_rc5
            	}
            	break;            
        	case AST_CONTROL_ANSWER:
	        	if (p->verbosity==100)//if (option_verbose >= 3)
    	        	ast_log(LOG_NOTICE, "dgv_indicate: AST_CONTROL_ANSWER condition = %d\n",condition);
        	    break;                             
        	case AST_CONTROL_HOLD:
				if (p)
				{		
	        		if (p->verbosity==100)
            			ast_log(LOG_NOTICE, "dgv_indicate: AST_CONTROL_HOLD condition = %d\n",condition);				
				
					if (option_verbose >= 3)
    	        		ast_log(LOG_NOTICE, "dgv_indicate: ast_moh_start on port %d \n", p->port);
            		
#ifdef CC_AST_HAS_INDICATE_DATA                        
	            	ast_moh_start(ast, data, DEFAULT_MOHINTERPRET);
#else//#ifdef CC_AST_HAS_INDICATE_DATA
    	        	ast_moh_start(ast, DEFAULT_MOHINTERPRET);
#endif//#ifdef CC_AST_HAS_INDICATE_DATA
				}						            
            	break;       
        	case AST_CONTROL_UNHOLD:
				if (p)
				{
        			if (p->verbosity==100)        
            			ast_log(LOG_NOTICE, "dgv_indicate: AST_CONTROL_UNHOLD condition = %d\n",condition);
            							
					if (option_verbose >= 3)
            			ast_log(LOG_NOTICE, "dgv_indicate: ast_moh_stop on port %d \n", p->port);
            		
            		ast_moh_stop(ast);            		
				}			                      
            	break;
        	case AST_CONTROL_VIDUPDATE://?????????????????????????
	        	if (option_verbose >= 3)
    	        	ast_log(LOG_NOTICE, "dgv_indicate: AST_CONTROL_VIDUPDATE condition = %d\n",condition);
        	    ast_mutex_lock(&p->lock); {
            	    dg_StopPlayBuffer(p->port);
            	} ast_mutex_unlock(&p->lock);
            	break;          
			case AST_CONTROL_FLASH:
				if (p->verbosity==100)//if (option_verbose >= 3)
            		ast_log(LOG_NOTICE, "dgv_indicate: AST_CONTROL_FLASH condition = %d\n",condition);
				if (p)
				{		
					if (p->verbosity==100)
            			ast_log(LOG_NOTICE, "dgv_indicate: AST_CONTROL_FLASH on port %d \n", p->port);
				}				                    
            	break;
    	}
		return 0;
	}

//-------------------------------------------------------------------------------------------
//
//-------------------------------------------------------------------------------------------
static int __unload_module(int reload)
{

	struct dgv_pvt *p;
	short ret;
	short card;
	
	ast_log(LOG_NOTICE, "  -- Unload module digivoice\n");


	//unregister CLI commands
	switch (dgv_conf.consolelanguage)
	{
		case 0://English
			ast_cli_unregister_multiple(dgv_cli_en, sizeof(dgv_cli_en) / sizeof(dgv_cli_en[0]));
			break;
		case 1://Portuguese
			ast_cli_unregister_multiple(dgv_cli_br, sizeof(dgv_cli_br) / sizeof(dgv_cli_br[0]));
			break;
		default:
			ast_cli_unregister_multiple(dgv_cli_en, sizeof(dgv_cli_en) / sizeof(dgv_cli_en[0]));
			break;		
	}					

	/* First, take us out of the channel loop */
	if (reload==0)
		ast_channel_unregister(&dgv_tech);

	if (!ast_mutex_lock(&iflock)) {
		/* Hangup all interfaces if they have an owner */
		p = iflast;
		while(p) {
			if (p->ast_ch)
			{
				ast_softhangup(p->ast_ch, AST_SOFTHANGUP_APPUNLOAD);
                p->reserved=0;
				ast_log(LOG_WARNING, "ast_softhangup for port %d\n",p->port);		
			}
			p = p->previous;
		}
		ast_mutex_unlock(&iflock);
	} else {
		ast_log(LOG_WARNING, "Unable to lock the monitor\n");
		return -1;
	}

  	//destroy channel linked list
	ast_mutex_lock(&iflock); 
  	{
		/* Destroy all the interfaces and free their memory */
		while(iflast) {
			p = iflast;		    
			ast_mutex_destroy(&p->record_lock);
			ast_mutex_destroy(&p->play_lock);
			ast_mutex_destroy(&p->lock);
            switch(p->sig)
            {
    
                case DG_SIG_R2_NDIS_FIXED:
                case DG_SIG_R2_NDIS_VARIABLE:
                    ret = dg_DestroyE1Thread(p->port);
                    if (ret==DG_EXIT_SUCCESS && p->verbosity==90) {
                        ast_log(LOG_NOTICE, "Destroying E1 thread for port %d\n",p->port);		
                    }
                    break;
                case DG_SIG_PRI:
                    break;
                case DG_SIG_CUSTOMCAS:
                    ret = dg_DestroyCustomCAS(p->port);
                    if (ret==DG_EXIT_SUCCESS && p->verbosity==90) {
                        ast_log(LOG_NOTICE, "Destroying Custom CAS thread for port %d\n",p->port);		
                    }
                    break;
                case DG_SIG_FXO:
                    break;
    			default:
    				break;
            }
            
			/*if ((dg_GetPortInterface(p->port) == DG_DIGITAL_INTERFACE)  && p->sig!=DG_SIG_PRI)
			{
				
				ret = dg_DestroyE1Thread(p->port);
				if (ret==DG_EXIT_SUCCESS && p->verbosity==90) {
					ast_log(LOG_NOTICE, "Destroying E1 thread for port %d\n",p->port);		
				}
			}*/
			//calls all destructor functions
			dg_DisableInputBuffer(p->port);
			dg_StopPlayBuffer(p->port);
            dg_DestroyCallProgress(p->port);

			
			iflast = iflast->previous;
			free(p);
		}
		iflast = NULL;
    } ast_mutex_unlock(&iflock);
#ifdef USE_LIBPRI
	    if (dchans_count>0) {
    	    thd_socket_running = 0;
        	thd_pri_running = 0;

        	for (ret=0; ret<dchans_count; ret++) {
	            //closes ccs device drivers
            	close(dchans[ret].fds[TX]);
            	close(dchans[ret].fds[RX]);
        	}
    	}
#endif//#ifdef USE_LIBPRI
	/* stops voicerlib engine */
	if (reload==0) {
		if (dg_DriverEnabled())
		{			
    		//sets alarmmode for each card
    		for (card = 1; card <= dg_GetCardsCount(); card++)
    		{
        		dg_SetAlarmMode(card, ALARM_MANUAL_NOTIFY);
    		}			
			
			ast_log(LOG_NOTICE, "  -- Shutdown VoicerLib\n");		
			dg_ShutdownVoicerlib();
		}
		ast_mutex_destroy(&iflock);
        
	}
	
  	ast_log(LOG_NOTICE, "  -- module unloaded\n");
	return 0;
}

int unload_module(void)
{
	ast_log(LOG_NOTICE, "unload_module: starting...\n");
	return __unload_module(0);
}

//------------------------------------------------------------
//starts all services and read configuration file
//------------------------------------------------------------
int setup_dgv(int reload)
{
	short ret, ch, card, sSpan;
	struct ast_config *cfg;
	struct ast_variable *v;
	struct dgv_pvt *tmp;
	char szTemp[TAMSTR];
	char szContext[TAMSTR];
	char szPort_ID[TAMSTR] = "";
	char szAccountCode[AST_MAX_ACCOUNT_CODE];		//Account code
	short group=0, sync;
	short category=1,txgain,rxgain,max_rx=4,min_rx=1,cid_pos,echo_can=2,interdigit_timeout=2000, echo_training=800;
    short gen_dtmf=2;
    short dialtone_timeout=3000;
    short wait_dialtone=1;
    short fax_detection=1;
    short dtmf_to_asterisk = 2;
    short gen_ringback=1;
    short gen_busy=1;
	//short auto_r2dmfc=1; //auto is default
	short signalling=0;
    short b_subscriber_hangup=1;
    short group_b_value=B_FREE_CALLING;
    short use_crc4=0;
    short silence_threshold=-36;
    char musicclass[TAMSTR] = "";
    
    char callprogress[TAMSTR];
    char language[MAX_LANGUAGE];
    short drop_tx_samples_factor = TX_CUT_FACTOR;
    short relaxdtmf=1;
    
    //fxs
    int ring_type=1;
    int flash_min_time=80;
    int flash_max_time=650;
    int immediate=0;//fxs and pri channels     
    char mailbox[AST_MAX_EXTENSION] = "1234\0";
        
    ast_group_t cur_pickupgroup = 0;
    pthread_attr_t      attr;

    unsigned int span_index;
    
    //arguments from span option
    int num_args;
    char *args[10];

	szAccountCode[0]='\0';

	cfg = ast_config_load(config);
	/* We *must* have a config file otherwise stop immediately */
	if (!cfg) {
		ast_log(LOG_ERROR,"\n********************************************************************\n");
        ast_log(LOG_ERROR, "FATAL ERROR: Unable to load /etc/asterisk/digivoice.conf (File exists?) *\n");
		ast_log(LOG_ERROR, "********************************************************************\n");
		return -1;
	}  

	/* Make sure we can register our channel type */
	if (reload==0) 
	{
		if (ast_channel_register(&dgv_tech)) {
			ast_log(LOG_ERROR, "Unable to register channel class %s\n", type);
			__unload_module(0);
			return -1;
		}
	}
	
	//default values for codecs
	dgv_tech.capabilities = AST_FORMAT_GSM;
	dgv_conf.allow_slinear=0;
	dgv_conf.allow_ulaw=0;
	dgv_conf.allow_alaw=0;
	dgv_conf.allow_gsm=1;

    //language console
    dgv_conf.consolelanguage = 0;
    
    //fxs_transfer
    features_conf.fxs_atxfer[0] = '\0';
    features_conf.fxs_atxfer[1] = '\0';
    features_conf.fxs_atxfer[2] = '\0';
    features_conf.fxs_atxfer[3] = '\0';
    features_conf.fxs_atxfer[4] = '\0';            
     
	//default values
	dgv_conf.auto_scan_port = 1;

	/* parse configuration file options */
	ast_mutex_lock(&iflock); {
		v = ast_variable_browse(cfg, "general");
		while (v)
		{
			if (strcasecmp(v->name, "auto_port") == 0) {
				dgv_conf.auto_scan_port = atoi(v->value);
				if (dgv_conf.auto_scan_port>=0 && dgv_conf.auto_scan_port<=1)
					ast_log(LOG_NOTICE,"Dgv Channel: scanning ports mode: [%s]\n",auto_scan_port_desc[dgv_conf.auto_scan_port]);
				else
					ast_log(LOG_NOTICE,"Dgv Channel: No scanning ports mode found. Assuming automatic mode\n");
			}
			else if (strcasecmp(v->name, "allow_slinear") == 0) {
				dgv_conf.allow_slinear = atoi(v->value);
				ast_log(LOG_NOTICE,"Dgv Channel: SLINEAR support %s\n", dgv_conf.allow_slinear ? "enabled" : "disabled");
			}
			else if (strcasecmp(v->name, "allow_ulaw") == 0) {
				dgv_conf.allow_ulaw = atoi(v->value);
				ast_log(LOG_NOTICE,"Dgv Channel: U_LAW support %s\n", dgv_conf.allow_ulaw ? "enabled" : "disabled");
			}
			else if (strcasecmp(v->name, "allow_alaw") == 0) {
				dgv_conf.allow_alaw = atoi(v->value);
				ast_log(LOG_NOTICE,"Dgv Channel: A_LAW support %s\n", dgv_conf.allow_alaw ? "enabled" : "disabled");
			}
			else if (strcasecmp(v->name, "allow_gsm") == 0) {
				dgv_conf.allow_gsm = atoi(v->value);
				ast_log(LOG_NOTICE,"Dgv Channel: GSM support %s\n", dgv_conf.allow_gsm ? "enabled" : "disabled");
			}
			else if (strcasecmp(v->name, "consolelanguage") == 0) {
				if (strcasecmp(v->value, "en")==0)
					dgv_conf.consolelanguage = 0;
				else
					dgv_conf.consolelanguage = 1;

				ast_log(LOG_NOTICE, "Dgv Channel: Console Language - %s \n", consolelanguage_desc[dgv_conf.consolelanguage]);				
			}
			v = v->next; /* next item */
		} /* end of while(v)*/

		//register CLI commands	
		switch (dgv_conf.consolelanguage)
		{
			case 0://English		
				ast_cli_register_multiple(dgv_cli_en, sizeof(dgv_cli_en) / sizeof(dgv_cli_en[0]));
				break;
			case 1://Portuguese		
				ast_cli_register_multiple(dgv_cli_br, sizeof(dgv_cli_br) / sizeof(dgv_cli_br[0]));
				break;
			default:		
				ast_cli_register_multiple(dgv_cli_en, sizeof(dgv_cli_en) / sizeof(dgv_cli_en[0]));
				break;
		}

		/* types to use with digivoice cards */
		v = ast_variable_browse(cfg, "allportsconfig");
		while (v)
		{
			if (strcasecmp(v->name, "default_callprogress") == 0) {
					strncpy(dgv_conf.szDefaultCallProgressFile,v->value,TAMSTR);
			}
			else if (strcasecmp(v->name, "detectiontype") == 0) {
				dgv_conf.default_detectiontype = atoi(v->value);

			}
			else if (strcasecmp(v->name, "dialtype") == 0) {
				dgv_conf.default_dialtype = atoi(v->value);
			}
			else if (strcasecmp(v->name, "afterdialpause") == 0) {
				dgv_conf.afterdialpause = atoi(v->value);
			}
			else if (strcasecmp(v->name, "ringbacktone1") == 0) {
				//get configuration name
				strncpy(dgv_conf.szRingbacktone1Name,v->value,TAMSTR);
				dgv_conf.ringbacktone1 = tone_lookup(v->value);
			}
			else if (strcasecmp(v->name, "ringbacktone2") == 0) {
				//get configuration name
				strncpy(dgv_conf.szRingbacktone2Name,v->value,TAMSTR);
				dgv_conf.ringbacktone2 = tone_lookup(v->value);
				
			}
			else if (strcasecmp(v->name, "busytone1") == 0) {
				//get configuration name
				strncpy(dgv_conf.szBusytone1Name,v->value,TAMSTR);
				dgv_conf.busytone1 = tone_lookup(v->value);
			}
			else if (strcasecmp(v->name, "busytone2") == 0) {
				//get configuration name
				strncpy(dgv_conf.szBusytone2Name,v->value,TAMSTR);
				dgv_conf.busytone2 = tone_lookup(v->value);
			}
			else if (strcasecmp(v->name, "ringbacktimes") == 0) {
				//get configuration name
				strncpy(szTemp,v->value,TAMSTR);
				sscanf(szTemp,"%u,%u,%u,%u",&dgv_conf.ringback_times[0],
																		&dgv_conf.ringback_times[1],
																		&dgv_conf.ringback_times[2],
																		&dgv_conf.ringback_times[3]);

			}
			else if (strcasecmp(v->name, "busytimes") == 0) {
				//get configuration name
				strncpy(szTemp,v->value,TAMSTR);
				sscanf(szTemp,"%d,%d,%d,%d",&dgv_conf.busy_times[0],
																		&dgv_conf.busy_times[1],
																		&dgv_conf.busy_times[2],
																		&dgv_conf.busy_times[3]);

			}
			v = v->next; /* next item */
		} /* end of while(v)*/


		if (option_verbose >= 3)
		{
			ast_log(LOG_NOTICE,"Configurations for Dgv Channel:\n");
			ast_log(LOG_NOTICE,"default detectiontype  = %d\n",dgv_conf.default_detectiontype);
			/*ast_log(LOG_NOTICE,"ringbacktone1 = %s (%d)\n",dgv_conf.szRingbacktone1Name, dgv_conf.ringbacktone1);
			ast_log(LOG_NOTICE,"ringbacktone2 = %s (%d)\n",dgv_conf.szRingbacktone2Name, dgv_conf.ringbacktone2);
			ast_log(LOG_NOTICE,"busytone1 = %s (%d)\n",dgv_conf.szBusytone1Name, dgv_conf.busytone1);
			ast_log(LOG_NOTICE,"busytone2 = %s (%d)\n",dgv_conf.szBusytone2Name, dgv_conf.busytone2);
	
			for (ch=0;ch<3;ch++)
				ast_log(LOG_NOTICE,"ringback_times[%d] = %d\n",ch, dgv_conf.ringback_times[ch]);
			for (ch=0;ch<3;ch++)
				ast_log(LOG_NOTICE,"busy_times[%d] = %d\n",ch, dgv_conf.ringback_times[ch]);
	*/
	
			ast_log(LOG_NOTICE,"default_call_progress = %s\n",dgv_conf.szDefaultCallProgressFile);
	
		}
	
		//set audio codecs capabilities
		dgv_tech.capabilities = (dgv_conf.allow_slinear ? AST_FORMAT_SLINEAR : 0) |
									(dgv_conf.allow_ulaw ? AST_FORMAT_ULAW : 0) |
									(dgv_conf.allow_alaw ? AST_FORMAT_ALAW : 0) |
									(dgv_conf.allow_gsm ? AST_FORMAT_GSM : 0) ;
		supported_fmts = dgv_tech.capabilities;
		ast_log(LOG_NOTICE,"audio capabilities = %x\n",supported_fmts);

	} ast_mutex_unlock(&iflock); /* config file parse end-block */	

	/* try to start voicerlib engine */
	if (reload==0) 
	{
		if ((ret = dg_StartVoicerlib(NULL)) != EXIT_SUCCESS)
		{
				ast_log(LOG_ERROR, "\nError starting voicerlib (code 0x%x)\n",ret);
				switch(ret)
				{
					case DG_ERROR_MEMORY_ALLOCATION:

						ast_log(LOG_NOTICE, "Cause: Memory allocation failed (low memory or /dev/vlibd* could not be created)\n");
						break;
					case DG_ERROR_FIRMWARE_NOT_FOUND:
						ast_log(LOG_NOTICE, "Cause: Firmware files not found at /var/lib/voicerlib/firmware!\n");
						break;
				}
				__unload_module(0);
				return -1;
		}
		else
			ast_log(LOG_NOTICE, "Number of ports detected = %d\n",dg_GetPortsCount());

        digivoice_sleep(500);

		sched = sched_context_create();
		if (!sched) {
			ast_log(LOG_ERROR, "Unable to create schedule context.\n");
			return -1;
		}
		else
		{
			ast_cond_init(&timer_pending_cond, NULL);
			
            pthread_attr_init(&attr);	
            pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);                
			if (pthread_create(&timer_thread, &attr, do_timer, NULL) < 0) {
				ast_log(LOG_ERROR, "Unable to start timer thread.\n");
				//ast_sched_del(sched, cdr_sched);
			}
            pthread_attr_destroy(&attr);	
		}

	}

    /* Set default spans to CAS */
    for(span_index=0;span_index<NUM_SPANS;span_index++) {
            span_type[span_index] = DG_SPAN_CAS;
    }


	//E1 configurations
	ast_mutex_lock(&iflock); {
		v = ast_variable_browse(cfg, "e1_config");
		while (v)
		{
			if (strcasecmp(v->name, "e1_sync") == 0) 
			{
				//get configuration name
				sync = atoi(v->value);
				//check values
				if (sync<1 || sync > 3)	//out of range sets internal as default
				{
						ast_log(LOG_WARNING,"e1_sync out of range.Using SYNC_INTERNAL as default\n");
						sync = SYNC_INTERNAL;
				}
			}
			else
            if (strcasecmp(v->name, "use_crc4") == 0) {

                use_crc4 = atoi(v->value);

                if (use_crc4<0 || use_crc4>3) {
                    use_crc4 = 0;      //wrong value means that don't use it
                }
            }
            else
                if (strcasecmp(v->name, "span") == 0) {

                    //span option in cas has only
                    num_args = ast_app_separate_args(v->value, ',' , args, 10);
                    //ast_log(LOG_WARNING,"Num Args = %d\n", num_args);

                    //sscanf(v->value,"%d,%s,%s,%s",&span_index,span_str,span_pritype_str,span_switchtype_str);
                    if (num_args>0) {

                        span_index = atoi(args[DG_SPAN_NUMBER]);
                        if (span_index && span_index<NUM_SPANS) {
                            if (!args[DG_FRAME_TYPE])
                                span_type[span_index-1] = DG_SPAN_CAS;
                            else
                              if (strcasecmp(args[DG_FRAME_TYPE], "ccs")==0) {
                                span_type[span_index-1] = DG_SPAN_CCS;
#ifdef USE_LIBPRI
	                                if (args[DG_SIDE]) {  //net/cpe
    	                                if (strcasecmp(args[DG_SIDE], "net")==0) {
        	                               span_pritype[span_index-1] = PRI_NETWORK;
            	                        }
                	                    else if (strcasecmp(args[DG_SIDE], "cpe")==0) {
                    	                   span_pritype[span_index-1] = PRI_CPE;                                       
                        	            }
                            	    }
                                	else
                                    	span_pritype[span_index-1] = PRI_CPE;

	                                	//switchtype
    	                            if (args[DG_SWITCHTYPE]) {                                    
        	                           	if (!strcasecmp(args[DG_SWITCHTYPE], "national")) 
            	        					span_switchtype[span_index-1] = PRI_SWITCH_NI2;
                    					else if (!strcasecmp(args[DG_SWITCHTYPE], "ni1"))
                    						span_switchtype[span_index-1] = PRI_SWITCH_NI1;
                    					else if (!strcasecmp(args[DG_SWITCHTYPE], "dms100"))
	                    					span_switchtype[span_index-1] = PRI_SWITCH_DMS100;
    	                				else if (!strcasecmp(args[DG_SWITCHTYPE], "4ess"))
        	            					span_switchtype[span_index-1] = PRI_SWITCH_ATT4ESS;
            	        				else if (!strcasecmp(args[DG_SWITCHTYPE], "5ess"))
                	    					span_switchtype[span_index-1] = PRI_SWITCH_LUCENT5E;
                    					else if (!strcasecmp(args[DG_SWITCHTYPE], "euroisdn"))
                    						span_switchtype[span_index-1] = PRI_SWITCH_EUROISDN_E1;
                    					else if (!strcasecmp(args[DG_SWITCHTYPE], "qsig"))
                    						span_switchtype[span_index-1] = PRI_SWITCH_QSIG;
                    					else {
                                        	ast_log(LOG_WARNING, "Unknown switchtype '%s' - assuming euroisdn\n", args[3]);
	                                        span_switchtype[span_index-1] = PRI_SWITCH_EUROISDN_E1;
    	                                }
        	                        }
            	                    else
                	                    span_switchtype[span_index-1] = PRI_SWITCH_EUROISDN_E1;

                    	            if (args[DG_OVERLAPDIAL]) {  //overlapdial
                        	            if (!strcasecmp(args[DG_OVERLAPDIAL], "overlapdial")) 
                            	            span_overlapdial[span_index-1] = 1;
                                	    else
                                    	    span_overlapdial[span_index-1] = 0;
                                	}
                                	else
	                                    span_overlapdial[span_index-1] = 0;

	                                if (args[DG_PRIDIALPLAN]) {  //pridialplan
    	                				if (!strcasecmp(args[DG_PRIDIALPLAN], "national")) {
        	            					span_pridialplan[span_index-1] = PRI_NATIONAL_ISDN + 1;
            	        				} else if (!strcasecmp(args[DG_PRIDIALPLAN], "unknown")) {
                	    					span_pridialplan[span_index-1] = PRI_UNKNOWN + 1;
                    					} else if (!strcasecmp(args[DG_PRIDIALPLAN], "private")) {
                    						span_pridialplan[span_index-1] = PRI_PRIVATE + 1;
                    					} else if (!strcasecmp(args[DG_PRIDIALPLAN], "international")) {
                    						span_pridialplan[span_index-1] = PRI_INTERNATIONAL_ISDN + 1;
                    					} else if (!strcasecmp(args[DG_PRIDIALPLAN], "local")) {
                    						span_pridialplan[span_index-1] = PRI_LOCAL_ISDN + 1;
	                    	 			} else if (!strcasecmp(args[DG_PRIDIALPLAN], "dynamic")) {
    	                 					span_pridialplan[span_index-1] = -1;
        	            				} 
            	                        else
                	                        span_pridialplan[span_index-1] = PRI_NATIONAL_ISDN + 1;
                    	            }
	
    	                            if (args[DG_PRILOCALDIALPLAN]) {  //prilocaldialplan
        	                            if (!strcasecmp(args[DG_PRILOCALDIALPLAN], "national")) {
            	                            span_prilocaldialplan[span_index-1] = PRI_NATIONAL_ISDN + 1;
                	                    } else if (!strcasecmp(args[DG_PRILOCALDIALPLAN], "unknown")) {
                    	                    span_prilocaldialplan[span_index-1] = PRI_UNKNOWN + 1;
                        	            } else if (!strcasecmp(args[DG_PRILOCALDIALPLAN], "private")) {
                            	            span_prilocaldialplan[span_index-1] = PRI_PRIVATE + 1;
                                	    } else if (!strcasecmp(args[DG_PRILOCALDIALPLAN], "international")) {
                                    	    span_prilocaldialplan[span_index-1] = PRI_INTERNATIONAL_ISDN + 1;
	                                    } else if (!strcasecmp(args[DG_PRILOCALDIALPLAN], "local")) {
    	                                    span_prilocaldialplan[span_index-1] = PRI_LOCAL_ISDN + 1;
        	                            } else if (!strcasecmp(args[DG_PRILOCALDIALPLAN], "dynamic")) {
            	                            span_prilocaldialplan[span_index-1] = -1;
                	                    } 
                    	                else
                        	                span_prilocaldialplan[span_index-1] = PRI_NATIONAL_ISDN + 1;
                            	    }
#endif//#ifdef USE_LIBPRI
                            }
                            else
                                span_type[span_index-1] = DG_SPAN_CAS;
    
                        }
                    }//num_args>0
                    else
                    {
                        //wrong arguments
                        //assume cas span to span[0]
                        span_type[0] = DG_SPAN_CAS;
                    }
                }
            else
			if (strcasecmp(v->name, "cards") == 0) 
			{
				//set card confifurations - need to restart
                
                if (reload<2) 
                {
                    if (set_card_config(v->value, sync, use_crc4)==0)
                    {
                        ast_log(LOG_ERROR, "\n************************************************\nFATAL ERROR - Please check [e1_config]\nsection at /etc/asterisk/digivoice.conf and\nif dgvfifo device driver is loaded.\n************************************************\n");
                        
    					//sets alarmmode for each card
    					for (card = 1; card <= dg_GetCardsCount(); card++)
    					{
        					dg_SetAlarmMode(card, ALARM_MANUAL_NOTIFY);
    					}                       
                        
                        dg_ShutdownVoicerlib();
                        return -1;

                    }

                }
			}
			v = v->next; /* next item */
		} /* end of while(v)*/
	} ast_mutex_unlock(&iflock); /* config file parse end-block */	

#ifdef USE_LIBPRI
    	//starts socket thread if there are spans in ccs mode
    	if (dchans_count>0 && reload==0) {
        	//starts socket thread
        	pthread_attr_init(&attr);
        	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
        	if (pthread_create(&thd_sock, &attr, dgfifo_thread, NULL)) {
	            ast_log(LOG_ERROR,"Error starting ccs fifo control thread\n");
        	}
        	pthread_attr_destroy(&attr);

	        if (thd_pri==0) {
	        	//starts pri thread
            	pthread_attr_init(&attr);
            	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);            	
            	if (pthread_create(&thd_pri, &attr, dchan, NULL)) {
	                ast_log(LOG_ERROR,"ERROR: Cannot start dchannel control thread!!!\n");
            	}
            	pthread_attr_destroy(&attr);
        	}
    	}
#endif//#ifdef USE_LIBPRI
	ast_log(LOG_NOTICE, "Loading ports........................ \n");
	//create private structure for each port
	if (reload==0) 
	{
		for (ch=1 ; ch<=dg_GetPortsCount() ; ch++) 
		{
			tmp = initialize_ports(ch);
			if (tmp)
			{
				if (ch==1) iffirst = tmp;	//save bottom list address

				/* port allocated */
				if (iflast!=NULL) iflast->next = tmp;
				tmp->previous = iflast;
				iflast = tmp;
			} 
			else 
			{
				ast_log(LOG_ERROR, "Unable to register channel '%s'\n", v->value);
			}
		}
	}
	ast_log(LOG_NOTICE, "Waiting synchronization....................\n");
	//digivoice_sleep(3000);
	
    //sets alarmmode for each card
    ast_log(LOG_NOTICE, "SetAlarmMode to ALARM_AUTOMATIC_NOTIFY \n");
    for (card = 0; card <= dg_GetCardsCount() - 1; card++)
    {
        dg_SetAlarmMode(card + 1, ALARM_AUTOMATIC_NOTIFY);		        
     
        for (sSpan = 0; sSpan <= 1; sSpan++)
        {
        	//reset e1alarm status      
			dgv_e1alarms[card][sSpan].nALARM_LOSS     = 0;   
			dgv_e1alarms[card][sSpan].nALARM_AISS     = 0;
			dgv_e1alarms[card][sSpan].nALARM_AIS16S   = 0;
			dgv_e1alarms[card][sSpan].nALARM_SYNC     = 0;
			dgv_e1alarms[card][sSpan].nALARM_MFSYNC   = 0;
			dgv_e1alarms[card][sSpan].nALARM_RAIS     = 0;  
  			dgv_e1alarms[card][sSpan].nALARM_CRC4SYNC = 0;
			dgv_e1alarms[card][sSpan].nALARM_RSLIP    = 0;   	
        }    
    }

    //Enable card framers
    //e1 card initializations
    
    //enable first the E1-a off all cards

    
/*    for (card = 1; card <= dg_GetCardsCount(); card++)
    {
        if ((dg_GetCardInterface(card) == DG_DIGITAL_INTERFACE)) 
        {
            ast_log(LOG_NOTICE, "Enabling Framer A of card %d...............\n", card);
            dg_EnableFramer(card, CFG_FRAMER_E1_A);
        }
    }
    
    digivoice_sleep(5000);*/
    //enable first the E1-a off all cards
    
/*    for (card = 1; card <= dg_GetCardsCount(); card++)
    {
        if ((dg_GetCardInterface(card) == DG_DIGITAL_INTERFACE)) 
        {
            ast_log(LOG_NOTICE, "Enabling Framer B of card %d...............\n", card);
            dg_EnableFramer(card, CFG_FRAMER_E1_B);
        }
    }*/
    digivoice_sleep(500);        

	//--------------------------------------------------
	//check individual port configuration
	//--------------------------------------------------
	ast_mutex_lock(&iflock); {
		v = ast_variable_browse(cfg, "port_config");
		while (v)
		{
			if (strcasecmp(v->name, "port_id") == 0) {
				strncpy(szPort_ID,v->value,TAMSTR);
			}
			else
			if (strcasecmp(v->name, "category") == 0) {
				category = atoi(v->value);
			}
			else
			if (strcasecmp(v->name, "max_digits_rx") == 0) {
				max_rx = atoi(v->value);
			}
			else
			if (strcasecmp(v->name, "min_digits_rx") == 0) {
				min_rx = atoi(v->value);
			}
			else
			if (strcasecmp(v->name, "interdigit_timeout") == 0) {
				interdigit_timeout = atoi(v->value);
			}
			else
			if (strcasecmp(v->name, "tx_gain") == 0) {
				txgain = atoi(v->value);
			}
			else
			if (strcasecmp(v->name, "rx_gain") == 0) {
				rxgain = atoi(v->value);
			}
			else
			if (strcasecmp(v->name, "cid_pos") == 0) {
				cid_pos = atoi(v->value);
			}
            else
			if (strcasecmp(v->name, "echocan") == 0) {
				echo_can = atoi(v->value);
			}
            else
            if (strcasecmp(v->name, "echotraining") == 0) {
                echo_training = atoi(v->value);
            }
            else
            if (strcasecmp(v->name, "drop_tx_samples_factor") == 0) {
                drop_tx_samples_factor = atoi(v->value);
            }
            else
            if (strcasecmp(v->name, "relaxdtmf") == 0) {
                relaxdtmf = atoi(v->value);
            }
            else
            if (strcasecmp(v->name, "wait_dialtone") == 0) {
                wait_dialtone = atoi(v->value);
            }
            else
            if (strcasecmp(v->name, "dialtone_timeout") == 0) {
                dialtone_timeout = atoi(v->value);
            }
            else
            if (strcasecmp(v->name, "gen_ringback") == 0) {
                gen_ringback = atoi(v->value);
            }
            else
            if (strcasecmp(v->name, "gen_busy") == 0) {
                gen_busy = atoi(v->value);
            }
            else
            if (strcasecmp(v->name, "gen_dtmf") == 0) {
                gen_dtmf = atoi(v->value);
            }
            else
            if (strcasecmp(v->name, "dtmf_to_asterisk") == 0) {
                dtmf_to_asterisk = atoi(v->value);
            }
            else
            if (strcasecmp(v->name, "silence_threshold") == 0) {
                silence_threshold = atoi(v->value);
            }
            else
            if (strcasecmp(v->name, "group_b_value") == 0) {
                group_b_value = atoi(v->value);

                //wrong value sets B_FREE_CALLING
                if (group_b_value < B_FREE_CALLING || group_b_value >  B_OUT_OF_SERVICE) 
                    group_b_value = B_FREE_CALLING;
            }
			else
            if (strcasecmp(v->name, "b_subscriber_hangup") == 0) {
                b_subscriber_hangup = atoi(v->value);

            }
            else
            if (strcasecmp(v->name, "callprogress") == 0) {
                 if (!ast_strlen_zero(v->name))
                     strlcpy(callprogress,v->value,TAMSTR);
                 else
                     strlcpy(callprogress,dgv_conf.szDefaultCallProgressFile,TAMSTR);
            }
            else
            if (strcasecmp(v->name, "language") == 0) {
                if (!ast_strlen_zero(v->name))
                    strlcpy(language,v->value,MAX_LANGUAGE);
                else
                    strlcpy(language, global_language, MAX_LANGUAGE);
            }
            else
			if (strcasecmp(v->name, "accountcode") == 0) {
				if (!ast_strlen_zero(v->name))
					strncpy(szAccountCode,v->value,AST_MAX_ACCOUNT_CODE);
				else
					szAccountCode[0]='\0';
			}
            else
            if (!strcasecmp(v->name, "musiconhold")) 
            {
                ast_copy_string(musicclass, v->value, sizeof(musicclass));
            }
			else
            if (!strcasecmp(v->name, "pickupgroup")) 
            {
                cur_pickupgroup = ast_get_group(v->value);
            } else 
			if (strcasecmp(v->name, "context") == 0) {
                   
				strncpy(szContext,v->value,TAMSTR);

                if (szContext[0]=='\0')
					strcpy(szContext,"default\0");
			}
			else
			if (strcasecmp(v->name, "signalling") == 0) {
                
                if (!strcasecmp(v->value, "r2mfc_ndis_fixed")) {
                   signalling = DG_SIG_R2_NDIS_FIXED;
                }
                else if (!strcasecmp(v->value, "r2mfc_ndis_variable")) {
                   signalling = DG_SIG_R2_NDIS_VARIABLE;
                }
#ifdef USE_LIBPRI
                	else if (!strcasecmp(v->value, "isdn")) {
                   		signalling = DG_SIG_PRI;
                	}
#endif//#ifdef USE_LIBPRI
                else if (!strcasecmp(v->value, "customcas")) {
                   signalling = DG_SIG_CUSTOMCAS;
                }
                else if (!strcasecmp(v->value, "fxo")) {
                   signalling = DG_SIG_FXO;
                }
                else if (!strcasecmp(v->value, "fxs")) {
                   signalling = DG_SIG_FXS;              	
                }
                else
                   signalling = DG_SIG_FXO; /* unkown signalling - set default FXO ?? */
			}
			else
#ifdef USE_LIBPRI
	            if (!strcasecmp(v->name, "internationalprefix")) 
    	        {
        	        ast_copy_string(internationalprefix, v->value, sizeof(internationalprefix));
            	}
            	else
            	if (!strcasecmp(v->name, "nationalprefix")) {
	                ast_copy_string(nationalprefix, v->value, sizeof(nationalprefix));
            	} 
            	else
            	if (!strcasecmp(v->name, "localprefix")) {
	                ast_copy_string(localprefix, v->value, sizeof(localprefix));
            	} 
            	else 
            	if (!strcasecmp(v->name, "privateprefix")) {
	                ast_copy_string(privateprefix, v->value, sizeof(privateprefix));
            	}
            	else
            	if (!strcasecmp(v->name, "unknownprefix")) {
	                ast_copy_string(unknownprefix, v->value, sizeof(unknownprefix));
				} 
				else
#endif//#ifdef USE_LIBPRI			
			if (!strcasecmp(v->name, "usecallerid")) {
				use_callerid = ast_true(v->value);
			}
			else			
            if (strcasecmp(v->name, "auto_r2dmfc") == 0) {
                ast_log(LOG_WARNING, 
                "\n+-------------------------------------------------------------+\n"
                "|   WARNING !!!                                               |\n"
                "|   You have an old version of digivoice.conf.                |\n"
                "|   'auto_r2dmfc' was replaced by 'signallig' configuration.  |\n"
                "|   Please, refer to CHANGELOG to update your digivoice.conf. |\n"
                "+-------------------------------------------------------------+\n");

                /*__unload_module(0);
                return -1;*/
                if (atoi(v->value))
                    signalling = DG_SIG_R2_NDIS_FIXED; 
                else
                    signalling = DG_SIG_R2_NDIS_VARIABLE;                 
            }
            else
            if (!strcasecmp(v->name, "customcascfg")) 
            {           	
                ast_copy_string(customcascfg, v->value, sizeof(customcascfg));
            }
            else
            if (strcasecmp(v->name, "ring_type") == 0) 
            {
                ring_type = atoi(v->value);                            		
            }
            else
            if (strcasecmp(v->name, "flash_min_time") == 0) 
            {
                flash_min_time = atoi(v->value);            		
            }
            else
            if (strcasecmp(v->name, "flash_max_time") == 0) 
            {
                flash_max_time = atoi(v->value);            		
            }
            else
			if (!strcasecmp(v->name, "immediate")) 
			{
				immediate = ast_true(v->value);
			}
            else
            if (!strcasecmp(v->name, "mailbox")) 
            {
                ast_copy_string(mailbox, v->value, sizeof(mailbox));
            }
            else
            if (!strcasecmp(v->name, "faxdetection")) 
            {
                fax_detection = atoi(v->value);
            }           
			else
			if (strcasecmp(v->name, "ports") == 0) 
			{
				//set port - reload=0 new one
				//1 - already allocated				
				set_port_config(reload, v->value, szPort_ID, szContext,category, min_rx, max_rx, interdigit_timeout, txgain,rxgain,
				 			    cid_pos,echo_can,szAccountCode, signalling, b_subscriber_hangup, group_b_value, silence_threshold,
                                musicclass, cur_pickupgroup, gen_dtmf, gen_ringback, callprogress, language, dtmf_to_asterisk, 
                                dialtone_timeout, wait_dialtone, gen_busy, drop_tx_samples_factor, relaxdtmf, echo_training,
                                ring_type, flash_min_time, flash_max_time, immediate, mailbox, fax_detection);
			}
			v = v->next; /* next item */
		} /* end of while(v)*/
	} ast_mutex_unlock(&iflock); /* config file parse end-block */

	//check group configuration
	ast_mutex_lock(&iflock); {
		v = ast_variable_browse(cfg, "groups");
		while (v)
		{
			if (strcasecmp(v->name, "group") == 0) {
				group = atoi(v->value);
			}
			else
			if (strcasecmp(v->name, "ports") == 0) {
				//find port
				set_port_group(group, v->value);
			}
			v = v->next; /* next item */
		} /* end of while(v)*/
	} ast_mutex_unlock(&iflock); /* config file parse end-block */	

	//set events callback
	if (reload==0) 
	{
		//set event handler function
		dg_SetEventCallback(event_handler,&event_context);
		//input buffer callback
		dg_SetAudioInputCallback(read_samples);
		//DEFAULT TYPE OF GSM files
		dg_SetGSMMode(GSM_RAW);
	}

	ast_config_destroy(cfg);
	
	/* We *must* have a features.conf file to transfer a call */
	cfg = ast_config_load(featurescfg);	
	if (!cfg)
	{
		ast_log(LOG_ERROR,"\n**********************************************************************\n");
        ast_log(LOG_ERROR, "FATAL ERROR: Unable to load /etc/asterisk/features.conf (File exists?)\n");
		ast_log(LOG_ERROR, "**********************************************************************\n");		
	}
	
	ast_mutex_lock(&iflock); 
	{
		v = ast_variable_browse(cfg, "featuremap");
		while (v)
		{
            if (!strcasecmp(v->name, "atxfer"))         	
                ast_copy_string(features_conf.fxs_atxfer, v->value, sizeof(features_conf.fxs_atxfer));            

			v = v->next; /* next item */
		} /* end of while(v)*/
		
		if (features_conf.fxs_atxfer[0] == '\0')
			ast_log(LOG_WARNING, "Dgv Channel: FXS transfer command not found!!! \n");
		else              
			ast_log(LOG_NOTICE, "Dgv Channel: atxfer: %s \n", features_conf.fxs_atxfer);		
	} 
	ast_mutex_unlock(&iflock); /* config file parse end-block */

	ast_config_destroy(cfg);	

	channel_up = 1;
	
	ast_log(LOG_NOTICE, "Requesting alarm notifications... \n\n");
	for (card = 1; card <= dg_GetCardsCount(); card++)
	{
		dg_GetAlarmStatus(card);
	}	
	
	return 0;
}

//-------------------------------------------------------------------------------------------
// This functions is called by ast_register_atexit 
//-------------------------------------------------------------------------------------------
void dgv_cleanup(void)
{
	channel_up = 0;
	ast_log(LOG_NOTICE, "dgv_cleanup: Start VoicerLib Shutdown...\n");
	__unload_module(0);
}

//-------------------------------------------------------------------------------------------
//
//-------------------------------------------------------------------------------------------
int load_module()
{
	ast_log(LOG_NOTICE, "load_module: starting for Dgv\n");
	ast_register_atexit(dgv_cleanup);
	ast_mutex_init(&iflock);
    ast_mutex_init(&timer_pending_lock);
#ifdef USE_LIBPRI
	    ast_mutex_init(&ccs_lock);
#endif//#ifdef USE_LIBPRI
	return setup_dgv(0);
}

int reload(void)
{
	channel_up = 0;
	ast_log(LOG_NOTICE, "Starting reload procedure for Dgv...\n");
	usleep(300*1000);
	setup_dgv(2);
	
	return 0;
}

/* this function isn't used but freepbx voicemail drop the call
if ast_read isn't created */
static struct ast_frame  *dgv_read(struct ast_channel *ast)
{	
//	struct dgv_pvt *p = (struct dgv_pvt *)ast->tech_pvt;

	static struct ast_frame f = { AST_FRAME_NULL, }; 

	f.src = (char *)type;
	return &f;
}

#ifdef CC_AST_HAS_VERSION_1_4
	AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_DEFAULT, tdesc,
    			    .load = load_module,
       				.unload = unload_module,
       				.reload = reload,
           			);
#else//#ifdef CC_AST_HAS_VERSION_1_4
	//-------------------------------------------------------------------------------------------
	//
	//-------------------------------------------------------------------------------------------
	int usecount()
	{
		return usecnt;
	}
	
	//-------------------------------------------------------------------------------------------
	//
	//-------------------------------------------------------------------------------------------
	char *description()
	{
		return (char *) desc;
	}
	
	//-------------------------------------------------------------------------------------------
	//
	//-------------------------------------------------------------------------------------------
	char *key()
	{
		return ASTERISK_GPL_KEY;
	}
#endif//#ifdef CC_AST_HAS_VERSION_1_4	
